/* ----------------------------------------------------------------------
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic
 
  Purpose: PPP client for real-time and post-procesing positioning applications
  Version: $ Rev: $
 
  2012-05-11 /JD: created
 
-*/

#include "gset/gcfg_sothis.h"

using namespace std;
using namespace gnut;

static int stop = 0;

void catch_signal(int){ stop = 1; cout << "Program interrupted by Ctrl-C [SIGINT,2]\n"; }      
void usage(const string& app){
  cout << app  << endl;
  cout << endl << "Usage: "
       << endl
       << endl << "    -h                     .. this help                          "
       << endl << "    -V int                 .. version                            "
       << endl << "    -v int                 .. verbosity                          "
       << endl << "    -x xml-file            .. configuration                      "
       << endl << "    -X                     .. default configuration              "
       << endl << endl;
  exit(0);
}

// MAIN
// ----------
int main(int argc, char** argv)
{
  signal(SIGINT,  catch_signal);

  t_gcfg_sothis gset;
  gset.app("G-Nut/Sothis", "0.9.0", "$Rev: 1058 $","(gnss@pecny.cz)", __DATE__, __TIME__);

  string conf("");
  t_glog glog;
    glog.tsys(t_gtime::GPS);
    glog.mask("/dev/stdout");  
    glog.verb(0);  
    glog.cache_size(99);
    gset.glog(&glog);

  int  maxepo  =  2880;  // maximum number of epochs in obs container

  // READ COMMAND-LINE
  for( int i=1; i<argc; i++ ){
    if (  !strcmp(argv[i],"--help") && i  <argc ){ usage(gset.app()); exit(0); }
    else if ( !strcmp(argv[i],"-h") && i  <argc ){ usage(gset.app()); exit(0); }
    else if ( !strcmp(argv[i],"-X") && i  <argc ){ gset.help(); exit(0); }
    else if ( !strcmp(argv[i],"-V") && i  <argc ){ cout << gset.app() << endl; exit(0); }
    else if ( !strcmp(argv[i],"-v") && i+1<argc ){ glog.verb(atoi(argv[++i])); }
    else if ( !strcmp(argv[i],"-x") && i+1<argc ){ conf    = argv[++i]; }

    else fprintf(stderr,"Unknown option: %s\n",argv[i]);
  }

  // READ XML SETTINGS
  if( conf.empty() ){ if( gset.read_istream(cin) ){ gset.help(); usage(gset.app()); exit(1); }}
  else{               if( gset.read(conf)        ){ gset.help(); usage(gset.app()); exit(1); }}
  gset.glog( &glog );

  t_gtime beg = dynamic_cast<t_gsetgen*>(&gset)->beg();
  t_gtime end = dynamic_cast<t_gsetgen*>(&gset)->end();

  set<string> sites = dynamic_cast<t_gsetgen*>(&gset)->rec();
  multimap<IFMT,string> inp = gset.inputs_all();
  int sample = dynamic_cast<t_gsetgen*>(&gset)->sampling();
  if(!sample) sample = dynamic_cast<t_gsetgen*>(&gset)->sampling_default();
   
  // COMMENTS
  glog.comment(0,"main","==========");
  glog.comment(0,"main","program = " + gset.app());
  glog.comment(1,"main","setting = " + conf);
  glog.comment(1,"main","logfile = " + glog.mask());
  glog.comment(1,"main","verbose = " + int2str(glog.verb()));
   
  glog.comment(1,"main","maxepo  = " + int2str(maxepo));
  glog.comment(1,"main","sample  = " + int2str(sample));
  glog.comment(1,"main","==========");

   
  // DECLARATIONS
  // ============
  vector<t_gio*>           gio;
  vector<t_gcoder*>     gcoder;
  vector<boost::thread*>  thrd;
  vector<t_gpppflt*>     vgppp;
  vector<t_gpreproc*>    vgpre;
  vector<boost::thread*> vtppp;
   
  t_gtime runepoch(t_gtime::GPS);
  t_gtime lstepoch(t_gtime::GPS);

  t_gallobs*    gobs  = new t_gallobs();
  t_gallpcv*    gpcv  = 0;
  t_gallnav*    gorb  = 0;
  t_gallotl*    gotl  = 0;
  t_gallprod*   gprd  = 0;
  t_gdata*      gdata = 0;
   
  if( gset.input_size("atx")  > 0 ) gpcv = new t_gallpcv;
  if( gset.input_size("blq")  > 0 ) gotl = new t_gallotl;
   
  t_gallobj*    gobj  = new t_gallobj(gpcv, gotl); gobj->glog( &glog );

  // set objects
  // ---> NOT UNIFIED XML+RINEXO OBJECT SETTINGS!
  // ---> TEMPORARY -> IF ADDED HERE, CANNOT BE CHANGED FROM V RINEX DECODERs!
  // =========================================================================
//  set<string>::const_iterator itOBJ;
//  set<string> obj = dynamic_cast<t_gsetrec*>(&gset)->objects();
//  for( itOBJ = obj.begin(); itOBJ != obj.end(); ++itOBJ ){
//    string name = *itOBJ;
//    shared_ptr<t_grec> rec = dynamic_cast<t_gsetrec*>(&gset)->grec(name);
//    rec->glog(&glog);
//    gobj->add( rec );
//  }

  // CHECK INPUTS
  if(       gset.input_size("sp3")    > 0 &&
            gset.input_size("rinexc") > 0 &&
            gset.input_size("rinexo") > 0 
  ){
    gorb = new t_gallprec; // glog.comment(0,"main", "GALLNAV type: t_gallprec");
    glog.append(false);
  }else{
     glog.comment(0,"main", "Error: incomplete input: rinexo+rinexc+sp3");
     glog.comment(0,"main", "Usage: ./" + gset.pgm() + " --default");
     usage(gset.app());
  }       

  if( gpcv ) gpcv->glog(&glog);
  if( gotl ) gotl->glog(&glog);
  if( gorb ) gorb->glog(&glog);
  if( gobs ) gobs->glog(&glog);

  runepoch = t_gtime::current_time(t_gtime::GPS);  // still returns full seconds (dsec=0.0)

  // DATA COLLECTING && PROCESSING
  // =============================
  // - single loop for offline processing (file://...)
  // - infinite loop (until stop=1) for real-time
   
  multimap<IFMT,string>::const_iterator itINP = inp.begin();
  for( size_t i = 0; i < inp.size() && itINP != inp.end(); ++i, ++itINP ){

     
    IFMT   ifmt(itINP->first);
    string path(itINP->second);
    string id("ID" + int2str(i));
    thrd.push_back(0);

    if(       ifmt == SP3_INP     ){ gdata = gorb; gcoder.push_back(new t_sp3("",8172));
    }else if( ifmt == RINEXO_INP  ){ gdata = gobs; gcoder.push_back(new t_rinexo("",4096));
    }else if( ifmt == RINEXC_INP  ){ gdata = gorb; gcoder.push_back(new t_rinexc("",4096));
    }else if( ifmt == RINEXN_INP  ){ gdata = gorb; gcoder.push_back(new t_rinexn("",4096));
    }else if( ifmt == ATX_INP     ){ gdata = gpcv; gcoder.push_back(new t_atx("",4096));
    }else if( ifmt == BLQ_INP     ){ gdata = gotl; gcoder.push_back(new t_blq("",4096));
//  }else if( ifmt == BNCRTCM_INP ){ gdata = gorb; gcoder.push_back(new t_bncrtcm("",4096));
//  }else if( ifmt == BNCOBS_INP  ){ gdata = gobs; gcoder.push_back(new t_bncobs("",4096));
    }else{ 
      glog.comment(0,"main","Error: unrecognized format " + int2str(ifmt) );
      gcoder.push_back(0);
      gdata = 0;
    }         
   
    if( path.substr(0,7) != "file://"  ){
      glog.comment(0,"main","invalid gio path - file supported only!");
      gio.push_back(0);
    }
     
    gio.push_back(new t_gfile);
    gio[i]->glog(&glog); 	 
    gio[i]->path(path);
        
    if( gcoder[i] && gio[i] ){	   
	   
      gio[i]->coder( gcoder[i] );
	   
      gcoder[i]->add_data(id, gdata);
      gcoder[i]->add_data("OBJ", gobj);	   
      gcoder[i]->glog(&glog);
      gcoder[i]->gset(&gset);	   

      runepoch = t_gtime::current_time(t_gtime::GPS);
      gio[i]->run_read();
      lstepoch = t_gtime::current_time(t_gtime::GPS);
      glog.comment(0, "main", "READ: " + path + " time: " + dbl2str( lstepoch.diff(runepoch) ) + " sec");
    }
  }
   
  // --------------     
  glog.comment(0, "main", "Filter started [PP]");

  // Start loop over all sites from xml
  for (set<string>::iterator it = sites.begin(); it != sites.end(); ++it)
  {
    runepoch = t_gtime::current_time(t_gtime::GPS);
    string site = *it;
    if( gobs->isSite(site) == false ){
       glog.comment(1, "main", site + runepoch.str_ymdhms(" observations not found: "));
       continue;
    }
    vgppp.push_back(0);
    vgpre.push_back(0);
    vtppp.push_back(0);
    int idx = vgppp.size() - 1;
    vgppp[idx] = new t_gpppflt(site, &gset);
    vgppp[idx]->setDAT(gobs, gorb);
    vgppp[idx]->setOUT(gprd);
    vgppp[idx]->setOTL(gotl);
//  vgppp[idx]->setPCV(gpcv);
    vgppp[idx]->setOBJ(gobj);
	  
    t_gtime epo(t_gtime::GPS);
    t_gtime beg_obs = gobs->beg_obs(site);
	  
    gobj->read_satinfo(beg_obs);
	  
    // pre-processing
    vgpre[idx] = new t_gpreproc(gobs, &gset); 
    vgpre[idx]->glog(&glog);
    vgpre[idx]->setNav(gorb);
    vgpre[idx]->preprocess(site, gobs->beg_obs(site), gobs->end_obs(site), sample);
    glog.comment(0, "main", site + " processing thread started [" + int2str(idx) + "]");
    glog.comment(0, "main", site + gobs->beg_obs(site).str_ymdhms("  beg: ")
                                 + gobs->end_obs(site).str_ymdhms("  end: "));

    vtppp[idx] = new boost::thread(&t_gpppflt::processBatch, vgppp[idx], gobs->beg_obs(site), gobs->end_obs(site));
  }
   
  // JOINING THREAD
  for( vector<boost::thread*>::iterator it = vtppp.begin(); it != vtppp.end(); ++it){
    if( *it ) (*it)->join();
  }

//for (int i=0; i<vgppp.size(); i++) vgppp[i]->debugPrint();
  lstepoch = t_gtime::current_time(t_gtime::GPS);
  glog.comment(0,"main","Filter ended [PP] duration " 
          + dbl2str( lstepoch.diff(runepoch)) + " sec");       

  // CLEANING
  for( size_t i = 0; i < inp.size(); i++ ){
     if ( gcoder[i] ) delete gcoder[i];
     if ( thrd[i]   ) delete thrd[i];
     if( gio[i] ){
       if(gio[i]->running() != 0 ) { 
         glog.comment(0,"main","client thread stopped: " + gio[i]->path() );
         gio[i]->stop(); t_gtime::gmsleep(1000);
         while( gio[i]->running() != 0 ) t_gtime::gmsleep(1000);
      }
      delete gio[i];
    }     
  }
  for( unsigned int i = 0; i < vgpre.size(); ++i ) { if (vgpre[i])  delete vgpre[i]; }
  for( unsigned int i = 0; i < vgppp.size(); ++i ) { if (vgppp[i])  delete vgppp[i]; }   
  for( unsigned int i = 0; i < vtppp.size(); ++i ) { if (vtppp[i])  delete vtppp[i]; }

  if( gobs ) delete gobs;
  if( gpcv ) delete gpcv;
  if( gotl ) delete gotl;
  if( gobj ) delete gobj;    
  if( gorb ) delete gorb;
  if( gprd ) delete gprd;

}
