
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <sstream>

#include "gall/gallobs.h"

using namespace std;

namespace gnut {  

// constructor
// ----------
t_gallobs::t_gallobs() 
  : t_gdata(),
    _nepoch(0),
    _overwrite(false)
{
  gtrace("t_gallobs::constructor");
  id_type(  t_gdata::ALLOBS );
  id_group( t_gdata::GRP_OBSERV );
}


// destructor
// ----------
t_gallobs::~t_gallobs()
{
  gtrace("t_gallobs::destructor");
  _mapobj.clear();
}


// Attribute/signal selection (quantitative priority)
// --------------------------------------------------
GOBSATTR t_gallobs::select_attr( const t_map_stt_sys& stt, GSYS gs, GOBSTYPE gt, GOBSBAND gb )
{   
  gtrace("t_gallobs::select_attr");

  GOBSATTR ga = ATTR;

  if( stt.find(gs) == stt.end() ) return ga;

  map<GOBSATTR,int> signal_sel;

  t_map_stt_obs::const_iterator itGOBS;
  t_map_stt_sat::const_iterator itSAT;

  for( itGOBS = stt.at(gs).begin(); itGOBS != stt.at(gs).end(); ++itGOBS )
  {
    GOBS gobs = itGOBS->first;
    string go = gobs2str(gobs);
     
    if( gb != str2gobsband(go) ||
        gt != str2gobstype(go)   ) continue;

    for( itSAT = stt.at(gs).at(gobs).begin(); itSAT != stt.at(gs).at(gobs).end(); ++itSAT )
    {
      signal_sel[str2gobsattr(go)] += itSAT->second.first;
    }     
  }

  // maximize
  int max_count = 0;
  for( map<GOBSATTR,int>::const_iterator it = signal_sel.begin(); it != signal_sel.end(); ++it )
  { 
    if( it->second > max_count ){ 
      ga = it->first;
      max_count = it->second;
//    cout << " minimize 1 : " << max_count << " " << gobsattr2str( ga ) << " " << gobsband2str( gb ) << endl;

    }
  }
   
  return ga;
}


// Observation statistics
// ----------------------
t_gallobs::t_map_stt_sys t_gallobs::stat_obs( string site,
				              double* smp,
                                              const t_gtime& beg,
			                      const t_gtime& end,
		 	                      t_map_sky_sat* sky_sat,
		 	                      t_map_stt_smp* smp_stt,
		 	                      t_map_stt_sys* sky_stt
					    )
{
  gtrace("t_gallobs::stat_obs");
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

   t_map_stt_smp smp_stt_loc;

   double sample = 0.0;
   if( smp ) sample = (*smp);

   t_map_stt_sys stat_obs;

   if( _mapobj.find(site) == _mapobj.end() ){ _gmutex.unlock(); return stat_obs; }
   
   t_map_stt_sys::const_iterator itGSYS;
   t_map_stt_sat::const_iterator itPRN;
   t_map_stt_obs::const_iterator itGOBS;
   
   t_map_oref::const_iterator itEpo = _mapobj[site].begin(); // start
   t_map_osat::const_iterator itSat;

   vector<GOBS>::iterator itVOBS;
   map<GSYS, map<string, int> > tot, all;
   t_gtime tt     = itEpo->first;
   t_gtime tt_sav = itEpo->first;
   set<string> list_sat, save_sat;
   
   for( ; itEpo != _mapobj[site].end(); itEpo++ ) // loop over epochs in interval
   {
      tt = itEpo->first;   
      
      if( tt > end ) break; // stop

      double dint = fabs(tt-tt_sav); 
      if( dint > 0.0 ){             // collect sampling histogram
	double tmp = round(dint);   // round to full seconds -> NEED A CHANGE IF >1Hz
        smp_stt_loc[tmp] += 1;
        tt_sav = tt;
      }

      list_sat.clear();
      for( itSat = _mapobj[site][tt].begin(); itSat != _mapobj[site][tt].end(); itSat++ ) // loop over satellites
      {
	 GSYS   sys = itSat->second->gsys();
	 string sat = itSat->second->sat();
#ifdef DEBUG	 
    cout << " ok sat [" << sat << "] " << tt.str_ymdhms() << " " << dint << endl;
#endif
	 double increment = 1;
	 list_sat.insert(sat);
	 
         // handle requested sampling (consider missing epochs with no elevation support)
	 if( sample != 0.0 && fabs(dint - sample) > DIFF_SEC ){  // POTENTIAL PROBLEM (only 1Hz now) NEED A CHANGE IF >1Hz
	   if( dint < sample ){                                  // skip the epoch
#ifdef DEBUG	      
              if( sat[0] == 'G' )
	      cout << " skip this epoch due to sample: " 
		   << fixed << setprecision(15)
		   << setw(20) << dint << setw(20) << sample << setw(20) << fabs(dint - sample) << endl;
#endif	      
	      continue;
	   }else if( save_sat.find(sat) != save_sat.end() ){  // satellite continuously observed (expect gap)
 	       increment = static_cast<int>(dint/sample);     // handle expected observation rate
#ifdef DEBUG
               cout << " gap for sat [" << sat << "] " << tt.str_ymdhms() << " " << dint << " " << sample << endl;
#endif    
	   }
	 }
	 
	 tot[sys][sat] += increment;
         tot[sys]["0"] += increment;
//
	 // INSTEAD check the satellites visibility (if available)
	 if( sky_sat && sky_sat->find(sat) != sky_sat->end() )
	 {
           map<t_gtime, t_gtime>::const_iterator itSKY = (*sky_sat)[sat].upper_bound(tt);
	   if( itSKY != (*sky_sat)[sat].begin() ) --itSKY;
	   if( tt < itSKY->first ||
	       tt > itSKY->second ){
#ifdef DEBUG
	        cout << "satellite not visible: " << sat << tt.str_ymdhms(" ") << endl;
#endif	        
	   }else all[sys]["0"] += increment;
         }

	 // existing GOBS in the epoch
	 vector<GOBS> obs = itSat->second->obs();
	 for( unsigned int i=0; i < obs.size(); i++ ){
	   if( itSat->second->getobs(obs[i]) != 0.0 ){
	     stat_obs[sys][obs[i]][sat].first += 1;

	     // add histogram of elevations if requested 
	     double ele = itSat->second->getele();
	     if( sky_stt != 0 && ele > 0 ){
	       int e = static_cast<int>(ele/10)*10;
	       string key = int2str( e );
 	       (*sky_stt)[sys][obs[i]][key].first += 1;
	     }
	   }
	 }

	 // expected GOBS considered the same for each satellite and system
	 for( itGOBS = stat_obs[sys].begin(); itGOBS != stat_obs[sys].end(); ++itGOBS ){
	   stat_obs[sys][itGOBS->first][sat].second = tot[sys][sat];
	    
	   // add histogram of elevations if requested 
	   int i = 0;
 	   while( sky_stt && i < 90 ){
             string key = int2str(i);

	     // fixed problem with all[sys] when not all satellite elevations are found (then use tot instead!)
             (*sky_stt)[sys][itGOBS->first][key].second = (all[sys]["0"] > tot[sys]["0"]) ? all[sys]["0"] : tot[sys]["0"];
	     i += 10;
	   }
	 }
      }
      save_sat.clear();
      save_sat = list_sat;
   }
  
   // identify sampling interval (auto)
   if( smp ){
     double count = 0;
     for( t_map_stt_smp::iterator it = smp_stt_loc.begin(); it != smp_stt_loc.end(); ++it ){
	
       if( smp_stt ) (*smp_stt)[it->first] = it->second; // copy to output parameter
	
       if( it->second > count ){
	  *smp = it->first;
	 count = it->second;
       }
     }
   }

   _gmutex.unlock(); return stat_obs;
}


// return set of available GNSS systems
// --------------------------
set<GSYS> t_gallobs::sys(string site)
{
  gtrace("t_gallobs::sys");
   t_map_oref::const_iterator itEpo;
   t_map_osat::iterator itSat;
   set<GSYS> temp;
   t_gtime t;
   
   for( itEpo = _mapobj[site].begin(); itEpo != _mapobj[site].end(); ++itEpo) // loop over epochs
   {
     t = itEpo->first;
     for( itSat = _mapobj[site][t].begin(); itSat != _mapobj[site][t].end(); ++itSat) // loop over satellites
     {
	temp.insert(itSat->second->gsys());
     }
   }   
   return temp;
}

// return used systems for epoch
// -------------------------------
set<GSYS> t_gallobs::sys(const string& site, const t_gtime& tt)
{
  gtrace("t_gallobs::sys");
  set<GSYS> all_sys;

  vector<t_gsatdata> satdata = obs(site, tt);
  _gmutex.lock();  
 
  vector<t_gsatdata>::iterator it;
  for (it = satdata.begin(); it != satdata.end(); it++){
//    cout << it->epoch().str_hms() << " " << it->sat() << endl;
    GSYS gsys = it->gsys();
    all_sys.insert(gsys);
  }

  _gmutex.unlock(); return all_sys;

}

// return list of available stations
// ----------
set<string> t_gallobs::stations()
{
  gtrace("t_gallobs::stations");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  set<string> all_sites;
  t_map_oobj::const_iterator itSITE = _mapobj.begin();
   
  while( itSITE != _mapobj.end() ){
    all_sites.insert(itSITE->first);
    itSITE++;
  }
  _gmutex.unlock(); return all_sites;
}


// return list of available satellites
// ----------
set<string> t_gallobs::sats(const string& site, 
                            const t_gtime& t, GSYS gnss)
{
  gtrace("t_gallobs::sats");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  set<string> all_sats;

  if( _mapobj.find(site)    == _mapobj.end() ||
      _mapobj[site].find(t) == _mapobj[site].end() ){ _gmutex.unlock(); return all_sats; }

  t_map_osat::const_iterator itSAT = _mapobj[site][t].begin();  
  while( itSAT != _mapobj[site][t].end() ){
    GSYS sys = itSAT->second->gsys();
    if(gnss == GNS) all_sats.insert(itSAT->first);
    else if(gnss == GPS && sys == GPS) all_sats.insert(itSAT->first);
    else if(gnss == GLO && sys == GLO) all_sats.insert(itSAT->first);
    else if(gnss == GAL && sys == GAL) all_sats.insert(itSAT->first);
    else if(gnss == BDS && sys == BDS) all_sats.insert(itSAT->first);
    else if(gnss == QZS && sys == QZS) all_sats.insert(itSAT->first);
    else if(gnss == SBS && sys == SBS) all_sats.insert(itSAT->first);     
    itSAT++;
  }
  _gmutex.unlock(); return all_sats;
}


// return list of available satellites
// ----------
vector<t_gsatdata> t_gallobs::obs(const string& site,
                                  const t_gtime& t)
{
  gtrace("t_gallobs::obs");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  vector<t_gsatdata> all_obs;
  t_gtime tt(t_gtime::GPS);

  if (_find_epo(site, t, tt) < 0) {
     _gmutex.unlock(); return all_obs;
  }

  t_map_osat::iterator itSAT = _mapobj[site][tt].begin();
  while( itSAT != _mapobj[site][tt].end() ){    
//    all_obs.push_back( itSAT->second ); 
    all_obs.push_back( *itSAT->second );
    itSAT++;
  }

  _gmutex.unlock(); return all_obs;
}


// TEMPORARY !!!
// return list of available satellites (POINTERS!)
// ----------
vector<t_spt_gobs> t_gallobs::obs_pt(const string& site, const t_gtime& t)
{
  gtrace("t_gallobs::obs_pt");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  vector<t_spt_gobs> all_obs;
  t_gtime tt(t_gtime::GPS);
   
  if( _find_epo(site, t, tt) < 0 ){
    _gmutex.unlock(); return all_obs;
  }
               
  t_map_osat::iterator itSAT = _mapobj[site][tt].begin();
  while( itSAT != _mapobj[site][tt].end() ){
     
    string sat = itSAT->first;
     
    // TESTING NEW METHOD
    all_obs.push_back( dynamic_pointer_cast<t_gobsgnss>(itSAT->second) );
 
    itSAT++;
  }

  _gmutex.unlock(); return all_obs;
}


// return single difference between satellites
// ---------
/*t_gsdbs t_gallobs::sdbs(const string& site,
                        const t_gtime& t)
{
   t_gsdbs diff;
   
   vector<t_gsatdata> gobs = this->obs(site, t);   
   diff.pivot( *gobs.begin() );
      
   for (unsigned int i = 1; i < gobs.size(); i++){
      gobs[i] = gobs[i] - gobs[0];      
   }        
   
   gobs.erase(gobs.begin()); // erase reference observation

   diff.data(gobs);
   
   return diff;
}
*/


// return list of available satellites
// ----------
vector<t_gtime> t_gallobs::epochs(const string& site)
{
  gtrace("t_gallobs::epochs");
   
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  vector<t_gtime> all_epochs;

  if( _mapobj.find(site) == _mapobj.end() ){
    // cout << "site not found: " << site << t.str_ymdhms() << endl; cout.flush();
     _gmutex.unlock(); return all_epochs;
  }
     
  t_map_oref::iterator it = _mapobj[site].begin();

  while( it != _mapobj[site].end() ){
    all_epochs.push_back( it->first );
    it++;
  }
  
  _gmutex.unlock(); return all_epochs;
}


// return first position for satellite
// ----------
t_gtime t_gallobs::beg_obs(const string& site)
{
  gtrace("t_gallobs::beg_obs");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  t_gtime tmp = t_gtime::last_time();
  if( _mapobj.find(site)    != _mapobj.end() &&
      _mapobj[site].begin() != _mapobj[site].end() ) tmp = _mapobj[site].begin()->first;

  _gmutex.unlock(); return tmp;
}


// return last position for satellite
// ----------
t_gtime t_gallobs::end_obs(const string& site)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  t_gtime tmp = t_gtime::first_time();
  if( _mapobj.find(site)    != _mapobj.end() &&
      _mapobj[site].begin() != _mapobj[site].end() )
    tmp = _mapobj[site].rbegin()->first;

  _gmutex.unlock(); return tmp;
}


// add observations ( both P and L in meters !!!)
// ----------
int t_gallobs::addobs(t_spt_gobs obs)
{
  gtrace("t_gallobs::addobs");
   
  t_gtime tt(obs->epoch());
  string site = obs->site();
  string sat  = obs->sat();
//  cout << "ADDING site: " << site << endl;

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  t_map_oref::const_iterator itEPO = _mapobj[site].find(tt);
  t_map_osat::const_iterator itSAT = _mapobj[site][tt].find(sat);

  // add new observations (or overwrite)
  // ===================================
  if( _overwrite || itEPO == _mapobj[site].end()        
                 || itSAT == _mapobj[site][tt].end()
  ){
#ifdef DEBUG
    cout << " ADDING " << site << " " <<  _mapobj[site].size() << " " << _nepoch // << " " <<  nepochs(site) // MUTEX !!!
         << " sat "    <<  sat << " " << tt.str_ymdhms() << " " << "\n";cout.flush();
#endif

    // delete old if exists
    // ====================
    if( _overwrite && itEPO != _mapobj[site].end()
                   && itSAT != _mapobj[site][tt].end()
    ){
        _mapobj[site][tt].erase(sat);
        if( _log && _log->verb() >= 2 ) _log->comment(2, "gallobs", site + tt.str_ymdhms(" obs replaced ") + sat);
    }

    // too many epochs (remove old)
    // ============================
    if( _nepoch > 0 && _mapobj[site].size() > _nepoch + 10 ){  // +10 .. reduce number of removals to every tenths

      t_map_oref::iterator itEPO = _mapobj[site].begin();
      while( itEPO != _mapobj[site].end() ){

        if( _mapobj[site].size() <= _nepoch ){ break; }
	
        t_gtime t = itEPO->first;
        if( _log && _log->verb() >= 2 )
	{	 	      
           string satells(" ");
           t_map_osat::iterator itSAT;
           for( itSAT = _mapobj[site][t].begin(); itSAT != _mapobj[site][t].end(); ++itSAT ){
             satells = satells + " " + itSAT->first;
           }
       
           _log->comment(2, "gallobs", site + t.str_ymdhms(" obs removed ") + satells);
	}

        _mapobj[site][t].erase(_mapobj[site][t].begin(),_mapobj[site][t].end());
        _mapobj[site].erase(itEPO++);
      }
    }

    // distinguish GNSS specific message
    // =================================
    if( obs->id_type() == t_gdata::OBSGNSS ){

      _mapobj[site][tt][sat] = obs;

    }else{
       if( _log ){  _log->comment(0,"gallobs","warning: t_gobsgnss record not identified!"); }
       _gmutex.unlock(); return 1;
    }
  }else{
    if( _log && _log->verb() >= 2 ) _log->comment(2,"gallobs", site + tt.str_ymdhms(" skipped ") + " " + sat);
    _gmutex.unlock(); return 0;
  }

  // comments
  if( _log && _log->verb() >= 3 ){

    vector<GOBS> v_obs = obs->obs();
    vector<GOBS>::iterator itOBS = v_obs.begin();
    ostringstream lg; lg << fixed << setprecision(3);
    for( ; itOBS != v_obs.end(); ++itOBS ) 
      lg << " " << gobs2str( *itOBS ) << ":" << obs->getobs( *itOBS );
  
    _log->comment(3,"gallobs",site + tt.str_ymdhms(" add obs: ") + " " + sat + lg.str() );
  }
  _gmutex.unlock(); return 0;
}


// clean function
// ----------
void t_gallobs::clean_outer( const  string& obj,
                             const t_gtime& beg, 
                             const t_gtime& end )
{
  gtrace("t_gallobs::clean_outer");
  if( end < beg ) return;

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  // loop over all object [default]
  t_map_oobj::const_iterator itOBJ  = _mapobj.begin();
  t_map_oobj::const_iterator itOBJ2 = _mapobj.end();
    
  if( _log && _log->verb() >= 2 ){
    _log->comment(2,"gallobs","obs clean request: " + obj + beg.str_ymdhms(": ") + end.str_ymdhms(" - "));
  }
     
  // loop over a single object (if obj defined !)
  if( ! obj.empty() ) itOBJ = itOBJ2 = _mapobj.find(obj);
  if( itOBJ2 != _mapobj.end() ) itOBJ2++;

  for( ; itOBJ != itOBJ2; ++itOBJ ){

    string site = itOBJ->first;

    // find and CLEAN all data (epochs) out of the specified period !
    t_map_oref::iterator it;
    t_map_oref::iterator itFirst = _mapobj[site].begin();
    t_map_oref::iterator itLast  = _mapobj[site].end();
    t_map_oref::iterator itBeg   = _mapobj[site].lower_bound(beg);  // greater|equal
    t_map_oref::iterator itEnd   = _mapobj[site].upper_bound(end);  // greater only!

    t_map_osat::iterator itOBS;

    // delete before BEGIN request
    if( itBeg != itFirst ){
      if( _log && _log->verb() >= 3  ){
        string lg("obs removed before: " + itBeg->first.str_ymdhms() + " " + site);
        _log->comment(3,"gallobs",lg);
      }

      for( it = itFirst; it != itBeg; ++it ){	 
        t_gtime tt(it->first);
	 
        _mapobj[site][tt].erase(_mapobj[site][tt].begin(),_mapobj[site][tt].end());
      }
       
      _mapobj[site].erase(itFirst,itBeg); // erase all but itBeg
    }

    // delete after END request
    if( itEnd != itLast ){ // && ++itEnd != itLast ){
      if( _log && _log->verb() >= 3 ){
        string lg("obs removed after : " + itEnd->first.str_ymdhms() + " " + site);
        _log->comment(3,"gallobs",lg);
      }

      for( it = itEnd; it != itLast; ++it ){
	    
        t_gtime tt(it->first);
	 
        _mapobj[site][tt].erase(_mapobj[site][tt].begin(),_mapobj[site][tt].end());	 
      }                   
      _mapobj[site].erase(itEnd,itLast);
    }
//    itOBJ++; // WHILE LOOP ONLY
  }
  _gmutex.unlock(); return;
}


// return list of available satellites
// ----------
t_gallobs::t_map_osat t_gallobs::find( string site, const t_gtime& t)
{
  gtrace("t_gallobs::find");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  t_gtime tt(t_gtime::GPS);
  t_map_osat  tmp;
  if(_find_epo(site, t, tt) < 0){
     _gmutex.unlock(); return tmp;
  }

//  shared_ptr<t_map_osat> tmp = &_mapobj[site][tt];
  _gmutex.unlock(); return _mapobj[site][tt]; // tmp;
}


// number of epochs for station
// ----------
unsigned int t_gallobs::nepochs(const string& site)
{
  gtrace("t_gallobs::nepochs");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  if( _mapobj.find(site) == _mapobj.end() ){ _gmutex.unlock(); return 0; }

/*
  if( site == "VSBO0" ){     
    t_map_oref::iterator itEPO = _mapobj[site].begin();
    while( itEPO != _mapobj[site].end() ){
       
      t_gtime t = itEPO->first;
      cout << site << " --- time: " << t.str_ymdhms() << " " << t.dsec() << endl;
      itEPO++;
    }
  }
*/
  unsigned int tmp = _mapobj[site].size();
  _gmutex.unlock(); return tmp;
}

// number of epochs for station, interval and sampling, list of OBS types
// ----------
unsigned int t_gallobs::nepochs(const string& site, const t_gtime& beg, const t_gtime& end, const int& sampl, map<GSYS, pair<int,int> >& n)
{
  gtrace("t_gallobs::nepochs");

  if( _mapobj.find(site) == _mapobj.end() ) return 0;   
   
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
  
   n.clear();
   
   t_gtime tt = beg;
   while (tt < end || tt == end)
   {
      // JE TREBA UPRAVIT!!!! MUTEX
#ifdef BMUTEX
      lock.unlock();
      vector<t_gsatdata> satdata = obs(site, tt);
      lock.lock();
#else
      _gmutex.unlock();
      set<GSYS> gnss = sys(site, tt);
      _gmutex.lock();
#endif      

      set<GSYS>::iterator itS;
      for(itS = gnss.begin(); itS != gnss.end(); itS++){
        n[*itS].second += 1;	 
      }
      
      tt = tt + sampl;
   }   

   map<GSYS, pair<int, int> >::iterator it;
   for (it = n.begin(); it != n.end(); it++){
     it->second.first = floor( (end - beg) / sampl + 0.5) + 1;
   }

   _gmutex.unlock(); return 1;
}


// gaps number
// ---------------------
int t_gallobs::gaps(string site, int gap, map<t_gtime, int>& len, double& min, double& max)
{
  gtrace("t_gallobs::gaps");

  if( _mapobj.find(site) == _mapobj.end() ) return 0;

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

   min = 86400.0;
   max =     0.0;
   
   t_map_oref::const_iterator itEpo;
   t_gtime t;
   t_gtime t_sav = _mapobj[site].begin()->first;

   for( itEpo = _mapobj[site].begin(); itEpo != _mapobj[site].end(); ++itEpo ) // loop over epochs
   {
      if( itEpo == _mapobj[site].begin() ) continue;

      t = itEpo->first;
      if( abs(t-t_sav) > gap ) len[t_sav] = abs(t-t_sav);
      if( abs(t-t_sav) < min ) min = abs(t-t_sav);
      if( abs(t-t_sav) > max ) max = abs(t-t_sav);
	 
      t_sav = t;
   }
   
   _gmutex.unlock(); return 1;
}


// short block of data
// ----------------------------
int t_gallobs::block(string site, int b, int g, map<t_gtime, int>& len)
{
  gtrace("t_gallobs::block");
   
  if( _mapobj.find(site) == _mapobj.end() ) return 0;

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
   t_map_oref::const_iterator itEpo;
   t_gtime t;
   t_gtime t_sav = _mapobj[site].begin()->first;
   t_gtime sb = _mapobj[site].begin()->first;
   t_gtime eb;
   
   for (itEpo = _mapobj[site].begin(); itEpo != _mapobj[site].end(); itEpo++) // loop over epochs
   {
      t = itEpo->first;
      if (abs(t-t_sav) > g){
	 eb = t_sav;      
        if (abs(eb-sb) < b){
	   len[sb] = abs(eb-sb);
	   sb = t;
        }
	else sb = t;
      }
            
      t_sav = t;
   }
   
   _gmutex.unlock(); return 1;
}


// number of code/phase bands
// ----------------------------------------
//int t_gallobs::nbands(string site, map<GSYS, map<t_gtime, map<string, pair<int,int> > > >& nb, int sampl)
int t_gallobs::nbands(string site, t_map_bsys& nb, int sampl)
{
  gtrace("t_gallobs::nbands");

  if (_mapobj.find(site) == _mapobj.end()) return -1;

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
 
   t_map_oref::const_iterator itEpoB = _mapobj[site].begin();
   t_map_oref::const_iterator itEpoE;
   t_map_osat::iterator itSat;

//   t_map_oref::const_iterator itEpoB = _mapobj[site].begin();
//   t_map_oref::const_iterator itEpoE;
//   map<string,t_gobsgnss>::iterator itSat;   
   t_gtime t,tt;

   // all epochs are involved
   if (sampl == 0){
     while (itEpoB != _mapobj[site].end())
     {
	t = itEpoB->first;
        for(itSat = _mapobj[site][t].begin(); itSat != _mapobj[site][t].end(); itSat++) // loop over satellites
        {
	  GSYS sys = itSat->second->gsys();
          string prn = itSat->first;
          itSat->second->nbands(nb[sys][t][prn]);
	   
  	  /* cout << " bands ok = "  << prn */
	  /*      << " " << nb.size() */
	  /*      << " " << nb[sys].size() */
	  /*      << " " << nb[sys][t].size() */
	  /*      << " " << nb[sys][t][prn].first   */
	  /*      << " " << nb[sys][t][prn].second */
  	  /*      << " " << fixed << setprecision(6) << t.dsec() */
	  /*      << endl; */
 
        }	
	itEpoB++;
     }
     _gmutex.unlock(); return 1; 
   }
   
   // just sampled epochs are involved
   itEpoB = _mapobj[site].begin();
   itEpoE = _mapobj[site].end(); itEpoE--;
   t = itEpoB->first;
   vector<t_gsatdata> v_obs;
   while (t < itEpoE->first || t == itEpoE->first)
   {
     // handle clock drift!!
#ifdef BMUTEX                          // JE TREBA UPRAVIT!!!! MUTEX
     lock.unlock();
#endif      
     _gmutex.unlock();
     v_obs = obs(site, t);
     _gmutex.lock();
#ifdef BMUTEX
     lock.lock();
#endif

     if( v_obs.size() > 0 )
     {
//     if( v_obs_pt[0] == 0 ){ cout << "empty point in the vector \n"; t = t+sampl; }
       tt = v_obs[0].epoch();
       for(itSat = _mapobj[site][tt].begin(); itSat != _mapobj[site][tt].end(); itSat++) // loop over satellites
       {
	GSYS sys = itSat->second->gsys();
        string prn = itSat->first;
        itSat->second->nbands(nb[sys][t][prn]);
       }
     }
     t = t+sampl;       //  relative drift (tt+sampl) - not good solution, but should work also for absolute drift (t+sampl)
   }  
   
  _gmutex.unlock(); return 1;
}


// add elevations
//---------------
int t_gallobs::add_elevations(string site, t_gtriple xyz_rec, t_gallnav* gnav )
{
  gtrace("t_gallobs::add_elevations");
   
  if( !gnav || xyz_rec.zero() || _mapobj.find(site) == _mapobj.end() ) return -1;

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  t_map_osat::iterator itSAT;
  t_map_oref::iterator itTIM;
  t_map_oobj::iterator itOBJ;
   
  t_gtriple top_sat;
  t_gtriple xyz_sat;
  double r_radi, s_radi, s_topo, ele;

  for(itTIM = _mapobj[site].begin(); itTIM != _mapobj[site].end(); ++itTIM )
  {
    t_gtime tt = itTIM->first;
     
    for(itSAT = _mapobj[site][tt].begin(); itSAT != _mapobj[site][tt].end(); ++itSAT )
    {
      string prn = itSAT->first;

//      cout << prn << tt.str_ymdhms(" ") << " elevace: ";

      if( gnav->pos( prn, tt, xyz_sat.crd_array(), NULL, NULL ) < 0 || xyz_sat.zero() ){
//	cout << " n/a \n";
      }else{ 

        top_sat = xyz_sat - xyz_rec;

        r_radi = xyz_rec.crd_cvect().norm_Frobenius()/1000.0; // km
        s_radi = xyz_sat.crd_cvect().norm_Frobenius()/1000.0; // km
        s_topo = top_sat.crd_cvect().norm_Frobenius()/1000.0; // km
     
        ele = -90 + acos( ( pow(s_topo, 2.0) + pow(r_radi, 2.0) - pow(s_radi, 2.0) )
	  		   / ( 2 * r_radi * s_topo ) ) / PI * 180;

        _mapobj[site][tt][prn]->addele(ele);
//        cout << ele << endl;
      }
    }

  }

  _gmutex.unlock(); return 1;
}


// return true if map contains the site and false if not
//-----------------------------------------------------
bool t_gallobs::isSite(string site)
{
  gtrace("t_gallobs::isSite");

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
 
  bool tmp = false;
  if ( _mapobj.find(site) != _mapobj.end() ) tmp = true;

  _gmutex.unlock(); return tmp;
}


// get all t_gsatdata for prn from epoch interval
// ----------------------------------------------
vector<t_spt_gobs> t_gallobs::obs_prn_pt(const string& site, const string& prn, 
			  	 	 const t_gtime& beg, const t_gtime& end)
{
  gtrace("t_gallobs::obs_prn_pt");

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
 
  vector<t_spt_gobs> all_obs;
  t_gtime tt(t_gtime::GPS);

  if( _mapobj.find(site) == _mapobj.end() ){
#ifdef DEBUG     
      cout << "site not found: " << site << t.str_ymdhms() << endl; cout.flush();
#endif
    _gmutex.unlock(); return all_obs;
  }       

  t_map_oref::iterator it1 = _mapobj[site].lower_bound(beg);  // greater || equal
  t_map_oref::iterator it2 = _mapobj[site].upper_bound(end);  // greater || equal   

  for ( ;it1 != it2; ++it1)
  {
     tt = it1->first;
     if( _mapobj[site].find(tt) != _mapobj[site].end() )
     {
       t_map_osat::iterator itSAT = _mapobj[site][tt].find(prn);
       if( itSAT != _mapobj[site][tt].end() ) all_obs.push_back( itSAT->second );
     }
  }
  _gmutex.unlock(); return all_obs;
}


// find epoch from the map w.r.t. DIFF_SEC
// ---------------------------------------
int t_gallobs::_find_epo(const string& site, const t_gtime& epo, t_gtime& tt)
{
  gtrace("t_gallobs::_find_epo");

  if( _mapobj.find(site) == _mapobj.end() ){
      cout << "site not found: " << site << " " << epo.str_ymdhms() << endl; cout.flush();
      return -1;
  }
   
  t_map_oref::iterator it1 = _mapobj[site].lower_bound(epo);  // greater || equal
  t_map_oref::iterator it0 = it1;                             // previous value

  if( it0 != _mapobj[site].begin() ) it0--;                               // set previous value  
  if( it1 == _mapobj[site].end() ) it1 = it0;

  if( it1 == _mapobj[site].end() && it0 == _mapobj[site].end() ){
    cout << site << " " << epo.str_ymdhms() << " " << epo.dsec() << " no observations !\n";
    return -1;
  }
     
#ifdef DEBUG
    cout << site << " " << epo.str_ymdhms() << " " << epo.dsec() << fixed << setprecision(6)
         << " t1_diff: " << setw(14) << it1->first.diff(epo)
         << " t0_diff: " << setw(14) << it0->first.diff(epo)
         << endl;
#endif

  if(      fabs(it1->first - epo) <= DIFF_SEC ) tt = it1->first;  // it = it1;                 // set closest value
  else if( fabs(it0->first - epo) <= DIFF_SEC ) tt = it0->first;  // it = it0;                 // set closest value
  else{
#ifdef DEBUG
     if( _log ){
       ostringstream lg;
       lg << site << tt.str_ymd()  << tt.dsec()
          << " it0 = " << it0->first.str_ymdhms() << " diff: " << it1->first.diff(epo)
          << " it1 = " << it1->first.str_ymdhms() << " diff: " << it0->first.diff(epo)
          << " not close observations found!";
       _log->comment(0,"gallobs",lg.str());
     }     
#endif
     return -1;                                     // not found !
  }
   
   return 1;
}

} // namespace
