/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

*/

#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <vector>
#include <map>

#include "gall/gallsurf.h"
#include "gset/gsetgen.h"
#include "gset/gsetnwm.h"
#include "gset/gsetout.h"
#include "gutils/gtypeconv.h"
#include "gmodels/gtropo.h"
#include "gmodels/gfunclin.h"
#include "gmodels/gfitlmm.h"
#include "gproj/gprojlcc.h"

using namespace std;

namespace gnut {  

// constructor
// ----------
t_gallsurf::t_gallsurf() 
 : t_ginterp(),
   _ver("1.0"),
   _overwrite(false),
   _reg_auto(true),
   _min_lat(MIN_LAT),
   _max_lat(MAX_LAT),
   _min_lon(MIN_LON),
   _max_lon(MAX_LON),
   _inc_lat(0.0),
   _inc_lon(0.0),
   _max_ver(0.0),
//   _proj(0), // NOT COMPATIBLE WITH OLD DEBIAN
   _set(0),
   _out(0),
   _sum(0),
   _grd(0),
   _srf(0),
   _fit(0)
{
  id_type(t_gdata::ALLSURF);
  id_group(t_gdata::GRP_PRODUCT);

  string str_DATA[] = { "TEMP", "PRES", "ZWD", "ZHD",  };  // default meteorological parameters
  set<string> param( str_DATA, str_DATA + 4);

  _set_param( param );
}



// constructor
// ----------
t_gallsurf::t_gallsurf(t_gsetbase* gset) // use only if NWM settings available !
 : t_ginterp(),
   _ver("1.0"),
   _overwrite(false),
   _reg_auto(true),
   _inc_lat(0.0),
   _inc_lon(0.0),
//   _proj(0), // NOT COMPATIBLE WITH OLD DEBIAN
   _set(gset),
   _out(0),
   _sum(0),
   _grd(0),
   _srf(0),
   _fit(0)
{
  id_type(t_gdata::ALLSURF);
  id_group(t_gdata::GRP_PRODUCT);

  double minLat = dynamic_cast<t_gsetnwm*>(gset)->min_lat();
  double maxLat = dynamic_cast<t_gsetnwm*>(gset)->max_lat();
  double minLon = dynamic_cast<t_gsetnwm*>(gset)->min_lon();
  double maxLon = dynamic_cast<t_gsetnwm*>(gset)->max_lon();

  set<string> param = dynamic_cast<t_gsetnwm*>(_set)->param();
   
  _set_param( param );
  _set_output();
  _set_region( minLat, maxLat, minLon, maxLon );
  _interp_1d = str_to_interp_1d( dynamic_cast<t_gsetnwm*>(_set)->interp_time()  );
  _interp_3d = str_to_interp_3d( dynamic_cast<t_gsetnwm*>(_set)->interp_space() );
  _interp_ht = str_to_interp_ht( dynamic_cast<t_gsetnwm*>(_set)->interp_vert() );
}


// constructor
// ----------
t_gallsurf::t_gallsurf(double minLat, double maxLat, double minLon, double maxLon)
 : t_ginterp(),
   _ver("1.0"),
   _overwrite(false),
   _reg_auto(true),
   _inc_lat(0.0),
   _inc_lon(0.0),
//   _proj(0), // NOT COMPATIBLE WITH OLD DEBIAN
   _set(0),
   _out(0),
   _sum(0),
   _grd(0),
   _srf(0),
   _fit(0)
{
  _set_region( minLat, maxLat, minLon, maxLon );
  if( _min_lat == 0.0 && _max_lat == 0.0 && _min_lon == 0.0 && _max_lon == 0.0 ) _reg_auto = false;
   
  string str_DATA[] = { "PRES", "TEMP", "ZWD", "ZHD" }; // default meteorological parameters
  set<string> param( str_DATA, str_DATA + 4);
  _set_param( param );
}


// destructor
// ----------
t_gallsurf::~t_gallsurf()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  t_map_met::iterator itEPO;
  for( itEPO = _map_met.begin(); itEPO != _map_met.end(); ++itEPO ){
     
    t_map_dat::iterator itGRID;
    for( itGRID = itEPO->second.begin(); itGRID != itEPO->second.end(); ++itGRID ){

#ifdef DEBUG
      cerr << " deleting profile " << fixed << setprecision(3)
           << setw(10) << itGRID->first.crd(0)
           << setw(10) << itGRID->first.crd(1)
	   << itEPO->first.str_ymdhms(" t_gallsurf: ")
	   <<  " " << (itGRID->second)->get_surf(LAT)
	   <<  " " << (itGRID->second)->get_surf(LON)
           << endl;
#endif

      delete itGRID->second;
    }
    itEPO->second.clear();
  }
  _map_met.clear(); // NOT NECESSARY FOR VALGRIND
   
  if( _grd ){ if( _grd->is_open() ){ _grd->close(); }; delete _grd; }
  if( _srf ){ if( _srf->is_open() ){ _srf->close(); }; delete _srf; }
  if( _fit ){ if( _fit->is_open() ){ _fit->close(); }; delete _fit; }
  if( _out ){ if( _out->is_open() ){ _out->close(); }; delete _out; }
  if( _sum ){ if( _sum->is_open() ){ _sum->close(); }; delete _sum; }
   
  _gmutex.unlock(); return;
}

   
// set projection (only if not set before)
// ----------
int t_gallsurf::gproj(const shared_ptr<t_gproj>& prj)
{  

  if( _proj ){ //  != make_shared<t_gproj>() ){
    if( *_proj == *prj ){ 
      if( _log ) _log->comment(0,"gallsurf","Specific projection already identified in NWM container" );
      return 1;
    }else{
    if( _log ) _log->comment(0,"gallsurf","Different projections identified in NWM container" );
    else               cerr << "gallsurf - Different projections identified in NWM container\n";
    return -1;
    }
  }

  if( _log ) _log->comment(0,"gallsurf","Set and use of specific projection in NWM container" );

  _proj = prj;

  return 1;
}


// add gnwm profile
// ----------
int t_gallsurf::add(t_gnwmsurf* met, bool proj)
{
// gtrace("t_gallsurf::add");
 
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
   
  _gmutex.lock();

  t_gpair LatLon( met->get_surf(LAT), met->get_surf(LON) );

  if( proj && _proj ){ //   != make_shared<t_gproj>() ){
    t_gpair LatLonXY( met->get_surf(YPRJ), met->get_surf(XPRJ) );  // TO CORRESPOND Y/X eq. LAT/LON
    LatLon = LatLonXY;
#ifdef DEBUG     
    cout << "adding profile using PROJECTION: " << LatLon[0] << " " << LatLon[1]
         << " " << met->get_surf(YPRJ) <<  " " << met->get_surf(XPRJ) 
	 << endl;
#endif
  }
   
  if( LatLon[0] == NWM_UNKNOWN || LatLon[1] == NWM_UNKNOWN ){
    cerr << "*** warn: met - not defined lat/lon of the profile!" << fixed
         << setw(10) << LatLon[0]
         << setw(10) << LatLon[1]
         << endl;
    _gmutex.unlock(); return -1;
  }

  if( LatLon[0] < _min_lat || LatLon[0] > _max_lat ||
      LatLon[1] < _min_lon || LatLon[1] > _max_lon    )
  {
    cout << "out of LatLon limits " << _min_lon << " " << _max_lon << " \n";
    _gmutex.unlock(); return 1;
  }

  if( _set )
  {	  
    set<string> surface = dynamic_cast<t_gsetnwm*>(_set)->surface();
    if( surface.find("surface") != surface.end() ) met->print_surface();
  }
   
  // profile does not exist or _overwrite (--> delete & new)
  t_gtime epo( met->epoch() );   
  if( _map_met[epo].find( LatLon ) == _map_met[epo].end() ){
     
    this->_new_element(epo,LatLon,met); // allocate in heap;
 // if( _log && _log->verb() >= 1 ) _log->comment(1,"gallsurf","ok - new profile: "+dbl2str(LatLon[0])+" "+dbl2str(LatLon[1]) );

  }else{

    if( _map_met.lower_bound(epo+60) == _map_met.end() ){ // LAST EPOCH ?
          epo.add_sec(+1); } // plus  1 sec (asume  forwards list of NWM files)
    else{ epo.add_sec(-1); } // minus 1 sec (asume backwards list of NWM files)

    if( _log && _log->verb() >= 1 ) _log->comment(1,"gallsurf","warning - cannot overwrite, epoch shifted: "+epo.str_ymdhms() );
    else                                    cerr << "gallsurf:  warning - cannot overwrite, epoch shifted: "+epo.str_ymdhms() << endl;     

    this->_new_element(epo,LatLon,met); // allocate in heap;
//  _gmutex.unlock(); return -1;
  }

#ifdef DEBUG
  cout << " adding profile " << fixed << setprecision(3)
       << setw(10) << met->get_surf(LAT)
       << setw(10) << met->get_surf(LON)
       << endl;
#endif   
  _gmutex.unlock(); return 0;
}
   

// get all epochs
// ----------
set<t_gtime> t_gallsurf::epochs()
{
  gtrace("t_gallsurf::epochs");
   
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  set<t_gtime> s_epochs;
   
  t_map_met::const_iterator itEPO = _map_met.begin();
  while( itEPO != _map_met.end() ){
    s_epochs.insert( itEPO->first );
    ++itEPO;
  }

  _gmutex.unlock();
  return s_epochs;
}


// get all pairs
// ----------
set<t_gpair> t_gallsurf::pairs( const t_gtime& epo )
{
  gtrace("t_gallsurf::pairs");
   
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  set<t_gpair> s_pairs;

  // not correct time
  if( _map_met.find(epo) == _map_met.end() ){
    _gmutex.unlock(); return s_pairs;
  }
     
  t_map_dat::const_iterator itGRID = _map_met[epo].begin();
  while( itGRID != _map_met[epo].end() ){
    s_pairs.insert( itGRID->first );
    ++itGRID;
  }

  _gmutex.unlock();
  return s_pairs;
}


// -----------------------------
// dLat, dLon grid step              ---->>>> POZOR, NEFUNGUJE SPRAVNE PRO VYBER STANIC!!!!!
// -----------------------------
int t_gallsurf::_get_grid_step()
{
   gtrace("t_gallsurf::get_grid_step");

   if( _inc_lat != 0.0 ) return 0;
   if( _inc_lon != 0.0 ) return 0;
     
   double Lon = NWM_UNKNOWN;
   double Lat = NWM_UNKNOWN;
   double StepLat = 9999;
   double StepLon = 9999;

   for(t_map_met::iterator iEP = _map_met.begin(); iEP != _map_met.end(); ++iEP)
   {
     t_map_dat dat = iEP->second;
     for(t_map_dat::iterator itD = dat.begin(); itD != dat.end(); ++itD)
     {
       if( Lat == NWM_UNKNOWN ) Lat = itD->first.crd(0);	
       if( Lon == NWM_UNKNOWN ) Lon = itD->first.crd(1);

       double dLat = fabs(itD->first.crd(0) - Lat);
       double dLon = fabs(itD->first.crd(1) - Lon);

       if( dLat > 0.0 && dLat < StepLat ) StepLat = fabs(itD->first.crd(0)-Lat);
       if( dLon > 0.0 && dLon < StepLon ) StepLon = fabs(itD->first.crd(1)-Lon);
     }
   }
   
#ifdef DEBUG
   cout << " dLat dLon  ... " <<   StepLat   << " : " <<   StepLon   << endl;
#endif
   _inc_lat = StepLat;
   _inc_lon = StepLon;
   return 0;
}


// -------------------------------------
// Get epochs for interpolation (single epoch)
// -------------------------------------
int t_gallsurf::_get_node_epochs(const t_gtime& epo, set<t_gtime>& set_dt)
{
   gtrace("t_gallsurf::get_node_epochs (single epoch)");
   
   t_map_met::iterator itEP, itEPforw, itEPback;
   itEP = _map_met.lower_bound(epo);
     
   // exact time (1 value)
   if( itEP->first == epo ){
      set_dt.insert(epo);
//      cerr << "EPOCHS -> " << epo.str_ymdhms() <<  " " << itEP->first.str_ymdhms() << endl;
//      if( _log ) _log->comment(0,"gallsurf","Warning - single epoch found: " + epo.str_ymdhms() + " " + int2str(set_dt.size()) );
//      else               cerr << "gallsurf:  Warning - single epoch found: " <<epo.str_ymdhms() <<" " <<set_dt.size() << endl;      
      return 0; 
   }
 
   itEPforw = itEPback = itEP;
   if( itEP == _map_met.begin() || itEP == _map_met.end() ){
     if( _log ) _log->comment(0,"gnwmsurf","no data for temporal interpolation");
     return -1;
   }else{
    // linear (2 values)
    --itEPback;
    set_dt.insert(itEPforw->first);
    set_dt.insert(itEPback->first);
    ++itEPforw;
   }
   
   if( itEPback == _map_met.begin() || itEPforw == _map_met.end() || _interp_1d != SPLINE ){
    --itEPforw;
   }else{
     // spline (4 values) 
     --itEPback;
     set_dt.insert(itEPforw->first);
     set_dt.insert(itEPback->first);
   }
#ifdef DEBUG
   cout << " node epochs found = "
        << " " << interp_1d_to_str( _interp_1d )
        << " " << interp_2d_to_str( _interp_2d )
        << " " << interp_3d_to_str( _interp_3d )
        << " " << interp_ht_to_str( _interp_ht )
        << " " << epo.str_ymdhms()
        << itEPback->first.str_ymdhms(" beg: ")
        << itEPforw->first.str_ymdhms(" end: ")
        << " " << set_dt.size() << endl;
#endif
   
#ifdef DEBUG
   cout << " epochs : ";
   for(itEP = _map_met.begin(); itEP != _map_met.end(); ++itEP ) cout << itEP->first.str_ymdhms(" ");
   cout << endl;
#endif

  return 0;
}


// -------------------------------------
// Get epochs for interpolation (interval)
// ------------------------------------- 
int t_gallsurf::_get_node_epochs(const t_gtime& beg, const t_gtime& end, set<t_gtime>& set_dt)
{
  gtrace("t_gallsurf::get_node_epochs (interval)");

#ifdef DEBUG
  cerr << " map_met epoch:"
       << beg.str_ymdhms(" ")
       << end.str_ymdhms(" ") << endl;
#endif

  // interface to a single epoch
  if( beg == end ) return _get_node_epochs(beg, set_dt);
     
  t_map_met::iterator itEPO;
  for( itEPO = _map_met.begin() ; itEPO != _map_met.end(); ++itEPO )
  {
//  cerr << " map_met epoch : " << itEPO->first.str_ymdhms() << endl;
    if( beg       == itEPO->first  ||  itEPO->first == end      ||
      (beg-6*3600  < itEPO->first  &&  itEPO->first  < end + 6*3600) ) // 6h reserve
    {
      set_dt.insert( itEPO->first );
       
#ifdef DEBUG
      cout << " node epochs found = " 
           << " " << interp_1d_to_str( _interp_1d )
           << " " << interp_2d_to_str( _interp_2d )
           << " " << interp_3d_to_str( _interp_3d )
           << " " << interp_ht_to_str( _interp_ht )
	   << itEPO->first.str_ymdhms(" epo: ")
           << beg.str_ymdhms(" beg: ")
           << end.str_ymdhms(" end: ")
           << " " << set_dt.size() << endl;
#endif
    }
  }   

  if( set_dt.size() == 0 ) return 1;

  if( _map_met.size() < 2 && beg != end ){
    if( _log ) _log->comment(0,"gallsurf","Warning - single epoch found. No temporal interpolation possible!" );
    else               cerr << "gallsurf:  Warning - single epoch found. No temporal interpolation possible!\n";
    return -1;
  }

  return 0;
}

// ---------------------------------------------
// get data points
// ---------------------------------------------
int t_gallsurf::_get_grid_points( const t_gpair& pt,
				  const set<t_gtime>& set_dt,
			                set<t_gpair>& set_pt )
{
  gtrace("t_gallsurf::get_grid_points");

  if( set_dt.size() <= 0   ||
    ( _inc_lat      <= 0.0 ||
      _inc_lon      <= 0.0 )){ set_pt.clear(); return -1; }

  // without cache
  t_gtime tt( *(set_dt.begin()) );
  unsigned int nPT = 0;

  t_map_met::iterator itMET = _map_met.find(tt);

  t_map_dat data = itMET->second;
//t_map_dat::reverse_iterator itDAT( data.upper_bound(pt) ); // ! not correct approach!!!
  t_map_dat::reverse_iterator itDAT = data.rbegin(); // combine forw/reverse iterator
   
  for( ; itDAT != data.rend(); ++itDAT)
  {     
    t_gpair node = itDAT->first;

#ifdef DEBUG     
    cout << " test = " <<   pt.crd(0)  << " " <<   pt.crd(1)
         <<        " " << node.crd(0)  << " " << node.crd(1) << endl;
#endif
   
    // 1 point
    if( pt == node )
    {     
//    _grid_points.insert(node);     // cache
      set_pt.insert(node);
      nPT = 1;  break;       
    }

    bool okLat(false), okLon(false);     
    if( fabs( pt[0] - node[0]    ) < _inc_lat   ) okLat = true;
    if( fabs( pt[1] - node[1]    ) < _inc_lon ||
        fabs( pt[1] - node[1]-360) < _inc_lon ||
        fabs( pt[1] - node[1]+360) < _inc_lon   ) okLon = true;

    // 2 points
    if( ( pt[0] == node[0] && okLon ) ||
	( pt[1] == node[1] && okLat ) )
    {
//    _grid_points.insert(node);     // cache
      set_pt.insert(node);
      nPT = 2;
      if( set_pt.size() == 2 )  break;       
      continue; 
    }

    if( okLat && okLon )
    {	   
//    _grid_points.insert(node);     // cache
      set_pt.insert(node);
      nPT = 4;
      if( set_pt.size() == 4 )  break;
      continue; 
    }
  }
   
  if( nPT != set_pt.size() ){ 
    set_pt.clear(); 
    return -1;
  }

#ifdef DEBUG     
  if( _log && _log->verb() >= 1 ){
    for( set<t_gpair>::iterator it = set_pt.begin(); it != set_pt.end(); ++it )
    {	  
      ostringstream os;
      os << fixed << setprecision(3)
         << " user: " << setw(8)  <<  pt.crd(0) << " " << setw(8) <<  pt.crd(1)
         << " grid: " << setw(8)  << it->crd(0) << " " << setw(8) << it->crd(1)
         << " " << set_pt.size();
      _log->comment(1,"gallsurf",os.str());
    }
  }
#endif
   
  if( _log && _log->verb() >= 1 ){
    ostringstream os;
    os <<  "inc_lat: " << _inc_lat 
       << " inc_lon: " << _inc_lon
       << " lat-lon: " << pt[0] << " " << pt[1]
       << " pt_size: " << set_pt.size()
       << " ep_size: " << set_dt.size()
       << " tot_epo: " << _map_met.size();
   _log->comment(1,"gallsurf", os.str());
  }

  return 0;
}


// ---------------------------------------------
// Create maps of GRID/REQ.PT for all NWM epochs
// ---------------------------------------------
int t_gallsurf::_get_grid_interp(const t_gtriple& pt,
			         const set<t_gtime>& set_epo,
			         const set<t_gpair>& set_pt,
				 const vector<MET_SURF> vec_surf,
			         t_map_met& map_grid,          // map of NWM epochs/pairs/profiles
			         t_map_met& map_prof)          // map of NWM epochs/reqpt/profiles req.point)
{
  gtrace("t_gallsurf::_get_grid_interp");
  // loop over NWM epochs (not requested epochs)
  set<t_gtime>::iterator itEPO;
  for( itEPO = set_epo.begin(); itEPO != set_epo.end(); ++itEPO )
  {  
     
#ifdef DEBUG     
    cout << "interpolate: NWM Epoch = " << itEPO->str_ymdhms()  << endl;
#endif     
     
    t_gtime tt( *itEPO );
    t_map_dat::iterator itDAT;
    t_map_met::iterator itMET = _map_met.find( tt );
    if( itMET == _map_met.end() ){ cout << "# --> warning: should not occure [interpolate]!\n"; continue; }
    int irc = 0;

    // GRID POINTS (all epochs): create map of gnwmsurf pointers (epochs/grid_pt)
    set<t_gpair>::iterator itPT;
    for( itPT = set_pt.begin(); itPT != set_pt.end(); ++itPT )
    {
      if( (itDAT = itMET->second.find( *itPT )) != itMET->second.end() ){
             map_grid[tt][itDAT->first] = itDAT->second;
      }else{ cerr << "error: missing point for bilinear interpolation !\n";
   	     map_grid.erase(*itEPO);
	     irc++; continue;
      }
    }

    if( irc > 0 ) continue; // skip problematic

    // REQ.POINT (all epochs): create map of NEW gnwmsurf pointers (epochs/req_pt)
    map_prof[tt][pt.gpair()] = new t_gnwmsurf();
    map_prof[tt][pt.gpair()]->add_surf(LAT, pt[0]);
    map_prof[tt][pt.gpair()]->add_surf(LON, pt[1]);
    map_prof[tt][pt.gpair()]->add_surf(HEL, pt[2]);
    map_prof[tt][pt.gpair()]->calc_surface();

    // interpolate basic parameters (surface)
    vector<MET_SURF>::const_iterator itSURF;
    for( itSURF = vec_surf.begin(); itSURF != vec_surf.end(); ++itSURF )
    {
      // PREPARE DATA for BILINEAR: interpolate basic surface parameters
      map<t_gpair, double> map_space;
      map<t_gtime, double> map_gtime;
      for( itPT = set_pt.begin(); itPT != set_pt.end(); ++itPT ){
        map_space[ *itPT ] = itMET->second[ *itPT ]->get_surf( *itSURF );
#ifdef DEBUG
        cout << "interp_aux1: "  << tt.str_ymdhms() << fixed << setprecision(5)
	     << setw(8) << t_gnwmbase::met_surf_str( *itSURF ) << " " << (*itPT)[0] << " " << (*itPT)[1]
	     << "  input: " << map_space[ *itPT ] << endl;
#endif
      }
      double res = NWM_UNKNOWN;
      if( bilinear( map_space, pt.gpair(), res ) == 0 ){ 
	map_prof[tt][pt.gpair()]->add_surf(*itSURF, res);
	 
	// save also basic parameters
//        if( m_surf ) (*m_surf)[tt][*itSURF] = res; // OBSOLETE !!!

#ifdef DEBUG	 
        cout << "interp_aux2: " << tt.str_ymdhms() << fixed << setprecision(5)
	     << setw(8) << map_prof[tt][pt.gpair()]->met_surf_str( *itSURF ) << " " << pt[0] << " " << pt[1]
	     << " output: " << res 
	     << endl;
#endif
      }
      map_space.clear();
    }
  }
  return 0;
}

// temporal interpolation (linear/spline)
// -----------------
int t_gallsurf::_interp_temporal(const t_gtime& beg, 
			 	 const t_gtime& end,
	 	                 const double& sampling,
	  		 	 const map<t_gtime,double>& inp,
	 	                       map<t_gtime,double>& out)
{ 
  map<t_gtime, double> map_gtime;   // local map
  map<t_gtime, double>::const_iterator itEPO,itE;

  t_gtime epo(beg);
  t_gtime epo_BEF0 = t_gtime::first_time();
  t_gtime epo_BEF  = t_gtime::first_time();
  t_gtime epo_AFT  = t_gtime::first_time();
  t_gtime epo_AFT1 = t_gtime::first_time();

  // loop over NWM epochs (shifting epochs map by 1)
  for( itEPO = inp.begin(); itEPO != inp.end(); ++itEPO )
  {
    t_gtime tt = itEPO->first; //  cerr << "tt = " << tt.str_ymdhms() << endl;
     
    // add current epoch (if not available)
    if( map_gtime.find(tt) == map_gtime.end() ) map_gtime[tt] = itEPO->second;
     
    // SINGLE EPOCH INTERPOLATION
    if( tt == beg && tt == end ){ out[tt] = inp.at(tt); return 0;}

    // skip the first epoch
    if( epo_BEF == t_gtime::first_time() ){ epo_BEF = tt; continue; }
     
    epo_AFT = tt;

    // support SPLINE if data available at epoBEF0 & epo_AFT1
    if( _interp_1d == SPLINE && map_gtime.size() > 2 ){
      itE = itEPO;

      // more data after available
      if( ++itE != inp.end() ){
         epo_AFT1 = itE->first;
	 double tmp = inp.at(epo_AFT1);
         map_gtime[epo_AFT1] = tmp;
//       cout << epo_AFT1.str_ymdhms("adding epoch AFT1: ") << endl;
      }
      // continue with interpolation
      else if( map_gtime.find(epo_BEF0) != map_gtime.end() ){ 
        map_gtime.erase(epo_BEF0); 
//      cout << epo_BEF0.str_ymdhms("removing epoch BEF0: ") << endl;
      }
    }

    // SHOULD NOT HAPPEN!
    if( map_gtime.size()%2 != 0 )
      cerr << tt.str_ymdhms("Warn - not correct map size: ") << " " << map_gtime.size() << " !\n";
     
#ifdef DEBUG
    cout << fixed << setprecision(4) <<t t.str_ymdhms("#EPO: ") << " " << map_gtime.size() << ": ";
    for( map<t_gtime, double>::iterator it = map_gtime.begin(); it != map_gtime.end(); ++it)
      cout << " ... " << it->first.str_ymdhms();
    cout << endl;
#endif

    // interpolate btw 2 NWP epochs
    while( epo == epo_BEF || epo == epo_AFT || ( epo > epo_BEF && epo < epo_AFT ) )
    {
      double res = 0.0;
      if( map_gtime.size() == 2 && linear( map_gtime, epo, res ) == 0 ){ out[epo] = res; }
      if( map_gtime.size() == 4 && spline( map_gtime, epo, res ) == 0 ){ out[epo] = res; }
#ifdef DEBUG
      cout << epo.str_ymdhms("#EPO: ") << setw(9) << xxx[0] << setw(9) << xxx[1]
           << setw(12) << res << endl;
#endif	
      epo = epo + sampling;
    }

    // shift by single NWM (source) epoch
    if( map_gtime.size() == 4 && _interp_1d == SPLINE ){ map_gtime.erase(epo_BEF0); }
    if( map_gtime.size() == 2 && _interp_1d == LINEAR ){ map_gtime.erase(epo_BEF ); }

    epo_BEF0 = epo_BEF;
    epo_BEF  = tt;

  } // loop over NWM epochs
  return 0;
}


// point interpolation - spatial + temporal
// -----------------
int t_gallsurf::interpolate_point( const t_gtriple& ell,
 	                           const t_gtime& beg,
			           const t_gtime& end,
			           map<t_gtime, t_map_metdata> *m_data,
				   map<t_gtime, t_map_metsurf> *m_surf,
				   map<t_gtime, t_gnwmsurf>    *m_prof,
			           double sampling )
{ 
  gtrace("t_gallsurf::interpolate_point");

  t_gtriple xxx(ell);
  if( _proj ){ xxx = _proj->ll2proj( ell ); } // USE THE PROJECTION!   
  t_gpair req_pt = xxx.gpair();

  if( _log && _log->verb() >= 2 ) _log->comment(2,"gallsurf","Req_pt coord: "+dbl2str(ell[0])+" "+dbl2str(xxx[1])+" "+dbl2str(xxx[2]));

  _max_ver = 0.0;

  if( m_data == 0 && m_surf == 0 && m_prof == 0 )  return 1;
  if( sampling <= 0 ) sampling = DEFAULT_SAMPLING;

  set<MET_SURF>    set_surf;
  vector<MET_SURF> vec_surf;  vector<MET_SURF>::iterator itSURF;
  vector<MET_DATA> vec_par;   vector<MET_DATA>::iterator itPAR;
   
  // user-requested parameters (from _vec_surf)
  for( itSURF = _vec_srf.begin(); itSURF != _vec_srf.end(); ++itSURF ){
    vec_surf.push_back(*itSURF);
    set_surf.insert(*itSURF);    // just to  list
  }

  // mandatory background surface parameters (for vertical reduction)
  if( set_surf.find(MP_Tp) == set_surf.end() ) vec_surf.push_back(MP_Tp);
  if( set_surf.find(MP_Mp) == set_surf.end() ) vec_surf.push_back(MP_Mp);
  if( set_surf.find(MP_Ep) == set_surf.end() ) vec_surf.push_back(MP_Ep);
  if( set_surf.find(MP_Wp) == set_surf.end() ) vec_surf.push_back(MP_Wp);
  if( set_surf.find(MP_Rt) == set_surf.end() ) vec_surf.push_back(MP_Rt);
  if( _interp_3d           == HOR2VER        ) vec_surf.push_back(HEL);   // interp. HEL of req.pt: 1.space 2.vert 3.time

  set<t_gpair> set_pt;         set<t_gpair>::iterator itPT;       // source grid pointers
  set<t_gtime> set_epo;        set<t_gtime>::iterator itEPO;      // source epochs

  t_map_met map_prof;          t_map_met::iterator itPROF;        // map (profiles) for interpolated point (fixed)
  t_map_met map_grid;          t_map_met::iterator itMET;         // map (profiles) for NWM grid points (BILINEAR:4)
                               t_map_dat::iterator itDAT;

  map<t_gpair, double> map_space;  
  map<t_gtime, double> map_gtime;
  map<t_gtime, double> map_results;
  map<t_gtime, double>::const_iterator itE;

  int irc = 0;
//irc += _get_node_epochs(epo, set_epo);                          // get epoch nodes (1,2,4)
  irc += _get_node_epochs(beg, end, set_epo);                     // get epoch nodes (1,2,4)
  irc += _get_grid_step();                                        // get grid step (if not set)  --> NEFUNGUJE SPRAVNE PRO VYBER STANIC!!
  irc += _get_grid_points(req_pt, set_epo, set_pt);               // get nearby grid points (1,2,4)

#ifdef DEBUG    // INTERPOLATING EPOCHS
  for( itEPO = set_epo.begin(); itEPO != set_epo.end(); ++itEPO )
    cerr << (*itEPO).str_ymdhms("interp_epochs: ") 
         << beg.str_ymdhms(" beg:") << end.str_ymdhms(" end:") << endl;
#endif

  if( irc != 0 ){ cout << "# --> warning: initialization problem [interpolate]!\n"; return -1; }

  // fill maps of epochs/data for GRID(4 x nEpo) & REQ_PT(1 x nEpo)
  irc += _get_grid_interp(xxx, set_epo, set_pt, vec_surf, map_grid, map_prof);

  // temporal for basic parameters (only if requested --> if( m_surf || m_prof )
  if( m_surf || m_prof ){
    map_gtime.clear();
    for( itSURF = vec_surf.begin(); itSURF != vec_surf.end(); ++itSURF )
    {
      for( itEPO = set_epo.begin(); itEPO != set_epo.end(); ++itEPO ){
        map_gtime[*itEPO] = map_prof[*itEPO][xxx.gpair()]->get_surf(*itSURF);
//      cerr << " map_gtime:  " << map_gtime.size()  << " " << itEPO->str_ymdhms() << endl;
      }

      _interp_temporal( beg, end, sampling, map_gtime, map_results );
     
      for( itE = map_results.begin(); itE != map_results.end(); ++itE )
      { t_gtime tt = itE->first;
        if( ! (tt<beg || tt>end) ) (*m_surf)[tt][*itSURF] = map_results[tt]; // cut requested period
      }

      map_gtime.clear();
    }
  }

  // loop over PARAMETERS (LEVEL data)
  // =================================
  for( itPAR = _vec_par.begin(); itPAR != _vec_par.end(); ++itPAR )
  {
#ifdef DEBUG
    cout << "interpolate param: " << t_gnwmbase::met_data_str(*itPAR,false) << endl;
#endif
    map_gtime.clear();

    if( *itPAR == ZTD ){ continue; }
      
    // loop over NWM original epochs
    for( itMET = map_grid.begin(); itMET != map_grid.end(); ++itMET )
    {
      t_gtime tt(itMET->first);
	     
      // loop over grid points
      for( itDAT  = itMET->second.begin();
	   itDAT != itMET->second.end() && *itPAR != ZTD; 
	 ++itDAT )
      {
	t_gnwmsurf* gsurf = itDAT->second;     // profile data
	double val = gsurf->get_surf(*itPAR);  // individual data
	double hel = xxx[2];                   // reference height
	
	if( fabs(hel - gsurf->get_surf(HEL)) > _max_ver ) _max_ver = fabs(hel - gsurf->get_surf(HEL));
		  
	if( _interp_3d == VER2HOR ){                         // 1.vert, 2.horiz, 3.time
	  if( _interp_ht == REDUCE ){                        // approximate with vertical reduction
	    val = gsurf->get_in_hel( *itPAR, xxx[2] );       // cout << "#[VER2HOR] VERT REDUCE\n";
	  }else if( _interp_ht == INTEGRATE ){               // directly searched for intergration
	    // cout << "#[VER2HOR] VERT INTEGR - not implemented!\n";
	    val = gsurf->get_integr( *itPAR, xxx[2] );
	  }
        }else if( _interp_3d == HOR2VER ){                   // 1.hoziz, 2.vert, 3.time
	    hel = map_prof[tt][req_pt]->get_surf(HEL);
	    val = gsurf->get_in_hel( *itPAR, hel );          // cout << "#[HOR2VER] VERT REDUCE\n";
        }
		  
	// add corrected/uncorrected VALUE for spatial interpolation
	map_space[itDAT->first] = val;
		  
#ifdef DEBUG
        cout << fixed    << setprecision(4) 
	     << "#GRD: " << tt.str_ymdhms()
             << " "      << itDAT->first.crd(0)
             << " "      << itDAT->first.crd(1)
             << " "      << itDAT->second->met_data_id( *itPAR )
             << " "      << itDAT->second->get_surf( *itPAR );
         if(!double_eq( val, itDAT->second->get_surf( *itPAR )) ){
	    cout << " -> " << val
	         << "    " << gsurf->get_surf( HEL )
	         << " -> " << hel;
	 }
	 cout << endl;
#endif
      }      
      // spatial interpolation for (un)corrected values
      double res = 0.0;
      if( bilinear( map_space, req_pt, res ) == 0 ){
	map_prof[tt][req_pt]->add_surf(*itPAR, res);

        if( _interp_3d == HOR2VER ){                                // 1.horiz, 2.vert, 3.time
          res = map_prof[tt][req_pt]->get_in_hel( *itPAR, xxx[2] ); // cout << "#[HOR2VER] VERT REDUCE \n";
        }	 
        map_gtime[tt] = res;

#ifdef DEBUG
        cout << fixed    << setprecision(4)
             << "#INT: " << tt.str_ymdhms()
	     << " "      << xxx[0]
	     << " "      << xxx[1]
             << " "      << map_prof[tt][req_pt]->met_data_id( *itPAR )
	     << " "      << map_prof[tt][req_pt]->get_surf(    *itPAR );
        if(!double_eq( res, map_prof[tt][req_pt]->get_surf(    *itPAR )) ){
  	  cout << " -> " << res
	       << "    " << map_prof[tt][req_pt]->get_surf( HEL )
	       << " -> " << xxx[2];
        }
        cout << " "      << map_gtime.size() << endl;
#endif

      }else{
        if(_log) _log->comment(0,"gallsurf", "Warning - bilinear interpolation failed");
      }
      map_space.clear();
    }

    // temporal interpolation
    _interp_temporal( beg, end, sampling, map_gtime, map_results );
     
    for( itE = map_results.begin(); itE != map_results.end(); ++itE )
    { 
      t_gtime tt = itE->first;
      if( ! (tt<beg || tt>end) )
	(*m_data)[tt][*itPAR] = map_results[tt]; // cut requested period

    }     

    map_gtime.clear();

  } // loop over PARAMETERS

  // HANDLE ZTD (cummulative parameter)
  for( itPAR = _vec_par.begin(); itPAR != _vec_par.end(); ++itPAR ){
    if( *itPAR == ZTD ){
      for( itE = map_results.begin(); itE != map_results.end(); ++itE ){
        t_gtime tt = itE->first;
        if( ! (tt<beg || tt>end) )
	   (*m_data)[tt][*itPAR] = (*m_data)[tt][ZHD] + (*m_data)[tt][ZWD]; // cummulate + cut ZTD
      }
    }
  }

  // clean local data
  for( itPROF = map_prof.begin(); itPROF != map_prof.end(); ++itPROF )
  {
    t_gtime tt = itE->first;
    t_gnwmsurf* tmpsurf = itPROF->second[req_pt];
      
    if( m_prof && ! (tt<beg || tt>end) ) (*m_prof)[tt] = *tmpsurf; // cut requested period

    delete tmpsurf;
  }
  map_prof.clear();

  return 0;
}


// ---------------------------------------------
// estimate gradients
// ---------------------------------------------
int t_gallsurf::analyse_gradient( const t_gtriple& ell,                         // ELL:BLH of point of interest
				  const double& cut_off,                        // [deg] elevation cut-off
				  map<t_gtime, map<MET_DATA, double> >& rf_data)// ref.point interpolated data
{   
  gtrace("t_gallsurf::analyse_gradient");

  map<t_gtime, map<MET_DATA, double> >::const_iterator itEPO;

   
#ifdef DEBUG  // WRITE INPUT DATA (ZHD, ZWD, ZTD)
  if( _proj ) cout << "USING THE LCC PROJECTION ! <---- \n";
  cout << "REQ POINT: " << fixed << setprecision(3) << ell[0] << " " << ell[1] << " " << ell[2] << endl;
  for( itEPO = rf_data.begin(); itEPO != rf_data.end(); ++itEPO)
  {
    cout << itEPO->first.str_ymdhms() << " ";
     
    map<MET_DATA, double>::const_iterator itDAT;
    for(itDAT = itEPO->second.begin(); itDAT != itEPO->second.end(); ++itDAT)
       cout << "  " << t_gnwmbase::met_data_id(itDAT->first) << ": " << itDAT->second;
     
    cout << endl;
  }
#endif

  map<t_gpair, double> OUT;
  map<t_gpair, double> plane_data;
  map<t_gpair, double> plane_datG;
  map<t_gpair, double> plane_datH;
  map<t_gpair, double> plane_datA;
   
  map<t_gpair, double>::const_iterator  itA;
  map<t_gpair, double>::const_iterator  itY;
  map<t_gpair, double>::const_iterator  itG;
  map<t_gpair, double>::const_iterator  itH;

  double deg2met = D2R * A_WGS;   // if applicable, degree to ~meter conversion!
  double hTropo = 12;             // [km] Troposphere height at REF POINT
  MET_DATA dtyp = ZTD;

  // loop over epochs (interpolated points)
  for( itEPO = rf_data.begin(); itEPO != rf_data.end(); ++itEPO)
  {
    t_gtime tt(itEPO->first);
    t_map_met::iterator itM = _map_met.find(tt);     

    if( itM == _map_met.end() ) cout << "*** Warning no data in map (should not happen): " << tt.str_ymdhms() << endl;
//  else                        cout << "Epoch: " << tt.str_ymdhms()  << endl;

    double rf_ztd = rf_data[tt][ZTD]; // [m]
    double rf_zhd = rf_data[tt][ZHD]; // [m]
    double rf_zwd = rf_data[tt][ZWD]; // [m]

    for(t_map_dat::iterator itD = itM->second.begin(); itD != itM->second.end(); ++itD)
    {
      t_gnwmsurf* gsurf = static_cast<t_gnwmsurf*>(itD->second);
       
      t_gtriple grd( gsurf->get_surf(LAT),gsurf->get_surf(LON),gsurf->get_surf(HEL) ); // ALWAYS USE ELL

#ifdef DEBUG
      if( _proj ){
        t_gtriple xyz = _proj->ll2proj( ell ); // USE THE PROJECTION!
        cout << "MSG: USING THE LCC PROJECTION !" << fixed << setprecision(3)
	     << setw(14) << ell[0] << setw(14) << ell[1] << setw(14) << ell[2] << " --> "
	     << setw(14) << xyz[0] << setw(14) << xyz[1] << setw(14) << xyz[2] << endl;
      }
#endif

      // relative coordinates w.r.t. point of interes, given in [m]
//    t_gpair ellgrd = itD->first;                                    // [deg] always LatLon !
      t_gpair ellgrd = grd.gpair();                                   // [deg] always LatLon !
      t_gpair ellrel(( ellgrd[0] - ell[0] )*deg2met,                  // [m]
                     ( ellgrd[1] - ell[1] )*deg2met );                // [m]
       
      double dist = sqrt(pow(ellrel[0],2) + pow(ellrel[1],2))/1000.0; // [km]
      double elev = R2D*atan(hTropo/dist);                            // [deg]
      double azim = 90.0 - R2D*atan2(ellrel[0],ellrel[1]);            // [deg]
      if( azim < 0 ) azim += 360.0;
#ifdef DEBUG
    cout << fixed << setprecision(1) 
	 << "dist: " << dist  << "  azi: " << azim
	 << "  ele: " << elev << "  cut: " << cut_off 
         << endl;
#endif

      // collect nearby grid points
      if( elev >= cut_off ){                                        // [deg]
        double zwd_norm = gsurf->get_in_hel( ZWD, ell[2] );
        double zhd_norm = gsurf->get_in_hel( ZHD, ell[2] );
        double value = 0.0;
//  	double correct  = tan(elev*D2R);
//	double correct  = 1/dist;
	switch ( dtyp ){
          case ZHD : value = (rf_zhd - zhd_norm           ); break; // [m]
	  case ZWD : value = (rf_zwd -            zwd_norm); break; // [m]
	  case ZTD : value = (rf_ztd - zhd_norm - zwd_norm); break; // [m]
	  default  : value = 0.0; break;
	}
        plane_data[ellrel] = value;                     // ZTD or ZHD or ZWD [m]
        plane_datG[ellrel] = value*1000*tan(elev*D2R);  // gradient [mm]
        plane_datH[ellrel] = value*1000/dist;           // ala gradient [mm/km]
        plane_datA[ellrel] = azim*D2R;                  // [deg]
              
//#ifdef DEBUG
       double hel      = gsurf->get_surf(HEL);
       double zwd      = gsurf->get_surf(ZWD);
       double zhd      = gsurf->get_surf(ZHD);
       cout << fixed << setprecision(0)
//	    << ellgrd[0] << " " << ellgrd[1]
	    << "  CRD:"  << setw(8) << ellrel[0] << " "   << setw(8) << ellrel[1] << setprecision(1)
	                 << setw(7) << hel       << " ->" << setw(7) << ell[2]    << setprecision(3)
	    << "  ZHD:"  << setw(6) << zhd       << " ->" << setw(6) << zhd_norm
            << "  ZWD:"  << setw(6) << zwd       << " ->" << setw(6) << zwd_norm
            << "  ZTD:"  << setw(6) << zhd+zwd   << " ->" << setw(6) << zhd_norm+zwd_norm
            << "  REF:"  << setw(6) << rf_ztd
	    << " "       << setw(6) << (zhd_norm + zwd_norm - rf_ztd)
//	    << " "       << setw(8) << dist      // [km]
	                 << setprecision(1)
	    << "  SKY: " << setw(5) << elev
	    << " "       << setw(6) << azim
	    << endl;
//#endif
      }
    }
     
//#ifdef DEBUG
    cout << "\n DATA FORE PLANE FIT,  REQ PT: " << fixed << ell[0] << " " << ell[1] << " " << ell[2] << endl;
    for( itA = plane_datA.begin(), itY = plane_data.begin(), itG = plane_datG.begin(), itH = plane_datH.begin();
	 itG != plane_datG.end(); ++itA, ++itY, ++itG, ++itH )
    {
        cout << fixed << setprecision(0) << setw(9) << itA->first[0] << setw(9) << itA->first[1]
                      << setprecision(5) << setw(9) << itG->second << " mm "
                      << setprecision(5) << setw(9) << itH->second << " ??"
                      << setprecision(5) << setw(9) << itY->second << " m \n";
    }
//#endif

    if( plane_data.size() < 4 ){ cerr << "*** ERROR: TOO FEW DATA\n"; return -1; }
    // OLD
    ColumnVector fit(3);  fit  << 0.0 << 0.0 << 0.0;
    ColumnVector rms(3);
    t_gfunc_plane Call_plane_LSQ;   Call_plane_LSQ.LSQ_plane(plane_data, OUT, fit);
     
    // NEW
    ColumnVector fitG(2);  fitG << 0.0 << 0.0;
    ColumnVector rmsG(2);  rmsG << 0.0 << 0.0;
    ColumnVector fitH(3);  fitH << 0.0 << 0.0 << 12.0; // hTropo [km]
    ColumnVector rmsH(3);  rmsH << 0.0 << 0.0 <<  0.0;
     
    ColumnVector X(plane_datA.size());     itA = plane_datA.begin();
    ColumnVector G(plane_datG.size());     itG = plane_datG.begin();
    ColumnVector H(plane_datH.size());     itH = plane_datH.begin();

    map<string, double> cf;
    int iter = 0;
    string msg;

    t_gfunc_linCS funcCS(cf);  
    t_gfitlmm gfit2(&funcCS);

    t_gfunc_linCSH funcCSH(cf);
    t_gfitlmm gfit3(&funcCSH);

    for(unsigned int i=0; i < plane_datA.size(); ++i, ++itA, ++itY, ++itG, ++itH){
      X[i+1] = itA->second;
      G[i+1] = itG->second;
      H[i+1] = itH->second;
    }

    if( gfit2.fit( X, G, fitG, rmsG, iter, msg ) < 0 ){ cerr << "*** ERROR: fitting problem!\n"; return -1; }
//  if( gfit3.fit( X, H, fitH, rmsH, iter, msg ) < 0 ){ cerr << "*** ERROR: fitting problem!\n"; return -1; }
	
    // gradients
    for( map<t_gtime, map<MET_DATA, double>>::iterator itI = rf_data.begin(); itI!= rf_data.end(); ++itI )
    {
      if( itEPO->first != itI->first ) continue;

      // DATA TYPE!
      for(map<MET_DATA, double>::iterator J = itI->second.begin(); J!=itI->second.end(); ++J)
      {
	 
	if(J->first == dtyp) // tu to treba zmenit!!!
	{
          double value = J->second;        // ZTD, ZHD, ZWD [m]
#ifdef DEBUG
	  cout << "\n ZTD  = " << rf_ztd*1000                     // ZTD [mm]
               << " "          << t_gnwmbase::met_data_str(dtyp) 
	       << " "          << value*1000                      // ZTD/ZHD/ZWD [mm]
	       << setprecision(5)
	       << " fit:" << setw(9) << fit(1) << setw(9) << fit(2) << setw(9) << fit(3) << endl;
#endif
//#ifdef DEBUG
	  cout << "\n "          << itEPO->first.str_ymdhms()
               << " " << setw(3) << itEPO->first.doy()
	              << setw(6) << itEPO->first.sod()
               << "  "           << t_gnwmbase::met_data_str(dtyp) 
	       << setprecision(3)
	       << "  "           << value                         // ZTD/ZHD/ZWD [m]
	       << setprecision(4)
	       << "  fit:"  << setw(9) << fitG(1) << setw(9) << fitG(2)
	       << "  rms:"  << setw(8) << rmsG(1) << setw(8) << rmsG(2)
	       << setprecision(4)
	       << "  fitH:" << setw(9) << fitH(1) << setw(9) << fitH(2) << setw(9) << fitH(3)
	       << "  rmsH:" << setw(8) << rmsH(1) << setw(8) << rmsH(2) << setw(9) << rmsH(3)
	       << endl;
//#endif
#ifdef DEBUG	   
	  cout << fixed << setprecision(2)
               << " DOY/SOD: "          << setw(3) << itEPO->first.doy()
	                                << setw(6) << itEPO->first.sod()
               << " "                   << t_gnwmbase::met_data_str(dtyp) << setprecision(1)
               << " #cut: "             << setw(5) << cut_off             << setprecision(4)  // [km]
               << " #N: "               << setw(9) << fit(1)*value*1000 // NS gradient [mm]
               << " #E: "               << setw(9) << fit(2)*value*1000 // EW gradient [mm]
               << endl;
#endif
#ifdef DEBUG
          cout << " FIT: " << fixed 
 	       << " EPO: " << itI->first.str_ymdhms() 
               << setprecision(0) << " cut: " << cut_off
	       << setprecision(8) << " FC: "  << fit(1) << "  " << fit(2) << "  " << fit(3) << endl;

          double SUMA = 0.0;
          cout << "---------------------------------------------------------------------------------------" << "\n";
          cout << "    SN               WE          Input    Output  =>  (In-Out) **     PAR. in epoch(t) " << "\n";
          cout << "---------------------------------------------------------------------------------------" << "\n";
          for(map<t_gpair, double>::iterator II = plane_datG.begin(); II!=plane_datG.end(); ++II)
          {
            for(map<t_gpair, double>::iterator JJ = OUT.begin(); JJ!=OUT.end(); ++JJ)
            {
              if(II->first==JJ->first)
              {
                SUMA += (II->second-JJ->second);
		ostringstream os;
                os   << fixed    << setprecision(4)
		     << setw(15) << II->first.crd(0)                 // SN
		     << setw(15) << II->first.crd(1)                 // WE
		     << setw(9)  << II->second                       // Input
		     << setw(9)  << JJ->second                       // Fitted
		     << "  => " << setw(8) << II->second-JJ->second  // Input - Fitted
		     << "  ** " << setw(8) << J->second              // Station
		     << " ==> " << setw(8) << II->second-J->second   // Input - Station
		     << endl;
		cout << os.str();
//#ifdef DEBUG
                ofstream out;
                out.open((itEPO->first.str_ymdhms()).c_str());
		out << os.str();
		out.close();
//#endif
	      }
	    }
	  }

          cout << "---------------------------------------------------------------------------------------" << "\n";
          cout << "SUMA of Differences:       " << "\n";
          printf("%e", SUMA);
          cout << "  "<< endl;
#endif

#ifdef DEBUG
          ofstream out;
          ostringstream ooo, aaa, ddd;
          double YE, DO, HO;
          YE = itEPO->first.year();
          DO = itEPO->first.doy();
          HO = itEPO->first.hour();
          ooo << YE;  string ye = ooo.str();
          aaa << DO;  string dd = aaa.str();
          ddd << HO;  string ho = ddd.str();

          out.open(("RES/"+ye+"-"+dd+"-"+ho).c_str());
          out << "  X  Y  Z  " << itEPO->first.str_ymdhms() << endl;
          for(map<t_gpair, double>::iterator II = plane_datG.begin(); II!=plane_datG.end(); ++II)
          {
	    for(map<t_gpair, double>::iterator JJ = OUT.begin(); JJ!=OUT.end(); ++JJ)
	    {
	      if(II->first==JJ->first)
	      {
	        out << fixed << setprecision(4)
		    << "  "  << II->first.crd(0)
		    << "  "  << II->first.crd(1) 
		    << "  "  << II->second
		    << "  " << JJ->second
		    << "  =>  " << II->second-JJ->second
		    << " ** " << J->second
		    << " ==> " << II->second-J->second 
		    << endl;
		}
	    }	     
	  }
          out.close();
#endif
	}
      }
    }
  }

  OUT.clear();
  plane_data.clear();
  plane_datG.clear();
  plane_datH.clear();
  plane_datA.clear();
   
#ifdef DEBUG
   // just for information: mf, mfg solutions
   _call_mf( );
#endif
   return 0;
}
   
void t_gallsurf::_call_mf()
     {
	/*
	 * 1/sin(eps);
	 * 1/tg()eps;
	 * 1/sin(eps)*tg(eps)+C
	 */
	for(int j = 1; j<=90; j++)
	  cout << fixed << setprecision(5) << j << "  mf: " << 1/sin(j/(180/M_PI)) << " mf_G: " << 1/tan(j/(180/M_PI)) << " : " << 1/((sin(j/(180/M_PI))*tan(j/(180/M_PI)))+0.0032) << endl;
     }
   
// CallRadius: function focused on selection of in-area grind points (or other points?)
// ------------
int t_gallsurf::_call_radius(const t_gtriple& ell, vector<double>& glat, vector<double>& glon, double& rad_fac, vector<t_gpair>& gLatLon)
{
  if(glat.size() == 0 || glon.size()==0) return -1;
	
  double lat = ell.crd(0);
  double lon = ell.crd(1);
	
  for(unsigned int i = 0; i<glon.size(); i++)
  {
     double temp = sqrt(pow((lat - glat[i]),2) + pow((lon - glon[i]),2));
     if(temp <= rad_fac)
     {
       t_gpair gtemp(glat[i], glon[i]);
       gLatLon.push_back(gtemp);
     }
  }
	
  return 0;
}
   

// CallRadius: function focused on selection of in-area grind points (or other points?)
// ------------
double t_gallsurf::_call_radius(const t_gtriple& ell, t_gpair& temp)
{

  double SD = sqrt(pow((ell[0] - temp[0]),2) + pow((ell[1] - temp[1]),2));
  return SD;
}


// VECTOR ANALYSE
// ------------
int t_gallsurf:: _call_vector_analyse(const t_gtriple& ell, ColumnVector& fit, string& dir, double& BETA, double& GRAD)
{ 
  /*tu som sa snazil urcit priesecnik roviny a priamky*/
  if(fit.size() == 0){ cerr << "Warning: No fitted plane coeficients" << endl; return -1;}
#ifdef DEBUG
  cout << ell[0] << "  " << ell[1] << "  " << ell[2] << "  >  " << fit(1) << "  " << fit(2) << "  " << fit(3) << endl;
#endif 
  // Over, ze ci existuje priesecnik priamky a roviny
  if(dir == "E")
  {
    // lambda, E, x
    /* double x = (ell[2]-fit(2)*ell[0]-fit(3))/fit(1); //cout << x << endl;*/
    BETA = 90.0-(180.0/M_PI)*acos((fit(1)*1.0)/(sqrt( fit(1)*fit(1) + fit(2)*fit(2) + 1.0 )));
    // BETA = 90.0-(180.0/M_PI)*acos((fit(1)*1.0 + fit(3)*ell[2])/(sqrt( (1.0+(ell[2]*ell[2]))*(fit(1)*fit(1) + fit(2)*fit(2) + 1.0) )));
    GRAD = ell[1]*(M_PI/180)*fit(2);
  }
  else if(dir == "N")
  {
    // phi, N, y
    /* double y = (ell[2]-fit(1)*ell[1]-fit(3))/fit(2); //cout << y << endl;*/
    BETA = 90.0-(180.0/M_PI)*acos((fit(2)*1.0)/(sqrt( fit(1)*fit(1) + fit(2)*fit(2) + 1.0 )));
    // BETA = 90.0-(180.0/M_PI)*acos((fit(2)*1.0 + fit(3)*ell[2])/(sqrt( (1.0 + (ell[2]*ell[2]))*(fit(1)*fit(1) + fit(2)*fit(2) + 1.0) )));
    GRAD = ell[0]*(M_PI/180)*fit(1);
  }	
  return 0;
}

   
// new element allocated in heap
// ----------
int t_gallsurf::_new_element(const t_gtime& epo,
	 	 	     const t_gpair& LatLon, 
	 		           t_gnwmsurf* met)
{
// gtrace("t_gallsurf::new_element");

//_map_met[epo][LatLon] = met; // TAKE OK
  _map_met[epo][LatLon] = new t_gnwmsurf( *(t_gnwmsurf*)(met) );  
//  cout << "new element SURF : " << met->get_surf(LAT) << " "  << met->get_surf(LON)  << " " 
//       << sizeof(_map_met[epo][LatLon]) << endl;
  return 0;
}


// set region of interest
// ----------
int t_gallsurf::_set_region(const double minLat,
	 	 	    const double maxLat,
	 		    const double minLon,
	 		    const double maxLon
			   )
{
  gtrace("t_gallsurf::set_region");

  _min_lat = minLat;
  _max_lat = maxLat;
  _min_lon = minLon;
  _max_lon = maxLon;
  return 0;
}


// Set param
// -----------------------------
void t_gallsurf::_set_param( const set<string>& param )
{
  gtrace("t_gallsurf::set_param");

  // the parameters requested in the interpolation
  if( param.find("TEMP") != param.end() ) _vec_par.push_back(TEMP); // MUST BE FIRST
  if( param.find("T")    != param.end() ) _vec_par.push_back(TEMP);
  if( param.find("PRES") != param.end() ) _vec_par.push_back(PRES); // MUST BE SECOND
  if( param.find("P")    != param.end() ) _vec_par.push_back(PRES);
  if( param.find("EPRE") != param.end() ) _vec_par.push_back(EPRE);
  if( param.find("E")    != param.end() ) _vec_par.push_back(EPRE);
  if( param.find("GEOM") != param.end() ) _vec_par.push_back(GEOM);
  if( param.find("HGT")  != param.end() ) _vec_par.push_back(GEOM);
  if( param.find("H")    != param.end() ) _vec_par.push_back(GEOM);
   
  if( param.find("IWV")  != param.end() ) _vec_par.push_back(IWV);
  if( param.find("ZHD")  != param.end() ) _vec_par.push_back(ZHD);
  if( param.find("ZWD")  != param.end() ) _vec_par.push_back(ZWD);
  if( param.find("ZTD")  != param.end() )
  {	
   if( param.find("ZHD") == param.end() ) _vec_par.push_back(ZHD);
   if( param.find("ZWD") == param.end() ) _vec_par.push_back(ZWD);
                                          _vec_par.push_back(ZTD);  // MUST BE AFTER ZHD+ZWD
  }
   
  if( param.find("SHUM") != param.end() ) _vec_par.push_back(SHUM);
  if( param.find("Q")    != param.end() ) _vec_par.push_back(SHUM);
  if( param.find("TM")   != param.end() ) _vec_par.push_back(TM);
  if( param.find("Tm")   != param.end() ) _vec_par.push_back(TM);
  if( param.find("GP")   != param.end() ) _vec_par.push_back(GEOP);
  if( param.find("G")    != param.end() ) _vec_par.push_back(GRAV);

  // not yet supported (due to resulting map from the interpolation)
  if( param.find("dE")    != param.end() ) _vec_srf.push_back(MP_Ep);
  if( param.find("dW")    != param.end() ) _vec_srf.push_back(MP_Wp);
  if( param.find("dT")    != param.end() ) _vec_srf.push_back(MP_Tp);
  if( param.find("dTm")   != param.end() ) _vec_srf.push_back(MP_Mp);
   
  if( param.find("RATIO") != param.end() ) _vec_par.push_back(RATX);
  if( param.find("RATX")  != param.end() ) _vec_par.push_back(RATX);
  if( param.find("dR")    != param.end() ) _vec_srf.push_back(MP_Rt);
/*
  if( param.find("MP_Ep") != param.end() ) _vec_srf.push_back(MP_Ep);
  if( param.find("MP_Eh") != param.end() ) _vec_srf.push_back(MP_Eh);
  if( param.find("MP_Wp") != param.end() ) _vec_srf.push_back(MP_Wp);
  if( param.find("MP_Wh") != param.end() ) _vec_srf.push_back(MP_Wh);
  if( param.find("MP_Tp") != param.end() ) _vec_srf.push_back(MP_Tp);
  if( param.find("MP_Th") != param.end() ) _vec_srf.push_back(MP_Th);
  if( param.find("MP_Mp") != param.end() ) _vec_srf.push_back(MP_Mp);
  if( param.find("MP_Mh") != param.end() ) _vec_srf.push_back(MP_Mh);
*/
  return;
}

// Set output
// -----------------------------
void t_gallsurf::_set_output()
{
  gtrace("t_gallsurf::set_output");

  if( ! _set ) return;

  string tmp;
  tmp = dynamic_cast<t_gsetout*>(_set)->outputs("out");
  if( ! tmp.empty() ){
//    substitute(tmp,"$(rec)",_site, false);
    _out = new t_giof;
    _out->tsys();
    _out->mask(tmp);
    _out->append( dynamic_cast<t_gsetout*>(_set)->append() );
  }
   
  tmp = dynamic_cast<t_gsetout*>(_set)->outputs("sum");
  if( ! tmp.empty() ){
//    substitute(tmp,"$(rec)",_site, false);
    _sum = new t_giof;
    _sum->tsys();
    _sum->mask(tmp);
    _sum->append( dynamic_cast<t_gsetout*>(_set)->append() );
  }

  tmp = dynamic_cast<t_gsetout*>(_set)->outputs("grd");
  if( ! tmp.empty() ){
    _grd = new t_giof;
    _grd->tsys();
    _grd->mask(tmp);
    _grd->append( dynamic_cast<t_gsetout*>(_set)->append() );
  }


  tmp = dynamic_cast<t_gsetout*>(_set)->outputs("srf");
  if( ! tmp.empty() ){
    _srf = new t_giof;
    _srf->tsys();
    _srf->mask(tmp);
    _srf->append( dynamic_cast<t_gsetout*>(_set)->append() );
  }

  tmp = dynamic_cast<t_gsetout*>(_set)->outputs("fit");
  if( ! tmp.empty() ){
    _fit = new t_giof;
    _fit->tsys();
    _fit->mask(tmp);
    _fit->append( dynamic_cast<t_gsetout*>(_set)->append() );
  }

  return;
}

} // namespace
