
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <cstring>
#include <memory>
#include <algorithm>

#include "gcoders/rinexo3.h"
#include "gdata/gobsgnss.h"
 
using namespace std;

namespace gnut {

// constructor
// ----------
t_rinexo3::t_rinexo3( string version, int sz )
  : t_rinexo2( version, sz )
{
}

   
// OBS-RINEX header
//   - read individual lines (tmpsize once used and consumed)
//   - read block of lines at once (tmpsize cummulated and then consumed)
// ----------
int t_rinexo3::_decode_head()
{ 
  gtrace("t_rinexo3::decode_head");

  // -------- "SYS / # / OBS TYPES" --------
  if( _line.find("SYS / # / OBS TYPES", 60) != string::npos ){

    // new system (read all lines at once)
    string sys(1,_line[0]);
    int fact = 1;
    if( sys != " " ){

      // guarantee that all relevant lines are available
      int num = str2int(_line.substr(3,3));
      int nlines = ceil(num/13.0);         // upto 13 values/line
      int addsize = 0;
      string tmp;
      for( int nl = 0; nl < nlines;  ++nl ){
        int irc = t_gcoder::_getline(tmp, _tmpsize + addsize);
        if( irc >= 0 ){ 
          addsize += irc;
        }else{ 
          // cout << "INCOMPLETE RECORD(OBS TYPES): " << tmp << "\n";
          _complete = false;
          break;
        }
      }

      if( ! _complete ) return -1;

      // sufficient number of lines is already guaranteed here
      for(int ii=0; ii<num; ++ii){
        // make sure that the obs is not already in t_vobstypes
        string go = _line.substr(7 + 4*(ii%(13)), 3);
        _fix_band( sys, go ); // FIX BAND
        GOBS obs = str2gobs( go );

        t_vobstypes::iterator it = _mapobs[sys].begin();

	// record exists
	for( it = _mapobs[sys].begin(); it != _mapobs[sys].end(); ++it )
	  if( it->first == obs ){ it->second = fact; break; }

	// new record
        if( it == _mapobs[sys].end() ) _mapobs[sys].push_back( make_pair(obs, fact) );

        if( (ii+1)%13 == 0 && num > 13) // get newline and check (read next for old system)
          _tmpsize += t_gcoder::_getline(_line, _tmpsize );

      } // loop over # observations
    }

    if( _log && _log->verb() >= 3 ){
      ostringstream lg;
      lg << "SYS / # / OBS TYPES: [" << sys << "] ";
	 
      t_vobstypes::const_iterator it = _mapobs[sys].begin();
      while( it != _mapobs[sys].end() ){ 
        lg << gobs2str( it->first ) << " ";
        it++;	 
      }
      _log->comment(3,"rinexo3",lg.str());
    }


  // -------- "SYS / SCALE FACTOR" --------
  }else if( _line.find("SYS / SCALE FACTOR",  60) != string::npos ){
       
    // new system (read all lines at once)
    string sys(1,_line[0]);
    int fact = str2int(_line.substr(2,4));

    if( sys != " " ){
	 
      // guarantee that all relevant lines are available
      int num = str2int(_line.substr(8,2));
      int nlines = ceil(num/13.0);         // upto 13 values/line
      int addsize = 0;
      string tmp;
      for( int nl = 0; nl < nlines;  ++nl ){
        int irc = t_gcoder::_getline(tmp, _tmpsize + addsize);
        if( irc >= 0 ){ 
          addsize += irc;
        }else{ 
          // cout << "INCOMPLETE RECORD(OBS TYPES): " << tmp << "\n";
	  _complete = false;
	  break;
        }	 
      }

      if( ! _complete ) return -1;

      // sufficient number of lines is already guaranteed here
      for(int ii=0; ii<num; ++ii){
	// make sure that the obs is not already in t_vobstypes
        string go = _line.substr(11 + 4*(ii%(13)), 3);
        _fix_band( sys, go ); // FIX BAND
        GOBS obs = str2gobs( go );	   
        t_vobstypes::iterator it = _mapobs[sys].begin();

        for( it = _mapobs[sys].begin(); it != _mapobs[sys].end(); ++it )
	  if( it->first == obs ){ it->second = fact; break; }

        if( it == _mapobs[sys].end() )
	  _mapobs[sys].push_back( make_pair(obs, fact) );

  	if( (ii+1)%13 == 0 && num > 13) // get newline and check (read next for old system)
	  _tmpsize += t_gcoder::_getline(_line, _tmpsize );

      } // loop over # observations
    }
	 
    if( _log && _log->verb() >= 3 ){
      ostringstream lg;
      lg << "SYS / SCALE FACTOR: [" << sys << "] ";
	 
      t_vobstypes::const_iterator it = _mapobs[sys].begin();
      while( it != _mapobs[sys].end() ){ 
	lg << " " << gobs2str( it->first );
        it++;	 
      }
      _log->comment(3,"rinexo3",lg.str());
    }

    if( _log ) _log->comment(3,"rinexo3","SYS / SCALE FACTOR: "     + string("not implemented"));

  // -------- "SIGNAL STRENGTH UNIT" --------
  }else if( _line.find("SIGNAL STRENGTH UNIT",60) != string::npos ){              
    _rnxhdr.strength( trim(_line.substr(0,20)) );
    if( _log ) _log->comment(3,"rinexo3","SIGNAL STRENGTH UNIT: " + _rnxhdr.strength());

     
  // -------- "SYS / DCBS APPL" --------
  }else if( _line.find("SYS / DCBS APPL",     60) != string::npos ){
    if( _log ) _log->comment(3,"rinexo3","SYS / DCBS APPL: "        + string("not implemented"));

     
  // -------- "SYS / PCVS APPL" --------
  }else if( _line.find("SYS / PCVS APPL",     60) != string::npos ){
   if( _log ) _log->comment(3,"rinexo3","SYS / PCVS APPL: "        + string("not implemented"));

     
  // -------- "GLONASS SLOT / FRQ # --------
  }else if( _line.find("GLONASS SLOT / FRQ #",60) != string::npos ){
    if( _log ) _log->comment(3,"rinexo3","GLONASS SLOT / FRQ #: "   + string("not implemented"));

     
  // --------"# OF SATELLITES" --------
  }else if( _line.find("# OF SATELLITES",     60) != string::npos ){
       
    _rnxhdr.numsats( str2int(_line.substr(0,6)) );
    if( _log ) _log->comment(3,"rinexo3","# OF SATELLITES: " + int2str(_rnxhdr.numsats()));

     
  // --------"PRN / # OF OBS" --------
  }else if( _line.find("PRN / # OF OBS",      60) != string::npos ){
    if( _log ) _log->comment(3,"rinexo3","PRN / # OF OBS: "         + string("not implemented"));

  }else if( t_rinexo2::_decode_head() );

  return 1;
}

   
// OBS-RINEX body
//   - read block of lines for each epoch (tmpsize cummulated and then consumed)
// ----------
int t_rinexo3::_decode_data()
{  
  gtrace("t_rinexo3::_decode_data");
   
  int irc = t_rinexo3::_read_epoch();

//if( _flag != '0' && _flag != '1' ){ // special event handling
//  if( _log ) _log->comment(3,"rinexo2",_epoch.str_ymdhms("RINEX 3.x special events not handled after "));
//  else                         cerr << _epoch.str_ymdhms("RINEX 3.x special events not handled after ") << "\n";

  // not ready to read
  if( irc < 0 ){ _complete = false; return -1; }

  // skip special lines
  if( irc > 0 ){
    for( int i = 0; i < _nsat; ++i ){
      int sz = t_gcoder::_getline( _line, _tmpsize );
      if( sz <= 0 ){ _complete = false; return -1; }
      _tmpsize += sz;
//      cout << "_line skip: " << _line;
    }

    t_gcoder::_consume(_tmpsize);
    _consume += _tmpsize;
    _tmpsize = 0;
    return 0;
  }
   

  // DATA READING 
  // ------------
  // loop over satellits records (x-lines)
  for( int i = 0; i < _nsat; i++ ){

    string sat = "";
    string key = "";
  
    // loop over all observation types for system/satellite key in _mapobs
    if( t_rinexo3::_read_obstypes( sat, key ) < 0 ){ _complete = 0; return -1; }
  }       

  if( _complete ){

    if( _log && _log->verb() >= 5 ) _log->comment(5,"rinexo3",_epoch.str_ymdhms("reading finished at "));
    _count += this->_fill_data();
    t_gcoder::_consume(_tmpsize);
    _consume += _tmpsize;
    _tmpsize = 0;	   
  }
 
  return 0;
}
   

// read epoch & number of satellites, return flag
// ----------
int t_rinexo3::_read_epoch()
{
  gtrace("t_rinexo3::_read_epoch");

  _nsat = 0;
  _flag = 'X';
   
  // not enough data to recognize standard epoch or special event
  if( _line.length() < 35 ) return _stop_read();
  _flag = _line[31]; //  cout << "_flag = [" << _flag << "]\n";
  _nsat = str2dbl( _line.substr(32,3) );

  if( _line.substr(0,6) == "      " ){
    switch ( _flag ){
     case '2' : 
        if( _log ) _log->comment(2,"rinexo3","Warning - start moving antenna identified, but not yet implemented!");
      //else              cerr <<  "rinexo3 - Warning - start moving antenna identified, but not yet implemented!\n";
        return 1;
        
     case '3' :
        if( _log ) _log->comment(2,"rinexo3","Warning - new site occupation identified, but not yet implemented!");
      //else              cerr <<  "rinexo3 - Warning - new site occupation identified, but not yet implemented!\n";
        return 1;

     case '4' :
        if( _log ) _log->comment(2,"rinexo3","Warning - new header information identified, but not fully implemented!");
      //else              cerr <<  "rinexo3 - Warning - new header information identified, but not fully implemented!\n"; 
        return 1;
 
     case '5' :
        if( _log ) _log->comment(2,"rinexo3","Warning - external event identified, but not yet implemented!");
      //else              cerr <<  "rinexo3 - Warning - external event identified, but not yet implemented!\n";
        return 1;
    }
  }
  string dummy;
  int yr,mn,dd,hr,mi;
  double sc;
  istringstream is(_line);
  is.clear();

  is >> dummy >> yr >> mn >> dd >> hr >> mi >> sc >> _flag >> _nsat;
  // double offs = 0.0; istr >> offs; istr.clear(); // optional receiver offset (currently not used)

  // check success
  if( is.fail() ){ // cout << "rinexo: problem reading EPOCH:" << _line;    
    return _stop_read();
  }
   
/* // ONLY A SLIGHTLY FASTER THEN THE ABOVE
  string tmp =  _line.substr(2,33);
  int irc = sscanf ( tmp.c_str(),"%4d %2d %2d %2d %2d %10lf %2c %2d", &yr, &mn, &dd, &hr, &mi, &sc, &_flag, &_nsat );
  
  if( irc < 8 ){ cout << "rinexo: problem reading EPOCH" << irc << ":" << tmp << " [" << _nsat << "]\n";
    return _stop_read();
  }
*/
  _epoch.from_ymdhms(yr, mn, dd, hr, mi, sc);
	 
  // filter out data
  if( _flag != '0' &&
      _flag != '1'  ) return 1;
  if( _epoch > _end ) return 2;
  if( _epoch < _beg ) return 2;

  // this filters sampling rate which, however, works up to 1Hz only!
  if( _int > 0 && int(round(_epoch.sod()+_epoch.dsec()))%int(_int) != 0 ) return 2; // ROUND FOR PLUS/MINUS DRIFT !!

#ifdef DEBUG
  if( _log ) _log->comment(0,"rinexo3","reading epoch: " + _epoch.str_ymd() + " " + _epoch.str_hms());
#endif

  return 0;
}


// read satellite observation types
// ----------
int t_rinexo3::_read_obstypes( const string& sat, const string& sys )
{
  gtrace("t_rinexo3::_read_obstypes");

  int ii = 0;
  int addsize = 0;
  unsigned int idx = 0;

  // general sys- or sat- specific group of observations
  string key = sat;
  map<string,t_vobstypes>::const_iterator itMAP = _mapobs.find(sat);
  if( itMAP == _mapobs.end() ){ key = sys; }

  string tmpsat("");

  // read new line for individual satellite (complete observations)
  if( ( addsize = t_gcoder::_getline( _line, _tmpsize ) ) >= 3 ){
    _tmpsize += addsize;
    tmpsat = t_gsys::eval_sat( _line.substr(1,2), t_gsys::char2gsys(_line[0]) );
  }else  return _stop_read();

  t_spt_gobs obs = make_shared<t_gobsgnss>(_rnxhdr.markname(), tmpsat, _epoch);

  // loop over sys-defined observation types
  t_vobstypes::const_iterator it = _mapobs[tmpsat.substr(0,1)].begin();
  unsigned int len;
  while( _complete && (it != _mapobs[tmpsat.substr(0,1)].end()) ){

    idx = 3 + 16*ii;
    len = _line.length();

    // getline succeed only if 'EOL' found
    if( len <= 0 && len < (idx + 14) ) return _stop_read();

    // check completness (excluding last SNR+LLI - sometimes missing)
    if( len < (idx + 14) ) _null_log( tmpsat, gobs2str(it->first) );
    else                   _read_obs( idx, it, obs );

    ii++; it++;

  } // while (over observation types)

  // filter GNSS and SAT
  string gs = t_gsys::char2str( tmpsat[0] ); // DONT USE SYS FROM ARG WHICH MAY BE EMPTY
  GSYS gsys = t_gsys::str2gsys( gs );        // DONT USE SYS FROM ARG WHICH MAY BE EMPTY
  if(( _sys.size()       > 0 &&       _sys.find(     gs ) == _sys.end()       ) ||
     ( _sat[gsys].size() > 0 && _sat[gsys].find( tmpsat ) == _sat[gsys].end() ))
  {
    if( _log ) _log->comment( 2, "rinexo3", "skip " + gs + " " + tmpsat);
  }else{ _vobs.push_back(obs); }

#ifdef DEBUG
  cout << " RNX: "     << _version 
       << " sys: "     << sys
       << " key: "     << key
       << " code/phase ID"
//       << " - band1:"  << " " << gobs2str( obs.code(1)  )
//                       << " " << gobs2str( obs.phase(1) )
//       << " - band2:"  << " " << gobs2str( obs.code(2)  )
//                       << " " << gobs2str( obs.phase(2) )
       << endl;
#endif     

  return 1;
}
 

// fix band
// ----------
int t_rinexo3::_fix_band(string sys, string& go)
{
  gtrace("t_rinexo3::_fix_band [" + sys + "]");

#ifdef DEBUG
  cout << " sys : " << sys << " " << t_gsys::str2gsys(sys) 
       << " gobs: "  << go << "  GO : " << go.substr(1,1)
       << " "  << str2gobsband(go)
       << endl;
#endif

  // EXCEPTION FOR BEIDOU (B2 --> B1)
  if( t_gsys::str2gsys(sys) == BDS &&  go[1] == '2' )
  {
    go[1] = '1'; 
    if( _version >= "3.02" )
          if( _log ) _log->comment(0,"rinexo3","warning - BDS band not valid for RINEX >= 3.02 [changed B2->B1]");
          else               cerr << "rinexo3:  warning - BDS band not valid for RINEX >= 3.02 [changed B2->B1]\n";
       
    else  if( _log ) _log->comment(0,"rinexo3","message - BDS band changed: B2->B1");
          else               cerr << "rinexo3:  message - BDS band changed: B2->B1\n";
  } 

  return 1;
}

} // namespace
