
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <sstream>

#include "gdata/gnavbds.h" 

using namespace std;

namespace gnut {

// constructor
// ----------
t_gnavbds::t_gnavbds()
  : t_gnav(),
    _iode(0),
    _iodc(0),
    _health(0),
    _toe(t_gtime::GPS),
    _toc(t_gtime::GPS),
    _tot(t_gtime::GPS),
    _a(0.0),
    _e(0.0),
    _m(0.0),
    _i(0.0),
    _idot(0.0),
    _omega(0.0),
    _OMG(0.0),
    _OMGDOT(0.0),
    _dn(0.0),
    _crc(0.0),
    _cic(0.0),
    _cuc(0.0),
    _crs(0.0),
    _cis(0.0),
    _cus(0.0),
    _f0(0.0),
    _f1(0.0),
    _f2(0.0),
    _acc(0.0)
{ 
  gtrace("t_gnavbds::t_gnavbds");
   
  _tgd[0] = _tgd[1] = 0.0;

  id_type(t_gdata::EPHBDS);
  id_group(t_gdata::GRP_EPHEM);
}


// desctructor
// ----------
t_gnavbds::~t_gnavbds(){
  gtrace("t_gnavbds::~t_gnavbds");
}


// operator <
// .. priority: satid, tot, iode
// ----------
/*
int t_gnavbds::operator<(const t_gnavbds& data) const
{       
  int _dtot = int( tot.diff( data.tot ) );
  int _dsat = strcmp( sat, data.sat );
  int _diod = int( iode - data.iode );

  if( _dsat != 0 ) return (_dsat < 0);  // satellite name
  if( _dtot != 0 ) return (_dtot < 0);  // time of transmition
  if( _diod != 0 ) return (_diod < 0);  // iode of ephemeris

  return false;
}
*/

// operator ==
// .. distinguish: satid, tot, iode
// ----------
/*
int t_gnavbds_struct::operator==(const t_gnavbds& data) const
{       
  int _dtot = int( tot.diff( data.tot ) );
  int _dsat = strcmp( _sat, data.sat );
  int _diod = int( _iode - data.iode );
  
  return ( _dtot == 0 && _dsat == 0 && _diod == 0 );
}
*/

// print function
// ----------
string t_gnavbds::str() const
{
  gtrace("t_gnavbds::str");   
   
  int w = 20;
  ostringstream tmp;

  tmp  << " " << setw(3) << sat()
       << " " << _toc.str("%Y-%m-%d %H:%M:%S")
//     << " " << _toe.str("%Y-%m-%d %H:%M:%S")
//     << " " << _tot.str("%Y-%m-%d %H:%M:%S")
       << scientific << setprecision(12)
       << setw(w) << _f0
       << setw(w) << _f1
       << setw(w) << _f2
       << setw(w) << _iode/1.0
       << setw(w) << _iodc/1.0
       << setw(w) << _acc/1.0
       << setw(w) << SQRT(_a)
       << setw(w) << _e
       << setw(w) << _i
       << setw(w) << _m
       << setw(w) << _idot
       << setw(w) << _omega
       << setw(w) << _OMG
       << setw(w) << _OMGDOT
       << setw(w) << _dn
       << setw(w) << _crc
       << setw(w) << _cic
       << setw(w) << _cuc
       << setw(w) << _crs
       << setw(w) << _cis
       << setw(w) << _cus
       << setw(w) << _tgd[0]
       << setw(w) << _tgd[1];

  return tmp.str();
}


// check message // NEED TO IMPROVE !
// ----------
int t_gnavbds::chk(string& msg)
{
  gtrace("t_gnavbds::chk");
   
  if( _toe == t_gtime::first_time() ){
    msg = "warning: navbds invalid toe" + _toe.str_ymdhms(" "+_sat+" ");
    return -1;
  }

  if( _iode < 0 ||  255 < _iode){
    msg = "warning: navbds invalid iode" + _toe.str_ymdhms(" "+_sat+" ");
    return -1;
  }
   
  if( _iodc < 0 || 1023 < _iodc ){	
    msg = "warning: navbds invalid iodc" + _toe.str_ymdhms(" "+_sat+" ");
    return -1;
  }  
   
  if( _a < 20000000.0 ){
    msg = "warning: navbds invalid _a" + _toe.str_ymdhms(" "+_sat+" ");
    return -1;
  }

  // .. other tests
  return 0;
}


// return URA
// ----------
/* int t_gnavbds::ura( double acc ) const */
/* { */
/*   gtrace("t_gnavbds::ura"); */
/*   unsigned int i; */
/*   for( i = 0; i < sizeof(ura_eph); i++) */
/*     if(acc > ura_eph[i]) break; */
/*   return i; */
/* } */


//
// t is GPS time of transmission,
// i.e. GPS time corrected for transit time (range/speed of light)
// ----------
int t_gnavbds::pos( const t_gtime& t, double xyz[], double var[], double vel[], bool chk_health )
{
  gtrace("t_gnavbds::pos");   
   
  if( sat().empty() ) return -1; // check if valid
  if( chk_health && _healthy() == false) return -1;
   
  xyz[0] = xyz[1] = xyz[2] = 0.0;
  if( var ) var[0] = var[1] = var[2] = 0.0;
  if( vel ) vel[0] = vel[1] = vel[2] = 0.0;

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  double Tk = t.diff(_toe);                         // T - toe difference
  while( Tk >  302400.0 ){ Tk -= 604800.0; }        // check the correct week
  while( Tk < -302400.0 ){ Tk += 604800.0; }        // check the correct week
 
  double Ek,dEk;
  _ecc_anomaly( Tk, Ek, dEk );                      // eccentric anomaly and its derivative

  double V = atan2(sqrt(1.0 - pow(_e,2.0))*sin(Ek), 
		                     cos(Ek) - _e); // true anomaly
  double U0 = V + _omega;                           // argument of latitude
   
  double sin2U = sin(2*U0);
  double cos2U = cos(2*U0);
   
  double Rk = _a*(1.0 - _e * cos(Ek));              // radius               
  double Uk = U0;                                   // argument of latitude 
  double Ik = _i + _idot*Tk;                        // inclination          
   
         Rk += _crs*sin2U + _crc*cos2U;             // corrected radius
         Ik += _cis*sin2U + _cic*cos2U;             // corrected inclination
         Uk += _cus*sin2U + _cuc*cos2U;             // corrected argument of latitude 
    
  double cosU   = cos(Uk);
  double sinU   = sin(Uk);
  double cosI   = cos(Ik);
  double sinI   = sin(Ik);
   
  double x = Rk*cosU;                               // position in orbital plane
  double y = Rk*sinU;
    
  double OMG = 0.0; 
  // MEO/IGS0 satellites 
  OMG = _OMG + (_OMGDOT - OMGE_DOT_BDS)*Tk 
                      - OMGE_DOT_BDS*(_toe.sow()-14);        // corrected long. of asc. node
   
  // GEO satellites
  if( t_gsys::bds_geo(sat()) ){
     OMG = _OMG + _OMGDOT*Tk 
                         - OMGE_DOT_BDS*(_toe.sow()-14);        // corrected long. of asc. node
  }

  double cosOMG = cos(OMG);
  double sinOMG = sin(OMG);
   
  xyz[0] = x*cosOMG - y*cosI*sinOMG;                // earth-fixed coordinates [m]
  xyz[1] = x*sinOMG + y*cosI*cosOMG;
  xyz[2] = y*sinI;

  if( t_gsys::bds_geo(sat()) ){
     double cosa = cos(-5.0*D2R);
     double sina = sin(-5.0*D2R);
     double cosb = cos(OMGE_DOT_BDS*Tk);
     double sinb = sin(OMGE_DOT_BDS*Tk);

     t_gtriple xyzG = xyz;
     xyz[0] =  xyzG[0]*cosb + xyzG[1]*sinb*cosa + xyzG[2]*sinb*sina;                // earth-fixed coordinates [m]
     xyz[1] = -xyzG[0]*sinb + xyzG[1]*cosb*cosa + xyzG[2]*cosb*sina;
     xyz[2] =               - xyzG[1]*sina      + xyzG[2]*cosa;     
  } 
   
  // error variance
  if( var ){ 
     *var = SQRT( _acc );
  }

  // velocities
  if( vel ){
    double n  = _mean_motion();

    // BNC implementation (begin)
    double tanv2 = tan(V/2);
    double dEdM  = 1 / (1 - _e*cos(Ek));
    double dotv  = sqrt((1.0 + _e)/(1.0 - _e)) / cos(Ek/2)/cos(Ek/2) / (1 + tanv2*tanv2) 
               * dEdM * n;
    double dotu  = dotv + (-_cuc*sin2U + _cus*cos2U)*2*dotv;
    double dotom = _OMGDOT - OMGE_DOT_BDS;
    double doti  = _idot + (-_cic*sin2U + _cis*cos2U)*2*dotv;
    double dotr  = _a * _e*sin(Ek) * dEdM * n
                + (-_crc*sin2U + _crs*cos2U)*2*dotv;
    double dotx  = dotr*cosU - Rk*sinU*dotu;
    double doty  = dotr*sinU + Rk*cosU*dotu;

    vel[0]  =     cosOMG*dotx  -    cosI*sinOMG*doty
             - x *sinOMG*dotom - y *cosI*cosOMG*dotom
                               + y *sinI*sinOMG*doti;

    vel[1]  =     sinOMG*dotx  +    cosI*cosOMG*doty
             + x *cosOMG*dotom - y *cosI*sinOMG*dotom
                               - y *sinI*cosOMG*doti;

    vel[2]  =     sinI*doty    + y *cosI*doti;     
    // BNC implementation (end)

/* 
    // bcccalc implementation (beg)
    // ----------------------------
    // ....... 4.1  compute the derivative of omegak and dlat w.r.t. time tk.
    // note:  the formula for the derivative of dlat is
    // fairly complicated to work out.

    double dOMG = _OMGDOT - OMGE_DOT_BDS;
    double dlat = _mean_motion() * sqrt(1.0 - _e*_e) * pow(_a*Rk,2.0);

    //....... 4.2  compute the derivative of the 2nd harmonic corrections
    // with respect to time tk. (i.e. compute the derivative of:
    // 1 - radius correction
    // 2 - argument of latitude correction
    // 3 - correction to inclination

    double dsin2U =  2.0 * cos2U * dlat;
    double dcos2U = -2.0 * sin2U * dlat;

    double ddRk = _crc * dcos2U + _crs * dsin2U;
    double ddIk = _cuc * dcos2U + _cus * dsin2U;
    double ddUk = _cic * dcos2U + _cis * dsin2U;

    //....... 4.3  compute the derivative of the following w.r.t. time tk:
    // 1) corrected argument of latitude
    // 2) corrected radius
    // 3) corrected inclination

    double corrRk  = _a*_e * sin(Ek)*dEk + ddRk;
    double corrIk  = ddIk;
    double corrUk  = dlat + ddUk;

    //....... 4.4  compute the derivative of the x and y positions in the
    // orbital plane, and finally, compute the velocity vector,
    // which is the derivative of the earth's fixed coordinates

    double dx = corrRk * cosU  -  Rk * sinU * corrUk;
    double dy = corrRk * sinU  +  Rk * cosU * corrUk;

    vel[0] =   ( dx            * cosOMG )
             - ( dy * cosI  * sinOMG )
             - (  x            * sinOMG * dOMG )
             - (  y * cosI  * cosOMG * dOMG )
             + (  y * sinI  * sinOMG * corrIk );

    vel[1] =   ( dx            * sinOMG )
             + ( dy * cosI  * cosOMG )
             + (  x            * cosOMG * dOMG )
             - (  y * cosI  * sinOMG * dOMG )
             - (  y * sinI  * cosOMG * corrIk );

    vel[2] =   ( dy * sinI )
             + (  y * cosI* corrIk );
    // bcccalc implementation (end)
 */ 
//    cout << " vel = " << fixed << setprecision(15) 
//         << " " << vel[0]
//         << " " << vel[1]
//         << " " << vel[2] << endl;
  }   

  _gmutex.unlock(); return 0;
}


//
// t is GPS time of transmission,
// i.e. GPS time corrected for transit time (range/speed of light)
// NOT CORRECTED FOR 2nd-order relativistic effect
// ----------
int t_gnavbds::clk( const t_gtime& t, double* clk, double* var, double* dclk, bool chk_health )
{   
  gtrace("t_gnavbds::clk");
   
  if( sat().empty() ) return -1;  // not valid !!!
  if( chk_health && _healthy() == false) return -1;   
   
#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  double Tk = t.diff(_toc);                      // difference from clk(!) epoch ref. time
  while( Tk >  302400.0 ){ Tk -= 604800.0; }     // check the correct GPS week
  while( Tk < -302400.0 ){ Tk += 604800.0; }     // check the correct GPS week

  double Ek, dEk;
  _ecc_anomaly( Tk, Ek, dEk );                   // eccentric anomaly
   
// SV time correction (including periodic relativity correction)
//  cout << "RELATIVISTIC EFFECT  uSec: " << - 2.0*sqrt(GM_WGS84*_a)*_e*sin(Ek)/CLIGHT/CLIGHT*1000000 << endl;

//  *clk = _f0 + Tk*_f1 + Tk*Tk*_f2  - 2.0*sqrt(GM_WGS84*_a)*_e*sin(Ek)/CLIGHT/CLIGHT;  // [sec]

// NOT CORRECTED FOR 2nd-order relativistic effect - at user side
  *clk = _f0 + Tk*_f1 + Tk*Tk*_f2;
   
  // check impossible clock correction
  if( clk && *clk > 1.0 ){
    cerr << "* warning: too large clock - skipped: " << fixed << setprecision(0) << *clk << endl; 
    _gmutex.unlock(); return -1;
  }
   
  // error variance
  if( var ){
     *var = SQRT( _acc );
  }

  // velocities
  if( dclk ){
     *dclk = 0.0;
  }

  _gmutex.unlock(); return 0;
}


// convert general gnavdata structure to gnavgps_struct
// 
//   ToC, ToE, ToT not definitely implemented!
// ----------
int t_gnavbds::data2nav( string sat, const t_gtime& ep, const t_gnavdata& data )
{
  gtrace("t_gnavbds::data2nav");   
   
#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  char tmp[12];
  _toe.tsys(t_gtime::GPS);
  _tot.tsys(t_gtime::GPS);
  _toc.tsys(t_gtime::GPS);
  _epoch.tsys(t_gtime::GPS);

  if( sat.substr(0,1) == "C" )  _sat =     sat;
  else                          _sat = "C"+sat;   

  // time of clocks (toc)
  _toc = ep; // (RINEX reference time)
  _epoch = ep;

  double data_21(data[21]);
  double data_27(data[27]);

  if( data_21 < 0 ){
   _toe = t_gtime::first_time(); // cerr << " incorrect BDS week number !\n";
   _gmutex.unlock(); return -1;
  }   
   
  // RINEX has a continuou week, thus next check is not necessary !
  data_21 += 1356; // conversion to GPS week
//  while( fabs(data_21 - _toc.gwk()) > 1 ) data_21 += 8191; // correct for the BDS week roll-over
   
  // time of ephemeris (toe)                [ week + seconds of week ]
  sprintf(tmp,"%4i %6i", static_cast<int>(data_21), static_cast<int>(data[11]));
  _toe.from_str("%W %v",tmp);

  // time of transmission (tot)
  int addwk = 0;
  if( data_27 > +302400 ){ addwk -= 1; }
  if( data_27 < -302400 ){ addwk += 1; }
  sprintf(tmp,"%4i %6i", _toe.gwk() + addwk, static_cast<int>(data_27));
  _tot.from_str("%W %v",tmp);
  

  _f0     =      data[0];
  _f1     =      data[1];
  _f2     =      data[2];
   
  _iode   = static_cast<int>(data[3]);
  _iodc   = static_cast<int>(data[28]);
  _health = static_cast<int>(data[24]);
   
  _a      =  SQR(data[10]);
  _e      =      data[8];
  _m      =      data[6];
  _i      =      data[15];
  _idot   =      data[19];
  _OMG    =      data[13];
  _OMGDOT =      data[18];
  _omega  =      data[17];
  _dn     =      data[5];

  _crs    =      data[4];
  _cus    =      data[9];
  _cis    =      data[14];
  _crc    =      data[16];
  _cuc    =      data[7]; 
  _cic    =      data[12];
   
  _acc    =      data[23];
  _tgd[0] =      data[25];
  _tgd[1] =      data[26];

  // Temporary conversion into GPS time
  // Must be replaced to t_gtime
  _toe = _toe + 14;
  _toc = _toc + 14;
  _tot = _tot + 14;

  _gmutex.unlock(); return 0;
}


// convert gnav_bds element to general gnavdata
// ----------
int t_gnavbds::nav2data( t_gnavdata& data )
{
  gtrace("t_gnavbds::nav2data");

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  data[0]  = _f0;
  data[1]  = _f1;
  data[2]  = _f2;   
  data[3]  = _iode;
  data[4]  = _crs;
  data[5]  = _dn;
  data[6]  = _m;
  data[7]  = _cuc;
  data[8]  = _e;
  data[9]  = _cus;
  data[10] = sqrt(_a);
  data[11] = _toe.sow();
  data[12] = _cic;
  data[13] = _OMG;
  data[14] = _cis;
  data[15] = _i;
  data[16] = _crc;
  data[17] = _omega;
  data[18] = _OMGDOT;
  data[19] = _idot;
  data[21] = _toe.gwk();
  data[23] = _acc;             // ura_eph[_acc]; // [m]
  data[24] = _health;
  data[25] = _tgd[0];
  data[26] = _tgd[1];
  data[27] = _tot.sow();
  data[28] = _iodc;

  while( data[27] <       0 ){ data[27] += 604800; }
  while( data[27] > +604800 ){ data[27] -= 604800; }

//cout << "=> DATA27: " << data[27] << endl;

  _gmutex.unlock(); return 0;
}



// printing 
// ----------
int t_gnavbds::print()
{
#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif

  _gmutex.lock();
  cout << "+GNAV\n" << str() << "\n-GNAV\n";
  _gmutex.unlock(); return 0;
}


// healthy check
bool t_gnavbds::_healthy() const
{   
   if( _health == 0 ) return true;
   return false;
}


// eccentric anomaly
// ----------
double t_gnavbds::_mean_motion()
{
  gtrace("t_gnavbds::_mean_motion");   
   
  double n0 = sqrt(GM_CGCS/(pow(_a,3.0)));      // computed mean motion [rad/sec]
  double n  = n0 + _dn;                          // corrected mean motion

  return n;
}


// eccentric anomaly
// ----------
void t_gnavbds::_ecc_anomaly( double dt, double& Ek, double& dEk )
{
  gtrace("t_gnavbds::_ecc_anomaly");   
   
  double n  = _mean_motion();                    // eccentric anomaly
  double Mk = _m + n*dt;                         // mean anomaly

  Ek = Mk;
  dEk = n;
  for( double E = 0; fabs(Ek-E)*_a > 0.001; ){
    E  = Ek;
    Ek = Mk + _e*sin(E);                         // kepler equation for eccentric anomaly [rad]
   dEk = n  + _e*cos(Ek)*dEk;                    // derivatives w.r.t. time
  }
}

} // namespace
