
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iostream>

#include "gmodels/gdop.h"
 
using namespace std;

namespace gnut {   

// constructor
// ----------
t_gdop::t_gdop()
{
   gtrace("t_gdop::t_gdop");
   
   _gnav = 0;
   _gobs = 0;
   _site = "";
   _Qx.ReSize(4);
}

t_gdop::t_gdop(t_gallnav* gnav, t_gallobs* gobs, string site)
{
   gtrace("t_gdop::t_gdop");
   
   _gnav = gnav;
   _gobs = gobs;
   _site = site;
   _Qx.ReSize(4);   
}


// Destructor
// ----------
t_gdop::~t_gdop()
{
   gtrace("t_gdop::~t_gdop");   
}

// set nav, obs, site
void t_gdop::set_data(t_gallnav* gnav, t_gallobs* gobs, string site)
{
   gtrace("t_gdop::set_data");   
   
   _gnav = gnav;
   _gobs = gobs;
   _site = site;   
}

// set log
void t_gdop::set_log(t_glog* glog)
{
   _log = glog;
}

// Calculate dop - _Qx
int t_gdop::calculate(t_gtime& epoch, t_gtriple& rec, GSYS gnss)
{  
   gtrace("t_gdop::calculate");   

   _rec = rec;
   
   if ( _gobs ) _sats = _gobs->sats(_site, epoch, gnss);
   else{
          cerr << "t_gdop:: WARNING - t_gallobs not set" << endl;
          if( _log ) _log->comment(1, "t_gdop","WARNING - t_gallobs not set");
          return -1;
       }

   if ( _sats.size() == 0 ) return -1;
   unsigned int Nsat = _sats.size();

   Matrix A(Nsat, 4); A = 0;
   int i = 0;
   for (set<string>::iterator it = _sats.begin(); it != _sats.end(); it++) {

      double xyz[3]  = {0.0, 0.0, 0.0};
      double vel[3]  = {0.0, 0.0, 0.0};
      double var[3]  = {0.0, 0.0, 0.0};

      int irc = _gnav->pos( *it, epoch, xyz, var, vel);

      if (irc < 0) continue;

      t_gtriple satpos(xyz);

      double rho = (_rec.crd_cvect() - satpos.crd_cvect()).norm_Frobenius();

      i++;
      A(i,1) = (_rec[0]-satpos[0]) / rho;
      A(i,2) = (_rec[1]-satpos[1]) / rho;
      A(i,3) = (_rec[2]-satpos[2]) / rho;
      A(i,4) = 1.0;
   }
   A = A.Rows(1,i);    // delete zero rows   
   if (A.Nrows() < 4) return -1;

   Matrix NN = A.t() * A;
   _Qx << NN.i();
   
   return 1;
}


// Position dilution of precision
double t_gdop::pdop()
{
   gtrace("t_gdop::pdop");      
   
   if ( _Qx.Ncols() != _Qx.Nrows() ) return -1.0;
   if ( _Qx.Ncols() != 4 ) return -1.0;
   
   return sqrt( _Qx(1,1)*_Qx(1,1) + _Qx(2,2)*_Qx(2,2) + _Qx(3,3)*_Qx(3,3) );
}

// Geom dilution of precision
double t_gdop::gdop()
{
   gtrace("t_gdop::gdop");      
   
   if ( _Qx.Ncols() != _Qx.Nrows() ) return -1.0;
   if ( _Qx.Ncols() != 4 ) return -1.0;
   
   return sqrt( _Qx(1,1)*_Qx(1,1) + _Qx(2,2)*_Qx(2,2) + _Qx(3,3)*_Qx(3,3) + _Qx(4,4)*_Qx(4,4));
}

// Time dilution of precision
double t_gdop::tdop()
{

  gtrace("t_gdop::tdop");

   
   if ( _Qx.Ncols() != _Qx.Nrows() ) return -1.0;
   if ( _Qx.Ncols() != 4 ) return -1.0;   
   
   return sqrt( _Qx(4,4)*_Qx(4,4) );
}

} // namespace
