
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdlib.h>
#include <iostream>
#include <iomanip>

#include "gmodels/gpcv.h"
#include "gutils/gsysconv.h"
#include "gutils/gtypeconv.h"
#include "gutils/gmatrixconv.h"
#include "gmodels/ginterp2d.h"

using namespace std;

namespace gnut {   

// -------------------------------
// T_GPCV - single antenn patterns
// -------------------------------

// constructor
// ----------
t_gpcv::t_gpcv()
 : _trans(true),    // transmitter (default yes)
   _anten(""),      // antenna type
   _ident(""),      // antenna identification
   _svcod(""),      // SVN code
   _method(""),     // calibartion method
   _source(""),     // source of calibration
   _snxcod("")      // sinex code 
{
  _beg.tsys(t_gtime::GPS);
  _end.tsys(t_gtime::GPS);
  id_type(t_gdata::PCV);   
}

// destructor
// ----------
t_gpcv::~t_gpcv(){}

// return correction ARP/CoM -> phase center
// zen [rad] - for receiver antenna, it is zenith angle to the satellite
//           - for satellite antenna, it is nadir angle (receiver zenith minus
//             geocentric (space) angle btw XYZsat and XYZrec)
// azi [rad] - for receiver antenna -> need to be implemented !
//           - for satellite antenna not interesting (only if horizontal eccentricity exists!)
// ----------
double t_gpcv::pco( double zen, double azi, GFRQ frq )
{
#ifdef BMUTEX   
   boost::mutex::scoped_lock lock(_mutex);
#endif 
   _gmutex.lock();

  double corr = 0.0;
  if( zen > PI ){
    ostringstream lg;
    lg << "not valid zenith angle:" << zen*R2D << endl;
    if( _log ) _log->comment(0,"gpcv",lg.str());
     _gmutex.unlock(); return corr;
  }
   
  // satellite only PCO (Z-offset) mapped to rec-sat direction (approximated)
  if( _mappco.find(frq) != _mappco.end() ){
     corr = _mappco[frq][2] * cos(zen);  // for satellite zen should to be zen-alfa
  }
   
  // SATELLITE HORIZONTAL ECCENTRICITIES NOT YET IMPLEMENTED !!!!
  // RECEIVER PCO NOT YET IMPLEMENTED (azimut/zenith dependent)
   
  _gmutex.unlock(); return corr;
}

// Update satellite coordinates according to pco
// -----------------------------------------------
int t_gpcv::pcoS( t_gsatdata& satdata, t_gtriple& dx, string LC)
{
#ifdef BMUTEX   
   boost::mutex::scoped_lock lock(_mutex);
#endif 
   _gmutex.lock();
   
  t_gtime epo         = satdata.epoch();
  t_gtriple Satcrd_t  = satdata.satcrd();
  t_gtriple Satvel_t  = satdata.satvel();      // TEMPORARY
  GSYS sys            = satdata.gsys();
  string sat          = satdata.sat();  
  ColumnVector Satcrd = Satcrd_t.crd_cvect();
  ColumnVector Satvel = Satvel_t.crd_cvect();     // TEMPORARY
   
  double MJD = epo.dmjd(); 
  t_gtriple apcf1, apcf2, apcf3, apcf4, apcLC;

  GFRQ     f1, f2;// f3, f4;
  GOBSBAND b1, b2, b3, b4;
  if (sys == GPS){
     f1 = G01;    f2 = G02;
     b1 = BAND_1; b2 = BAND_1; b3 = BAND_5;
  }else if(sys == GLO){
     f1 = R01;    f2 = R02;     
     b1 = BAND_1; b2 = BAND_1; b3 = BAND_3;
  }else if(sys == GAL){

     b1 = BAND_1; b2 = BAND_5; b3 = BAND_6; b4 = BAND_7;
  }else if(sys == BDS){

     b1 = BAND_1; b2 = BAND_2; b3 = BAND_6; b4 = BAND_7;
  }else if(sys == QZS){

     b1 = BAND_1; b2 = BAND_2; b3 = BAND_5; b4 = BAND_6;
  }  
   
   if( _mappco.find(f1) != _mappco.end() && _mappco.find(f2) != _mappco.end() ){
      apcf1 = _mappco[f1];
      apcf2 = _mappco[f2];
   }else{ 
      t_map_offs offs = GNSS_PCO_OFFSETS();      
      t_map_offs::iterator itSYS = offs.find(sys);
      if(itSYS != offs.end()){
	 t_map_pcos::iterator itb1 = itSYS->second.find(b1);
	 t_map_pcos::iterator itb2 = itSYS->second.find(b2);
	 t_map_pcos::iterator itb3 = itSYS->second.find(b3);
	 t_map_pcos::iterator itb4 = itSYS->second.find(b4);
	 
	 if(itb1 != itSYS->second.end()) {
	    apcf1[0] = itb1->second[0];
	    apcf1[1] = itb1->second[1];
	    apcf1[2] = itb1->second[2];	    
	 }
	 
	 if(itb2 != itSYS->second.end()) {
	    apcf2[0] = itb2->second[0];
	    apcf2[1] = itb2->second[1];
	    apcf2[2] = itb2->second[2];
	 }
	 
	 if(itb3 != itSYS->second.end()) {
	    apcf3[0] = itb3->second[0];
	    apcf3[1] = itb3->second[1];
	    apcf3[2] = itb3->second[2];
	 }	 
	 
	 if(itb4 != itSYS->second.end()) {
	    apcf4[0] = itb4->second[0];
	    apcf4[1] = itb4->second[1];
	    apcf4[2] = itb4->second[2];
	 }	 
	 
      }
  }
   
  for (int i=0; i<=2; i++){
     apcf1[i] /= 1000.0;
     apcf2[i] /= 1000.0;
     apcf3[i] /= 1000.0;
     apcf4[i] /= 1000.0;

  }   
   
   // Satelite-Earth unit vector
   ColumnVector k = -Satcrd / Satcrd.NormFrobenius();

   // Along solar panel unit vector
   ColumnVector j;
   ColumnVector sun = _ephplan.sunPos(MJD).crd_cvect();
                sun = sun / sun.NormFrobenius(); 
                  j = crossproduct(k,sun);

   // BDS: GEO -> Y is perpendicular to the orbital plane
   if( sys == BDS || t_gsys::bds_geo(sat) ){
      ColumnVector v = Satvel / Satvel.NormFrobenius();
      j = crossproduct(k,v);
   }   

   // complete to satelite fixed right-hand coord system
   // GPS: Sun in +X hemisphere
   // GLO: Sun in +X hemisphere (IGS conventions)
   // BDS: MEO/IGSO -> Sun in +X hemisphere
   ColumnVector i = crossproduct(j,k);      
   // GAL: Sun in -X hemisphere
   if( sys == GAL ) i = -1 * i;

   // QZS: |beta| > 20 deg -> Sun in -X hemisphere; |beta| < 20 deg -> Y is perpendicular to the orbital plane, +y in anti-orbital momentum direction (south of orbital plane)
   if( sys == QZS){
      double beta = satdata.beta(sun);
      if(beta <= 20 && beta >= -20 ) i = -1 * i;
   }
   // PCO for linear combination
    if (LC == "L3"){
      double koef1 = 0.0;
      double koef2 = 0.0;       
      satdata.coef_ionofree(BAND_1, koef1, BAND_2, koef2);
      apcLC = apcf1*koef1 + apcf2*koef2;
    }else apcLC = apcf1;
      
   dx[0] = apcLC[0]*i(1) + apcLC[1]*j(1) + apcLC[2]*k(1);
   dx[1] = apcLC[0]*i(2) + apcLC[1]*j(2) + apcLC[2]*k(2);
   dx[2] = apcLC[0]*i(3) + apcLC[1]*j(3) + apcLC[2]*k(3);
   
#ifdef DEBUG
   cout << "PCO calculation for " << anten() << " " << epo.str_hms() << endl;
   cout << fixed << setprecision(4);
   cout << "pco f1: " << apcf1 << " m" << endl
        << "pco f2: " << apcf2 << " m" << endl      
        << "pco lc: " << apcLC << " m" << endl
        << fixed << setprecision(5)
        << "Sat dX:\n" << dx << endl;
#endif
   
   _gmutex.unlock(); return 1;   
}

// Update satellite coordinates according to pco
// -----------------------------------------------
int t_gpcv::pcoR( t_gsatdata& satdata, t_gtriple& dx, t_gtriple& site, string LC)
{
#ifdef BMUTEX   
   boost::mutex::scoped_lock lock(_mutex);
#endif 
   _gmutex.lock();
   

  GSYS sys = satdata.gsys();
     
  t_gtriple apcf1, apcf2, apcLC;

  GFRQ f1, f2;
  if (sys == GPS){
      f1 = G01;
      f2 = G02;
  }else if(sys == GLO){
      f1 = R01;
      f2 = R02;     
  }
   
  if( _mappco.find(f1) == _mappco.end() || _mappco.find(f2) == _mappco.end() ){
    _gmutex.unlock(); return -1;
  }

  apcf1 = _mappco[f1];
  apcf2 = _mappco[f2];
   
  for (int i=0; i<=2; i++){
     apcf1[i] /= 1000.0;
     apcf2[i] /= 1000.0;     
  }
   
   // PCO for linear combination
    if (LC == "L3"){
      double koef1 = 0.0;
      double koef2 = 0.0;
      satdata.coef_ionofree(BAND_1, koef1, BAND_2, koef2);              
      apcLC = apcf1*koef1 + apcf2*koef2;
    }else apcLC = apcf1;
   
   t_gtriple ell(0.0, 0.0, 0.0);
   xyz2ell(site, ell);
   neu2xyz(ell, apcLC, dx);
   
#ifdef DEBUG
   cout << "PCO calculation for " << anten() << endl;
   cout << fixed << setprecision(4);
   cout << "pco f1: " << apcf1 << " m" << endl
        << "pco f2: " << apcf2 << " m" << endl      
        << "pco lc: " << apcLC << " m" << endl
        << fixed << setprecision(5)
        << "Sat dX:\n" << dx << endl;
#endif
   
   _gmutex.unlock(); return 1;   
}

// return type
// ---------------
int t_gpcv::pco_proj(double& corr, t_gsatdata& satdata, t_gtriple& site, t_gtriple& dx, string LC)
{
   gtrace("t_gpcv::pco_proj");

//   t_gtriple dx(0.0, 0.0, 0.0);
//   int irc = this->pcoR(satdata, dx, LC);
   
   t_gtriple satcrd = satdata.satcrd();
   t_gtriple SatRec(0.0, 0.0, 0.0);
   SatRec = satcrd - site;
   ColumnVector e = SatRec.unitary();
   
   corr = DotProduct(dx.crd_cvect(),e);
   
   return 1;
}


// return type
// ----------
int t_gpcv::pcvS( double& corrLC, t_gsatdata& satdata, t_gtriple& site, string LC)
{   
  
#ifdef BMUTEX   
   boost::mutex::scoped_lock lock(_mutex);
#endif 
   _gmutex.lock();

  GSYS sys = satdata.gsys();
     
  GFRQ f1, f2;
  if (sys == GPS){
      f1 = G01;
      f2 = G02;
  }else if(sys == GLO){
      f1 = R01;
      f2 = R02;     
  }
   
  if( _mapzen.find(f1) == _mapzen.end() || _mapzen.find(f2) == _mapzen.end() ){
    _gmutex.unlock(); return -1;
  }
   
   map<double, double> mapDataf1;
   mapDataf1 = _mapzen[f1];
   
   map<double, double> mapDataf2;
   mapDataf2 = _mapzen[f2];   
     
   double eleS = satdata.ele();     
   double rS = satdata.satcrd().crd_cvect().NormFrobenius();
   double rR =         site.crd_cvect().NormFrobenius();
   
   double sinz = (rR/rS)*cos(eleS);
   double zen = asin(sinz) * R2D;

   double corrf1 = 0.0;
   double corrf2 = 0.0;
   
   t_ginterp2d interp;
   if ( interp.linear(mapDataf1, zen, corrf1) < 0 || interp.linear(mapDataf2, zen, corrf2) < 0 ){
      _gmutex.unlock(); return -1;
   }

   // [mm] -> [m]
   corrf1 /= 1000.0;
   corrf2 /= 1000.0;   
   
   // PCV for linear combination
    if (LC == "L3"){
      double koef1 = 0.0;
      double koef2 = 0.0;       
      satdata.coef_ionofree(BAND_1, koef1, BAND_2, koef2);              
      corrLC = corrf1*koef1 + corrf2*koef2;
    }else corrLC = corrf1;   
   
#ifdef DEBUG
   cout << "Sat pcv interpolation" << endl;
   cout << "Nadir = "    << zen 
	<< " corrf1 =  " << corrf1*1000 << " mm, "
	<< " corrf2 =  " << corrf2*1000 << " mm, "
	<< " corrLC =  " << corrLC*1000 << " mm, "	
	<< " PRN: " << anten() << endl;
#endif
   
   _gmutex.unlock(); return 1;
}

// return type
// ----------
int t_gpcv::pcvR( double& corrLC, t_gsatdata& satdata, string LC)
{   
#ifdef BMUTEX   
   boost::mutex::scoped_lock lock(_mutex);
#endif 
   _gmutex.lock();

   GSYS sys = satdata.gsys();
     
  GFRQ f1, f2;
  if (sys == GPS){
      f1 = G01;
      f2 = G02;
  }else if(sys == GLO){
      f1 = R01;
      f2 = R02;     
  }
   
  if( _mapzen.find(f1) == _mapzen.end() || _mapzen.find(f2) == _mapzen.end() ){
    _gmutex.unlock(); return -1;
  }

   map<double, double> mapDataf1;
   mapDataf1 = _mapzen[f1];
   
   map<double, double> mapDataf2;
   mapDataf2 = _mapzen[f2];         
     
   double zen = M_PI/2.0 - satdata.ele();        
          zen *= R2D;

   double corrf1 = 0.0;
   double corrf2 = 0.0;   
   
   t_ginterp2d interp;
   if ( interp.linear(mapDataf1, zen, corrf1) < 0 || interp.linear(mapDataf2, zen, corrf2) < 0 ){
      _gmutex.unlock(); return -1;
   }

   // [mm] -> [m]
   corrf1 /= 1000.0;
   corrf2 /= 1000.0;
   
   // PCV for linear combination
    if (LC == "L3"){
      double koef1 = 0.0;
      double koef2 = 0.0;       
      satdata.coef_ionofree(BAND_1, koef1, BAND_2, koef2);              
      corrLC = corrf1*koef1 + corrf2*koef2;
    }else corrLC = corrf1;   
  
#ifdef DEBUG
   cout << "Rec pcv interpolation" << endl;
   cout << "Zenith = "    << zen 
	<< " corrf1 =  " << corrf1*1000 << " mm, "
	<< " corrf2 =  " << corrf2*1000 << " mm, "
	<< " corrLC =  " << corrLC*1000 << " mm, "	
	<< " Ant: " << anten() << endl;
#endif   
   
   _gmutex.unlock(); return -1;   
}

} // namespace
