
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include "gmodels/gpppmodel.h"

namespace gnut {   

// Constructors
// -------------------
t_gpppmodel::t_gpppmodel(string site, t_glog* glog, t_gsetbase* settings)
 : t_gsppmodel(site, glog, settings),
   _gallobj(0)
{
   gtrace("t_gpppmodel::t_gpppmodel");   
   
   _site = site;
   _settings = settings;
   _log = glog;

   if (_trpModStr.compare("nwm")          == 0) tropoModel = new t_gtropo();

   _phase            = dynamic_cast<t_gsetproc*>(_settings)->phase();
   _grad_mf          = dynamic_cast<t_gsetproc*>(_settings)->grad_mf();

}

// Destructor
// --------------------------
t_gpppmodel::~t_gpppmodel()
{
   gtrace("t_gpppmodel::~t_gpppmodel");
}


// wind-up correction
// ----------
double t_gpppmodel::windUp(t_gtime& epoch, const string prn, const ColumnVector& rSat, const ColumnVector& rRec)
{
  gtrace("t_gpppmodel::windUp");   
  double Mjd = epoch.mjd(true) + epoch.sod(true) / 86400.0;

  // First time - initialize to zero
  // -------------------------------
  map<string, double>::iterator it = _windUpTime.find(prn);  
  if( it == _windUpTime.end() ){
      _windUpTime[prn] = Mjd;
      _windUpSum[prn]  = 0.0;
     }
 
  // Compute the correction for new time
  // -----------------------------------
  else if (_windUpTime[prn] != Mjd) {
         
    _windUpTime[prn] = Mjd;
      
    // Unit Vector GPS Satellite --> Receiver
    // --------------------------------------
    ColumnVector rho = rRec - rSat;
    rho /= rho.norm_Frobenius();

    // GPS Satellite unit Vectors sz, sy, sx
    // -------------------------------------
    ColumnVector sz = -rSat / rSat.norm_Frobenius();
  
    ColumnVector xSun = _ephplan.sunPos(Mjd).crd_cvect();
    xSun /= xSun.norm_Frobenius();
  
    // Along solar panel unit vector
    ColumnVector e = (xSun-rSat) / sqrt(DotProduct(xSun-rSat, xSun-rSat));
    ColumnVector sy = crossproduct(sz,e);     
     
     
//  ColumnVector sy = crossproduct(sz, xSun);
    ColumnVector sx = crossproduct(sy, sz);

    // Effective Dipole of the GPS Satellite Antenna
    // ---------------------------------------------
    ColumnVector dipSat = sx - rho * DotProduct(rho,sx) - crossproduct(rho, sy);

    // Receiver unit Vectors rx, ry
    // ----------------------------
    ColumnVector rx(3);
    ColumnVector ry(3); 

    double recEll[3]; xyz2ell(rRec.data(), recEll) ;
    double neu[3];

    neu[0] = 1.0;
    neu[1] = 0.0;
    neu[2] = 0.0;
    neu2xyz(recEll, neu, rx.data());

    neu[0] =  0.0;
    neu[1] = -1.0;
    neu[2] =  0.0;
    neu2xyz(recEll, neu, ry.data());
 
    // Effective Dipole of the Receiver Antenna
    // ----------------------------------------
    ColumnVector dipRec = rx - rho * DotProduct(rho,rx) + crossproduct(rho, ry);
 
    // Resulting Effect
    // ----------------
    double alpha = DotProduct(dipSat,dipRec) /(dipSat.norm_Frobenius() * dipRec.norm_Frobenius());

    if (alpha >  1.0) alpha =  1.0;
    if (alpha < -1.0) alpha = -1.0;

    double dphi = acos(alpha) / 2.0 / M_PI;  // in cycles

    if( DotProduct(rho, crossproduct(dipSat, dipRec)) < 0.0 ) dphi = -dphi;

    _windUpSum[prn] = floor(_windUpSum[prn] - dphi + 0.5) + dphi;
  }

#ifdef DEBUG
   cout << "Wind up " << prn << " " << _windUpSum[prn] << endl;
#endif   
   
  return _windUpSum[prn];
}


// set gallobj pointer
// --------------------
void t_gpppmodel::setOBJ(t_gallobj* obj)
{
   _gallobj = obj;
}

// model computed range value (phase/code)
// ----------
double t_gpppmodel::cmpObs(t_gtime& epoch, t_gallpar& param, t_gsatdata& gsatdata, bool phase, bool com)
{   
   gtrace("t_gpppmodel::cmpObs");   

   double spp_model = t_gsppmodel::cmpObs(epoch, param, gsatdata, phase);
   
  // Cartesian coordinates to ellipsodial coordinates
  t_gtriple xyz, ell;
  ColumnVector cRec(3);

  if ( param.getCrdParam(_site, xyz) > 0 ){
    cRec = xyz.crd_cvect();
  }else{
    xyz  = _aprox_crd_xml;
    cRec = xyz.crd_cvect();
  }	   
  xyz2ell(xyz, ell);   

  t_gtriple satcrd = gsatdata.satcrd();
  ColumnVector cSat = satcrd.crd_cvect();     
     
  // Wind up correction 
  double wind = 0.0;
  if (phase) wind = windUp(epoch, gsatdata.sat(), cSat, cRec) * gsatdata.wavelength_L3();
   
  // Phase center variation correction
  double pcv_R = 0.0;  // variation - receiver
  double pcv_S = 0.0;  // variation - satellite
  double pco_R = 0.0;  // offset    - receiver
  if( _gallobj != 0 && phase){
    shared_ptr<t_gobj> objnull;
    shared_ptr<t_gobj>  sat_obj = _gallobj->obj( gsatdata.sat() );
    shared_ptr<t_gobj> site_obj = _gallobj->obj( _site          );
     
    shared_ptr<t_gpcv>  pcvnull;
    shared_ptr<t_gpcv>  sat_pcv;
    shared_ptr<t_gpcv> site_pcv;

    if(  sat_obj != objnull )  sat_pcv =  sat_obj->pcv(epoch);
    if( site_obj != objnull ) site_pcv = site_obj->pcv(epoch);
           
    if(  sat_pcv != pcvnull && com){
      sat_pcv->pcvS(pcv_S, gsatdata, xyz, "L3");
      // sat pco is applied in t_gpppfilter::satPos
    }
              
    if( site_pcv != pcvnull ){
       site_pcv->pcvR(pcv_R, gsatdata, "L3");
       t_gtriple dx(0.0, 0.0, 0.0);       
       if ( site_pcv->pcoR(gsatdata, dx, xyz, "L3") > 0 ) site_pcv->pco_proj(pco_R, gsatdata, xyz, dx, "L3");
       pco_R *= -1;
    }     
  }   


#ifdef DEBUG
   cout << gsatdata.epoch().str_hms() <<  " " 
        << gsatdata.sat()  << fixed << setprecision(20)
        << "  rho: "       << gsatdata.rho()
        << "  ele: "       << setw(6) << gsatdata.ele()*180.0/PI
        << "  H:   "       << setw(6) << ell[2]  
        << "  SPP_model "  << setw(15)<< spp_model
        << "  wind: "      << setw(6) << wind
        << "  lambda3: "   << setw(6) << gsatdata.wavelength_L3()
        << "  Rec PCO: "   << setw(6) << pco_R          
        << "  Rec PCV: "   << setw(6) << pcv_R
        << "  Sat PCV: "   << setw(6) << pcv_S
        << endl;
//   int ooo; cin >> ooo;
#endif          
   
   // Return value
   return spp_model      +
          wind           +
          pco_R          +
          pcv_R          +
          pcv_S;
}


// Compute troposperic delay
// -----------
double t_gpppmodel::tropoDelay(t_gtime& epoch, t_gallpar& param, t_gtriple ell, t_gsatdata& satdata)
{        
   gtrace("t_gpppmodel::tropoDelay");   
      
   if (tropoModel == 0){
      if( _log ) _log->comment(0, "gppp","Tropo Model setting is not correct [" + _trpModStr + "] default used! Check config.");
      else                cerr << "gppp - Tropo Model setting is not correct [" + _trpModStr + "] default used! Check config.\n";
      
      tropoModel = new t_saast();
   }
   
   double ele = satdata.ele();
   double azi = satdata.azi();   
   
   double delay = 0.0;
   double zwd   = 0.0;
   
   int i, j, k;
   i = param.getParam(_site, t_gpar::TRP, "");
   j = param.getParam(_site, t_gpar::GRD_N, "");
   k = param.getParam(_site, t_gpar::GRD_E, "");   
   if ( i >= 0 ){
      zwd = param[i].value;
   }else zwd = tropoModel->getZWD(ell, epoch);

   if ( _tropo_mf.compare("gmf") == 0 ){
      double gmfh, gmfw, dgmfh, dgmfw;
	t_gmf mf;
      mf.gmf( epoch.mjd(), ell[0], ell[1], ell[2], M_PI/2.0-ele,
              gmfh, gmfw, dgmfh, dgmfw );
      delay = gmfh * tropoModel->getZHD(ell, epoch) + gmfw * zwd;
      
	if (j >= 0){
	   double grdN = 0;
         if ( _grad_mf.compare("tilt") == 0 )
	     grdN = param[j].value * dgmfw * cos(azi);
	   else if ( _grad_mf.compare("chh") == 0 ){
	     grdN = param[j].value*1000.0 * (1.0/(sin(ele)*tan(ele)+0.0031)) * cos(azi);
	     grdN /= 1000.0;
	   }
	   
	   delay += grdN;
	}
	if (k >= 0){
	   double grdE = 0;
         if ( _grad_mf.compare("tilt") == 0 )
	     grdE = param[k].value * dgmfw * sin(azi);
	   else if ( _grad_mf.compare("chh") == 0 ){
	     grdE = param[k].value*1000.0 * (1.0/(sin(ele)*tan(ele)+0.0031)) * sin(azi);
	     grdE /= 1000.0; 
	   }
	   
	   delay += grdE;
	}      
		
#ifdef DEBUG
    cout << epoch.str("EPOCH: %H:%M:%S") << endl << fixed << setprecision(3);
    cout << "Ell:" << ell[0] << " " << ell[1] << " " << ell[2]
	 << " Hydrostatic part: " << tropoModel->getZHD(ell, epoch)
	 << " Wet part: "        << zwd
	 << " gmfh: " << gmfh
	 << " gmfw: " << gmfw
         << " Delay: " << delay << endl << endl;	
#endif
   }else if ( _tropo_mf.compare("cosz") == 0 ){
      delay = 1/sin(ele) * tropoModel->getZHD(ell, epoch) +1/sin(ele) * zwd;
	double df = -( cos(ele) ) / ( sin(ele)*sin(ele) );
	if (j >= 0){
	   double grdN = 0;
         if ( _grad_mf.compare("tilt") == 0 )
	     grdN = param[j].value * df * cos(azi);
	   else if ( _grad_mf.compare("chh") == 0 ){
	     grdN = param[j].value*1000.0 * (1.0/(sin(ele)*tan(ele)+0.0031)) * cos(azi);
	     grdN /= 1000.0;
	   }
	   
	   delay += grdN;
	}
	if (k >= 0){
	   double grdE = 0;
         if ( _grad_mf.compare("tilt") == 0 )
	     grdE = param[k].value * df * sin(azi);
	   else if ( _grad_mf.compare("chh") == 0 ){
	     grdE = param[k].value * (1.0/(sin(ele)*tan(ele)+0.0031)) * sin(azi);
	     grdE /= 1000.0;
	   }
	   
	   delay += grdE;
        }           
   }
   
   return delay;
}

} // namespace
