
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/


#include "gmodels/gsppmodel.h"

namespace gnut {   

// Global variables
// -----------------------------------------------------------
const double   MAXRES_CODE      =   15.0;
const double   MAXRES_PHASE     =   0.04;
const double   MAXRES_CODE_GLO  =   30;
const double   MAXRES_PHASE_GLO =   0.08;

// -----------------------------------------------------------

// Constructors
// -------------------
t_gsppmodel::t_gsppmodel(string site, t_glog* glog, t_gsetbase* settings)
{
   gtrace("t_gsppmodel::t_gsppmodel");
   
   tropoModel = 0;
   
   _settings = settings;
   _site = site;
   _log = glog;
   _phase = false;
   
   _sigCodeGLO       = dynamic_cast<t_gsetproc*>(_settings)->sigma_C_GLO();
   _aprox_crd_xml[0] = dynamic_cast<t_gsetrec*>(_settings)->X();
   _aprox_crd_xml[1] = dynamic_cast<t_gsetrec*>(_settings)->Y();
   _aprox_crd_xml[2] = dynamic_cast<t_gsetrec*>(_settings)->Z();
   
   _tropo_mf         = dynamic_cast<t_gsetproc*>(_settings)->tropo_mf();      
   _trpModStr        = dynamic_cast<t_gsetproc*>(_settings)->tropo_model();

   if      (_trpModStr.compare("saastamoinen") == 0) tropoModel = new t_saast();
   else if (_trpModStr.compare("davis")        == 0) tropoModel = new t_davis();
   else if (_trpModStr.compare("hopfield")     == 0) tropoModel = new t_hopf();
}

// Destructor
// ---------------------
t_gsppmodel::~t_gsppmodel()
{
   gtrace("t_gsppmodel::~t_gsppmodel");
  if(tropoModel) delete tropoModel;      
}


// Outliers detection
// ----------
int t_gsppmodel::outlierDetect(   vector<t_gsatdata>& data,
			             SymmetricMatrix& Qx,
			       const SymmetricMatrix& Qsav, 
			       const ColumnVector& v,
			      vector<GOBSTYPE>& obstypes)
{
   gtrace("t_gsppmodel::outlierDetect");
   
   vector<t_gsatdata>::iterator it;
   
   double maxVcode     = 0.0;
   double maxVphase    = 0.0;
   double maxVcodeGLO  = 0.0;
   double maxVphaseGLO = 0.0;   
      
   vector<t_gsatdata>::iterator itMaxVcode;
   vector<t_gsatdata>::iterator itMaxVphase;
   vector<t_gsatdata>::iterator itMaxVcodeGLO;
   vector<t_gsatdata>::iterator itMaxVphaseGLO;   

   int ii = 0;
   for( it = data.begin(); it != data.end(); it++ ){
     ++ii;

     if (it->gsys() == GPS){
       if( maxVcode == 0.0 || abs(v(ii)) > maxVcode ){
	
         maxVcode = abs(v(ii));
         itMaxVcode = it;		     
       }
     }else if (it->gsys() == GLO && _sigCodeGLO < 999){
       if( maxVcodeGLO == 0.0 || abs(v(ii)) > maxVcodeGLO ){
	
         maxVcodeGLO = abs(v(ii));
         itMaxVcodeGLO = it;		     
       }	
     }else{
       if( maxVcode == 0.0 || abs(v(ii)) > maxVcode ){
	
         maxVcode = abs(v(ii));
         itMaxVcode = it;		     
       }	
     }

     if (it->gsys() == GLO && _sigCodeGLO >= 999) ii--;
      
     if( _phase ){
       ++ii;

	if (it->gsys() == GPS){	
         if( maxVphase == 0.0 || abs(v(ii)) > maxVphase ){

	   maxVphase = abs(v(ii));
	   itMaxVphase = it;
         }
       }else if (it->gsys() == GLO){
         if( maxVphaseGLO == 0.0 || abs(v(ii)) > maxVphaseGLO ){
       
	   maxVphaseGLO = abs(v(ii));
	   itMaxVphaseGLO = it;
         }
       }else{
         if( maxVphase == 0.0 || abs(v(ii)) > maxVphase ){

	   maxVphase = abs(v(ii));
	   itMaxVphase = it;
         }
       }
	
     }

   }
#ifdef DEBUG
    cout << "Max res range: " << maxVcode << endl;
    cout << "Max res phase: " << maxVphase << endl;
#endif

  if( maxVcodeGLO > MAXRES_CODE_GLO ){

     if( _log)
        _log->comment(2, "gppp", _site + " outlier (range) " + itMaxVcodeGLO->sat()
		                       + " size:" + int2str(data.size())
	  	                       + " v: "   + dbl2str(maxVcodeGLO)
	                               + " " + itMaxVcodeGLO->epoch().str_hms() );
    data.erase(itMaxVcodeGLO);
    Qx = Qsav;
    return 1;
  }      

  if( maxVphaseGLO > MAXRES_PHASE_GLO ){

    if( _log ) 
        _log->comment(2, "gppp", _site + " outlier (phase) " + itMaxVphaseGLO->sat()
		                       + " size:" + int2str(data.size())
		                       + " v: "   + dbl2str(maxVphaseGLO)
	                               + " " + itMaxVphaseGLO->epoch().str_hms() );	 
    data.erase(itMaxVphaseGLO);
    Qx = Qsav;
    return 1;
  }      

  if( maxVcode > MAXRES_CODE ){

    if( _log)
        _log->comment(2, "gppp", _site + " outlier (range) " + itMaxVcode->sat() 
		                       + " size:" + int2str(data.size())
	  	                       + " v: "   + dbl2str(maxVcode)
		                       + " ele: " + dbl2str(itMaxVcode->ele_deg())
	                               + " " + itMaxVcode->epoch().str_hms() );
    data.erase(itMaxVcode);
    Qx = Qsav;
    return 1;
  }

  if( maxVphase > MAXRES_PHASE ){

    if( _log )
        _log->comment(2, "gppp", _site + " outlier (phase) " + itMaxVphase->sat()
		                       + " size:" + int2str(data.size())
		                       + " v: "   + dbl2str(maxVphase)
		                       + " ele: " + dbl2str(itMaxVphase->ele_deg())		      
	                               + " " + itMaxVphase->epoch().str_hms() );
    data.erase(itMaxVphase);
    Qx = Qsav;
    return 1;
  }

  return 0;
}

// model computed range value 
// ----------
double t_gsppmodel::cmpObs(t_gtime& epoch, t_gallpar& param, t_gsatdata& gsatdata, bool phase, bool cmp)
{   
   gtrace("t_gsppmodel::cmpObs");   
   
  // Cartesian coordinates to ellipsodial coordinates
  t_gtriple xyz, ell;
  ColumnVector cRec(3);

  if ( param.getCrdParam(_site, xyz) > 0 ){
    cRec = xyz.crd_cvect();
  }else{
    xyz  = _aprox_crd_xml;
    cRec = xyz.crd_cvect();
  }	   
  xyz2ell(xyz, ell);   

  t_gtriple satcrd = gsatdata.satcrd();
  ColumnVector cSat = satcrd.crd_cvect();          

  string prn = gsatdata.sat(); 
   
  // Tropospheric wet delay correction
  double trpDelay = 0;
  trpDelay = tropoDelay(epoch, param, ell, gsatdata);   
  if(trpDelay > 50) trpDelay = 0;
   
  // Receiver clock correction 
  double clkRec = 0.0;
  int i;
  i = param.getParam(_site, t_gpar::CLK, "");
  if ( i >= 0 ) clkRec = param[i].value;
  else
    if( _log && _log->verb() >= 1) 
        _log->comment(1, "gppp", _site + " ! warning:  Receiver Clock is not included in parameters!");   
   
  // GLONASS system time offset
  double glonass_offset = 0;
  if (gsatdata.gsys() == GLO){
    int i;
    i = param.getParam(_site, t_gpar::GLO_ISB, "");
    if ( i >= 0 ) {
       glonass_offset = param[i].value;
//cout << "GLO Offset: " << prn << " " << glonass_offset << " " << gsatdata.epoch().str_hms() << endl;
    }     

  }
   
  // Galileo system time offset
  double galileo_offset = 0;
  if (gsatdata.gsys() == GAL){
    int i;
    i = param.getParam(_site, t_gpar::GAL_ISB, "");
    if ( i >= 0 ) { galileo_offset = param[i].value;
//cout << "GAL Offset: " << prn << " " << galileo_offset << " " << gsatdata.epoch().str_hms() << endl;       
    }
     
  }
   
  // BaiDou system time offset
  double beidou_offset = 0;
  if (gsatdata.gsys() == BDS){
    int i;
    i = param.getParam(_site, t_gpar::BDS_ISB, "");
    if ( i >= 0 ) {beidou_offset = param[i].value;
//cout << "BDS ISB: " << prn << " " << param[i].value << " " << gsatdata.epoch().str_hms() << endl;
    }
  }
   
  // QZSS system time offset
  double qzss_offset = 0;
  if (gsatdata.gsys() == QZS){
    int i;
    i = param.getParam(_site, t_gpar::QZS_ISB, "");
    if ( i >= 0 ) qzss_offset = param[i].value;     
  }   


#ifdef DEBUG
   cout << gsatdata.epoch().str_hms() <<  " " 
        << gsatdata.sat()  << fixed << setprecision(20)
        << "  rho: "       << gsatdata.rho()
        << "  ele: "       << setw(6) << gsatdata.ele()*180.0/PI
        << "  H:   "       << setw(6) << ell[2]     
        << "  rec clk: "   << setw(6) << clkRec
        << "  sat clk: "   << setw(6) << gsatdata.clk()
        << "  Tropdelay: " << setw(6) << trpDelay
        << "  GLO_ISB " << setw(6) << glonass_offset
        << "  GAL_ISB " << setw(6) << galileo_offset
        << "  BDS_ISB " << setw(6) << beidou_offset
        << "  QZS_ISB " << setw(6) << qzss_offset     
        << endl;
//   int ooo; cin >> ooo;
#endif          
   
   // Return value
   return gsatdata.rho() + 
          clkRec         - 
          gsatdata.clk() + 
          trpDelay       +
          glonass_offset +
          galileo_offset +
          beidou_offset  +     
          qzss_offset;
}

// Compute troposperic delay
// -----------
double t_gsppmodel::tropoDelay(t_gtime& epoch, t_gallpar& param, t_gtriple ell, t_gsatdata& satdata)
{        
   gtrace("t_gsppmodel::tropoDelay");   
      
   if (tropoModel == 0){
      if( _log ) _log->comment(0, "gppp","Tropo Model setting is not correct [" + _trpModStr + "] default used! Check config.");
      else                cerr << "gppp - Tropo Model setting is not correct [" + _trpModStr + "] default used! Check config.\n";
      
      tropoModel = new t_saast();
   }
   
   double ele = satdata.ele();
   
   double delay = 0.0;
   double zwd   = 0.0;
   
   int i;
   i = param.getParam(_site, t_gpar::TRP, "");
   if ( i >= 0 ){
      zwd = param[i].value;
   }else zwd = tropoModel->getZWD(ell, epoch);

   if ( _tropo_mf.compare("gmf") == 0 ){
      double gmfh, gmfw, dgmfh, dgmfw;
	t_gmf mf;
      mf.gmf( epoch.mjd(), ell[0], ell[1], ell[2], M_PI/2.0-ele,
              gmfh, gmfw, dgmfh, dgmfw );
      delay = gmfh * tropoModel->getZHD(ell, epoch) + gmfw * zwd;
      
		
#ifdef DEBUG
    cout << epoch.str("EPOCH: %H:%M:%S") << endl << fixed << setprecision(3);
    cout << "Ell:" << ell[0] << " " << ell[1] << " " << ell[2]
	 << " Hydrostatic part: " << tropoModel->getZHD(ell, epoch)
	 << " Wet part: "        << zwd
	 << " gmfh: " << gmfh
	 << " gmfw: " << gmfw
         << " Delay: " << delay << endl << endl;	
#endif
   }else if ( _tropo_mf.compare("cosz") == 0 ){
      delay = 1/sin(ele) * tropoModel->getZHD(ell, epoch) +1/sin(ele) * zwd;
   }
   
   return delay;
}

} // namespace
