
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <cmath>
#include <iomanip>

#include "gmodels/gtropo.h"
//#include "gmodels/gpt.h"
#include "gutils/gconst.h"
#include "gutils/gtypeconv.h"

using namespace std;

namespace gnut {   

// ---------
// Base class
// ----------
t_gtropo::t_gtropo()
 : _nwm(0)
{}
   
t_gtropo::~t_gtropo()
{}

void t_gtropo::nwm( t_gallsurf* nwm )
{ 
  _nwm = nwm;
}

// ----------
double t_gtropo::getZHD(const t_gtriple& ell, const t_gtime& epo) // ell v RADIANECH !! TREBA SJEDNOTIT
{   
  gtrace("t_gtropo::getZHD");
  
  // in future this may be significantly optimized!
  map<t_gtime, map<MET_DATA, double> > map_nwm;

  if( _nwm ){
    t_gtriple tmp( ell[0]*R2D, ell[1]*R2D, ell[2]);             // DOCASNE PREVEDU NA DEGREE
    _nwm->interpolate_point(tmp, epo, epo, &map_nwm );
  }else{
    cerr << "t_gtropo:getZHD not available data, use other model \n";
  }
   

  if( map_nwm.find(epo)      != map_nwm.end() &&
      map_nwm[epo].find(ZHD) != map_nwm[epo].end() )
  {
#ifdef DEBUG
    cout << "NWM data: " << fixed << setprecision(3)
         << " ZHD " << setw(12) << map_nwm[epo][ZHD] << endl;
#endif   
     return map_nwm[epo][ZHD];
  }
//  cout << "t_gtropo::getZHD UNKNOWN" << endl;   
  return NWM_UNKNOWN;
}

// ----------
double t_gtropo::getZWD(const t_gtriple& ell, const t_gtime& epo) // ell v RADIANECH !! TREBA SJEDNOTIT
{ 
  gtrace("t_gtropo::getZWD");

  // in future this may be significantly optimized!
  map<t_gtime, map<MET_DATA, double> > map_nwm;

  if( _nwm ){
    t_gtriple tmp( ell[0]*R2D, ell[1]*R2D, ell[2]);           // DOCASNE PREVEDU NA DEGREE
    _nwm->interpolate_point(tmp, epo, epo, &map_nwm );
  }else{
    cerr << "t_gtropo:getZWD not available data, use other model \n";
  }
   
  if( map_nwm.find(epo) != map_nwm.end() &&
      map_nwm[epo].find(ZWD) != map_nwm[epo].end() )
  {
#ifdef DEBUG
    cout << "NWM data: " << fixed << setprecision(3) 
         << " ZWD " << setw(12) << map_nwm[epo][ZWD] << endl;
#endif   
    return map_nwm[epo][ZWD];
  }
   
//  cout << "t_gtropo::getZWD UNKNOWN" << endl;
  return NWM_UNKNOWN;
}



// ---------
// Saastamoinen 1972
// ----------
double t_saast::getSTD(const double& ele, const double& height)
{    
  gtrace("t_saast::getSTD"); 
   
  double pp =  1013.25 * pow(1.0 - 2.26e-5 * height, 5.225) ;
  double TT =  18.0 - height * 0.0065 + 273.15;
  double hh =  50.0 * exp(-6.396e-4 * height);
  double ee =  hh / 100.0 * exp(-37.2465 + 0.213166*TT - 0.000256908*TT*TT);

#ifdef DEBUG
  cout << "meteo data: " << fixed << setprecision(3)
       << " pp " << setw(12) << pp
       << " TT " << setw(12) << TT
       << " hh " << setw(12) << hh
       << " ee " << setw(12) << ee << endl;
#endif
      
  double h_km = height / 1000.0;

  if (h_km < 0.0) h_km = 0.0;
  if (h_km > 5.0) h_km = 5.0;
  int    ii   = int(h_km + 1);
  double href = ii - 1;
	
  double bCor[6];
  bCor[0] = 1.156;
  bCor[1] = 1.006;
  bCor[2] = 0.874;
  bCor[3] = 0.757;
  bCor[4] = 0.654;
  bCor[5] = 0.563;

  double BB = bCor[ii-1] + (bCor[ii]-bCor[ii-1]) * (h_km - href);
	
  double zen   = PI/2.0 - ele;
  double delay = (0.002277/cos(zen)) * (pp + ((1255.0/TT)+0.05)*ee - BB*(tan(zen)*tan(zen)));
  return delay;
}

// ----------
double t_saast::getZHD(const t_gtriple& Ell, const t_gtime& epoch)
{   
  gtrace("t_saast::getZHD");
   
  double P, T, N;
  _gpt.gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P,T,N);
   
  double delay = (0.002277 * P) /
                 (1.0 - 0.00266 * cos(2.0*Ell[0]) - 0.00000028 * Ell[2]);

#ifdef DEBUG
  cout << "GPT data: " << fixed << setprecision(3)
       << " P " << setw(12) << P
       << " T " << setw(12) << T
       << " N " << setw(12) << N
       << " ZHD " << setw(12) << delay << endl;           
#endif   
   
  return delay;
}

// ----------
double t_saast::getZWD(const t_gtriple& Ell, const t_gtime& epoch)
{   
  gtrace("t_saast::getZWD");   
   
  double P, T, N;   
  _gpt.gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P, T, N);

  T += 273.15;   
   
  double hh =  50.0 * exp(-6.396e-4 * Ell[2]);
  double e =  hh / 100.0 * exp(-37.2465 + 0.213166*T - 0.000256908*T*T);   
   
  double delay = (0.0022768*e*(1255.0/T + 0.05))
               / (1.0 - 0.00266 * cos(2.0*Ell[0]) - 0.00000028 * Ell[2]);
  return delay;
}


// ---------
// Davis 1985 model (Saast + Thayer 1974)
// ----------
double t_davis::getZHD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_davis::getZHD");   
   
  double P, T, N;
  _gpt.gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P,T,N);
   
  double delay = (0.0022768 * P)
               / (1.0 - 0.0026 * cos(2.0*Ell[0]) - 0.00000028 * Ell[2]);

#ifdef DEBUG
  cout << "GPT data: " << fixed << setprecision(3)
       << " P " << setw(12) << P
       << " T " << setw(12) << T
       << " N " << setw(12) << N
       << " ZHD " << setw(12) << delay << endl;           
#endif

  return delay;
}

// ---------
double t_davis::getZWD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_davis::getZWD");   
   
  // not implemented yet
  return 0.0;
}


// ----------
// Hopfield model
// ----------
double t_hopf::getZHD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_hopf::getZHD");   
   
  double P, T, N;

  _gpt.gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P, T, N);
     
  double delay = (1e-06/5.0)*(77.64*(P/T))*(40136.0+148.72*T);
  return delay;
}

// ----------
double t_hopf::getZWD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_hopf::getZWD");   
   
  double P, T, N;
  double hh =  50.0 * exp(-6.396e-4 * Ell[2]);
  _gpt.gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P, T, N);
      
  double e =  hh / 100.0 * exp(-37.2465 + 0.213166*T - 0.000256908*T*T);

  double delay = (1.e-06/5.0)*((-12.96)*(e/T)+(3.718*1.e05)*(e/(T*T)))*11000.0;
  return delay;
}


// ---------
// Baby model
// ---------
double t_baby::getZHD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_baby::getZHD");   
   
  double P, T, N;
  _gpt.gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P, T, N);

  T += 273.15;    // [K]

  double gs = 9.81;         // [ms^2] surface gravity !!!
  double rs = A_WGS+Ell[2];
  double sigma = Eps/T;
  double mu = gs/(Rd*Eps)*(1.0-(2.0/(rs*sigma)));
      
  double delay = (0.022277*P/gs)*(1.0+(2.0/(rs*sigma*(mu+1.0))));     
  return delay;
}

// ---------
double t_baby::getZWD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_baby::getZWD");   
   
  // not implemeted yet
  return 0.0;
}


// Chao model
// --------
double t_chao::getZHD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_chao::getZHD");   
   
  // not implemeted yet
  return 0.0;
}


// Chao model (ZWD), Mendes pp. 85
// --------
double t_chao::getZWD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_chao::getZWD");   
   
  double P, T, N, alpha;
  double hh =  50.0 * exp(-6.396e-4 * Ell[2]);   
  _gpt.gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P, T, N);

  T += 273.15;
  alpha = 0.0065; // model is not very sensitive to the temperature lapse rate   
     
  double e =  hh / 100.0 * exp(-37.2465 + 0.213166*T - 0.000256908*T*T);   
      
  double delay = 4.70*100.0*(pow(e,1.23)/(T*T))+1.71*1.e6*(pow(e,1.46)/(T*T*T))*alpha;
  return delay;
}


// ---------
// Ifadis model
// ---------
double t_ifad::getZHD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_ifad::getZHD");   
   
  // not implemeted yet
  return 0.0;
}

// ---------
double t_ifad::getZWD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_ifad::getZWD");   
   
  double P, T, N;   
  double hh =  50.0 * exp(-6.396e-4 * Ell[2]);   
  _gpt.gpt_v1( epoch.mjd(), Ell[0], Ell[1], Ell[2], P, T, N);

  T += 273.15;
      
  double e =  hh / 100.0 * exp(-37.2465 + 0.213166*T - 0.000256908*T*T);   
      
  double delay = 0.00554-0.0880*1.e-4*(P - 1000.0)+0.272*1.e-4*e+2.771*(e/T);
  return delay;
}


// ---------
// Askne Nordius model
// ---------
double t_askn::getZHD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_askn::getZHD");   
   
  return 0.0;   // not implemeted yet
}

// ---------
double t_askn::getZWD(const t_gtriple& Ell, const t_gtime& epoch)
{
  gtrace("t_askn::getZWD");   
   
  return 0.0;   // not implemeted yet (WV pressure not available)
}

// ---------  BACHA TADY JE TO V degrees !!!!!
double t_askn::getZWD(const t_gtriple& Ell,
		      double  T,             // temperature [K]
		      double  E,             // water vapour [hPa]
     		      double  dT,            // T lapse rate [K/km]
		      double  dE,            // E decay parameter [-]
		      double  dW,            // ZWD decay parameter [-]
		      double  ref,           // ZWD reference value [m]
                      double& ratio,         // combi ratio of E/ZWD decays [%]
		      double  tm
		     )
{
  gtrace("t_askn::getZWD");   
   
  double lat = Ell[0];                              // lat [deg]
  double hgt = Ell[2]/1000;                         // height [km]
  double gm = 9.784/( 1 - 0.00266 * cos(2*lat*D2R) 
		        - 0.00028 * hgt);           // [m/s] Askne-Nordius original
  double TmE = T * ( 1 - dT/1000 * Rd/gm/(dE + 1)); // [K] mean temperature
//double TmW = T * ( 1 - dT/1000 * Rd/gm/(dW + 1)); // [K] mean temperature
  double Tm  = TmE;
  double dO;

  if( dT <= 0 ) dT = 6.5;                           // [K/km]
   
  double coef1 = 0.0, coef2 = 0.0, delay = 0.0, decay = 0.0;

  if( ratio == -1 && ref != 0.0 )
  {
    if( fabs(dE - dW) <= RATIO_LIM ){
       ratio = RATIO_EW;
    }else{
      coef1 = ratio/100;
      coef2 = 1.0-coef1;
      decay = coef1*dE + coef2*dW;  // if( decay < 0.0 ) decay = 0.0; !!!! TADY NE

      if( tm != NWM_UNKNOWN ){
	 Tm = tm;          // use requested Tm (argument tm)
         double dO = 1e-06 * (K2_THA - K1_THA*Eps + K3_THA/TmE) * Rd/gm/ref * E - 1;
         ratio = (dO-dW)/(dE-dW)*100.0;
         if( ratio > 100.0 ) ratio = 100.0;
         if( ratio <   0.0 ) ratio =   0.0;
      }else{
         Tm = T * ( 1 - dT/1000 * Rd/gm/(dE + 1) ); // [K] mean temperature
         dO = 1e-06 * (K2_THA - K1_THA*Eps + K3_THA/Tm) * Rd/gm/ref * E - 1;
         ratio = (dO-dW)/(dE-dW)*100.0;
         if( ratio > 100.0 ) ratio = 100.0;
         if( ratio <   0.0 ) ratio =   0.0;
       
         Tm = T * ( 1 - dT/1000 * Rd/gm/(dO + 1) ); // [K] mean temperature
         ratio = (dO-dW)/(dE-dW)*100.0;
         if( ratio > 100.0 ) ratio = 100.0;
         if( ratio <   0.0 ) ratio =   0.0;
 /*       
      double TmE = T * ( 1 - dT/1000 * Rd/gm/(dE + 1) ); // [K] mean temperature
      double TmW = T * ( 1 - dT/1000 * Rd/gm/(dW + 1) ); // [K] mean temperature
      double zwdE = 1e-06 * (K2_THA - K1_THA*Eps + K3_THA/Tm) * Rd/gm/(dE + 1) * E;  // [m]
      double zwdW = 1e-06 * (K2_THA - K1_THA*Eps + K3_THA/Tm) * Rd/gm/(dW + 1) * E;  // [m]
      cout << " test " << ratio
	   << fixed << setprecision(1)
	   << " dE   " << dE
	   << " dW   " << dW
	   << " dO   " << dO	
	   << " rat  " << ratio
	   << " dZWD " << (zwdE-zwdW)*1000
	   << " zwdE " << zwdE*1000 
	   << " zwdW " << zwdW*1000
	   << " ref  "  << ref*1000 
	   << "  tm  "  << tm
	   << "  Tm  "  << Tm
	   << endl;
 */
      }       
    }
  }
   
  coef1 = ratio/100;
  coef2 = 1.0-coef1;
  decay = coef1*dE + coef2*dW; if( decay < 0.0 ) decay = 0.0;   // ???? CHECK IT LATER ON AGAIN 
//TmE = T * ( 1 - dT/1000 * Rd/gm/(dE    + 1));   // [K] mean temperature
//TmW = T * ( 1 - dT/1000 * Rd/gm/(dW    + 1));   // [K] mean temperature
  Tm  = T * ( 1 - dT/1000 * Rd/gm/(decay + 1));   // [K] mean temperature
//Tm  = TmE;
  if( tm != NWM_UNKNOWN ){ Tm = tm;  }            // replace by argument
  delay = 1e-06 * (K2_THA - K1_THA*Eps + K3_THA/Tm) * Rd/gm/(decay + 1) * E; // [m]

#ifdef DEBUG
  cout << fixed << setprecision(3)
       << " ratio = " << ratio
       << " decay = " << decay
       << "  e    = " << E
       << "  T    = " << T
       << "  Tm   = " << Tm
       << "  zwd  = " << delay
       << "  ref  = " << ref
       << endl;
#endif

  return delay; // # [m]
}

   
// ---------  BACHA TADY JE TO V degrees !!!!!
double t_askn::getZWD_OLD(const t_gtriple& Ell,
		      double  T,             // temperature [K]
		      double  E,             // water vapour [hPa]
     		      double  dT,            // T lapse rate [K/km]
		      double  dE,            // E decay parameter [-]
		      double  dW,            // ZWD decay parameter [-]
		      double  ref,           // ZWD reference value [m]
                      double& ratio,         // combi ratio of E/ZWD decays [%]
		      double  tm
		     )
{
  gtrace("t_askn::getZWD");   
   
  double lat = Ell[0];                              // lat [deg]
  double hgt = Ell[2]/1000;                         // height [km]
  double gm = 9.784/( 1 - 0.00266 * cos(2*lat*D2R) 
		        - 0.00028 * hgt);           // [m/s] Askne-Nordius original
  double TmE = T * ( 1 - dT/1000 * Rd/gm/(dE + 1)); // [K] mean temperature
//double TmW = T * ( 1 - dT/1000 * Rd/gm/(dW + 1)); // [K] mean temperature
  double Tm  = TmE;

  if( dT <= 0 ) dT = 6.5;                           // [K/km]
   
  double coef1 = 0.0, coef2 = 0.0, delay = 0.0, decay = 0.0;

  if( ratio == -1 && ref != 0.0 )
  {
    double min_zwd = 9.999;
    double est_rat = 0.0;
    double min_rat = 0.0;
    double max_rat = 100.0;
    double inc_rat = 1.0; // 10.0
    double tmp;

    if( fabs(dE - dW) <= RATIO_LIM ){ min_rat = max_rat = RATIO_EW; }
     
    for( ratio = min_rat; ratio <= max_rat; ratio += inc_rat )
    {
      coef1 = ratio/100;
      coef2 = 1.0-coef1;
      decay = coef1*dE + coef2*dW;  // if( decay < 0.0 ) decay = 0.0; !!!! TADY NE

//    TmW = T * ( 1 - dT/1000 * Rd/gm/(dW    + 1) ); // [K] mean temperature	 
      TmE = T * ( 1 - dT/1000 * Rd/gm/(dE    + 1) ); // [K] mean temperature      
      Tm  = T * ( 1 - dT/1000 * Rd/gm/(decay + 1) ); // [K] mean temperature	 
//    Tm  = TmE;

      if( tm != NWM_UNKNOWN ){ Tm = tm;  }         // replace Tm

      tmp = 1e-06 * (K2_THA - K1_THA*Eps + K3_THA/Tm) * Rd/gm/(decay + 1) * E;  // [m]

// /*       
      double dZWD = fabs(tmp-ref);
      cout << " test " << ratio
	   << fixed << setprecision(1)
	   << " dE   " << dE
	   << " dW   " << dW
	   << " dZWD " << dZWD*1000
	   << " min  " << min_zwd*1000 
	   << " tmp  " << tmp*1000 
	   << " ref  "  << ref*1000 
	   << "  tm  "  << tm
	   << "  Tm  "  << Tm
	   << endl;
//*/
      if( fabs(tmp-ref) < min_zwd ){ 
//	 cout << "ratio = " << ratio << endl;
 	 est_rat =  ratio;
	 min_zwd =  fabs(tmp - ref);
      }
       
#ifdef DEBUG
  cout << fixed << setprecision(3)
       << " ratio = " << ratio
       << " decay = " << decay
       << "  zwd  = " << tmp
       << "  ref  = " << ref
       << "  dif  = " << fabs(tmp-ref)
       << "  min  = "  << est_rat
       << endl;
#endif
    }

    ratio = est_rat; // return estimated ratio
  }
   
  coef1 = ratio/100;
  coef2 = 1.0-coef1;
  decay = coef1*dE + coef2*dW; if( decay < 0.0 ) decay = 0.0;   // ???? CHECK IT LATER ON AGAIN 
//TmE = T * ( 1 - dT/1000 * Rd/gm/(dE    + 1));   // [K] mean temperature
//TmW = T * ( 1 - dT/1000 * Rd/gm/(dW    + 1));   // [K] mean temperature
  Tm  = T * ( 1 - dT/1000 * Rd/gm/(decay + 1));   // [K] mean temperature
//Tm  = TmE;
  if( tm != NWM_UNKNOWN ){ Tm = tm;  }            // replace by argument
  delay = 1e-06 * (K2_THA - K1_THA*Eps + K3_THA/Tm) * Rd/gm/(decay + 1) * E; // [m]

//#ifdef DEBUG
  cout << fixed << setprecision(3)
       << " ratio = " << ratio
       << " decay = " << decay
       << "  e    = " << E
       << "  T    = " << T
       << "  Tm   = " << Tm
       << "  zwd  = " << delay
       << "  ref  = " << ref
       << endl;
//#endif

  return delay; // # [m]
}

} // namespace
