/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iostream>
#include <sstream>

#include "gproc/gpppflt.h"
#include "gutils/gmatrixconv.h"

namespace gnut {

// Global variables
// -----------------------------------------------------------
const unsigned int MINOBS = 5;
// -----------------------------------------------------------

// constructor
// ----------
t_gpppflt::t_gpppflt(string mark, t_gsetbase* set)
: t_gspp(mark, set),
 t_gppp(mark, set),
  t_gsppflt(mark, set),
   _flt(0),
   _smt(0)
{
  gtrace("t_gpppflt::t_gpppflt");   

   t_gppp::_get_settings();   
   
  this->_setOut();       // set output && identify if smoothing requested

  _epoch.tsys(t_gtime::GPS);

  _grdStoModel = new t_randomwalk();
  _grdStoModel->setq(dynamic_cast<t_gsetflt*>(_set)->rndwk_grd() );      

  int ipar = _param.parNumber();   

  // Add tropospheric gradient parameters
  if( dynamic_cast<t_gsetproc*>(_set)->gradient() ){
     t_gpar par_grdN(_site, t_gpar::GRD_N, ++ipar, "");
     t_gpar par_grdE(_site, t_gpar::GRD_E, ++ipar, "");     
     par_grdN.setMF( _mf );
     par_grdE.setMF( _mf );     
     _param.addParam(par_grdN);
     _param.addParam(par_grdE);     
  }   

  // Filling init parameter covariance matrix
 SymmetricMatrix tmp = _Qx; 
 _Qx.ReSize( _param.parNumber() ); _Qx = 0.0;
 for(int i = 1; i <= tmp.Nrows(); i++) _Qx(i,i) = tmp(i,i);
 double grdInit = 0.0005;
 for (unsigned int i=1; i<=_param.parNumber(); i++) {
   if (_param[i-1].parType == t_gpar::GRD_N)
     _Qx(i,i) = grdInit * grdInit;
   else if (_param[i-1].parType == t_gpar::GRD_E)
     _Qx(i,i) = grdInit * grdInit;
 }

 _reset_amb = dynamic_cast<t_gsetflt*>(_set)->reset_amb();
 _reset_par = dynamic_cast<t_gsetflt*>(_set)->reset_par();   
   
} // end constructor


// destructor
// ----------
t_gpppflt::~t_gpppflt()
{ 
  gtrace("t_gpppflt::~t_gpppflt");   
   
  if( _grdStoModel ) delete _grdStoModel;

  if( _flt ){ if( _flt->is_open() ){ _flt->close(); }; delete _flt; }
}


// Process observation batch
// ----------
int t_gpppflt::processBatch(const t_gtime& beg, const t_gtime& end)
{

  gtrace("t_gpppflt::processBatch");

  if( beg > end ){ cerr << _site << " - processing not started [beg > end]\n"; return -1; }

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif 
  _gmutex.lock();

  _dop.set_data(_gnav, _gobs, _site);   
   
  t_gtime begT( beg );
  t_gtime endT( end );

  _running = true;
  _param.setSite(_site);
  double ZTD, ZWD, rmsZTD;
  t_gtriple xyz, neu, ell;

  double subint = pow(10,floor(log10(_sampling)));
  int sign  = fabs(_sampling)/_sampling;

  string smtModStr(dynamic_cast<t_gsetflt*>(_set)->method_smt() );
  string fltModStr(dynamic_cast<t_gsetflt*>(_set)->method_flt() );
  int smt_delay = dynamic_cast<t_gsetflt*>(_set)->smt_delay();      
     
   t_gtime now( begT );

#ifdef DEBUG
   if( _log ) _log->comment(1,"t_gpppflt", begT.str_ymdhms( "Filtering: " ) + endT.str_ymdhms(" "));
#endif

   t_gtime timeT(now); 
     
//   cout << _site << ": Start filtering: " << _site << now.str_ymd() << endl;           
   while(now < end || now == end)
   {
      // synchronization
     if(now != end){     
	if( now.sod()%_sampling != 0){
	   now.add_dsec( sign * subint/100 ); // add_dsec used for synchronization!

	   continue;
	}
	now.reset_dsec();
     }      

     // clean/collect/filter epoch data
     _data.erase( _data.begin(), _data.end() );
     _data = _gobs->obs( _site, now );                
      
     if( _data.size() == 0){
       if( _log && _log->verb() >= 1 )
	   _log->comment(1,"gppp", _site + now.str_ymdhms(" no observation found at epoch: "));

       now.add_sec( sign * _sampling );  // not yet add_dsec() ! not support >1Hz data
       continue;
     }

      t_gtime obsEpo = _data.begin()->epoch();       
     _timeUpdate(obsEpo); 

     // reseting params according to conf
     if (_reset_amb > 0) {
	if (now.sod() % _reset_amb == 0) {
	  vector<int> ind = _param.delAmb();
	  Matrix_rem(_Qx, ind);
	}	
     }
     if (_reset_par > 0) {
	if (now.sod() % _reset_par == 0) {
	  vector<int> ind = _param.delAmb();
	  Matrix_rem(_Qx, ind);	   
         _initialized = false;
	}
	
     }

     int irc_epo = _processEpoch(obsEpo);

     if( irc_epo < 0 ) {
       if( _log && _log->verb() >= 1 ) 
	   _log->comment(1,"gppp", _site + now.str_ymdhms(" epoch ") + " was not calculated" );

       now.add_sec( sign * _sampling ); // not yet add_dsec() ! not support >1Hz data
       continue;
     }      

     if( _log && _log->verb() >= 1 )
	 _log->comment(1,"gppp", _site + now.str_ymdhms(" processing epoch: "));

     if( _param.getCrdParam(_site, xyz) > 0 ){
     }else if( _valid_crd_xml() ) xyz = _aprox_crd_xml;
      
     xyz2ell(xyz, ell);
     
     int idx = _param.getParam(_site, t_gpar::TRP, "");
     if( idx >= 0 ) {
	rmsZTD = sqrt(_Qx(idx+1,idx+1));     // _param is indexed from 0, but _Qx from 1 !
	ZWD = _param[idx].value;
     }else{
	ZWD = _gModel->tropoModel->getZWD(ell, obsEpo);
	rmsZTD = 0.0;
     }

     ZTD = _gModel->tropoModel->getZHD(ell, obsEpo) + ZWD;           

     int ign = _param.getParam(_site, t_gpar::GRD_N, "");
     double grdN = 0.0;
     double rmsN = 0.0;
     if (ign >= 0) {
	grdN = _param[ign].value;
	rmsN = sqrt(_Qx(ign+1,ign+1));      // _param is indexed from 0, but _Qx from 1 !
     }      

     int ige = _param.getParam(_site, t_gpar::GRD_E, "");
     double grdE = 0.0;
     double rmsE = 0.0;
     if (ige >= 0) {
        grdE = _param[ige].value;
	rmsE = sqrt(_Qx(ige+1,ige+1));      // _param is indexed from 0, but _Qx from 1 !
     }
      
     int iclk = _param.getParam(_site, t_gpar::CLK, "");
     double clk = 0;
     if (iclk >= 0) clk = _param[iclk].value;

     double nsat = _data.size();
     double gdop = -1;
     if (_dop.calculate(now, xyz) >=0 ) gdop = _dop.gdop();
      
     if (_allprod != 0){
       shared_ptr<t_gprodcrd> prdcrd = make_shared<t_gprodcrd>(now); // now is rounded value !!
       prdcrd->xyz(xyz);
       _allprod->add(prdcrd, _site);
	
       shared_ptr<t_gprodtrp> prdtrp = make_shared<t_gprodtrp>(now); // now is rounded value !!
       prdtrp->set_val("GDOP",gdop);                                   // add gdop
       prdtrp->set_val("NSAT",nsat);                                   // add nsat
       prdtrp->ztd( ZTD, rmsZTD );
       prdtrp->grd( grdN, grdE, rmsN, rmsE);
       _allprod->add(prdtrp, _site);
     }

     ostringstream os;
     os       << fixed << setprecision(4)
	      << now.str_ymdhms()
	      << " "   << setw(13) << xyz[0]
              << " "   << setw(13) << xyz[1] 
	      << " "   << setw(13) << xyz[2]
	      << " "   << setw(7)  << ZTD
              << " "   << setw(7)  << grdN*1000     // [mm]
              << " "   << setw(7)  << grdE*1000     // [mm]
              << " "   << setw(15) << clk           // [m]	
	      << fixed << setprecision(5)
	      << " "   << setw(13) << sqrt(_Qx(1,1))
	      << " "   << setw(9)  << sqrt(_Qx(2,2)) 
	      << " "   << setw(9)  << sqrt(_Qx(3,3))
	      << " "   << setw(7)  << rmsZTD
	      << fixed << setprecision(0)
              << " "   << setw(2)  << nsat
              << " "   << setw(3)  << gdop
              << endl;

     if( _flt ){
       _flt->write(os.str().c_str(), os.str().size());
       _flt->flush();
     }

     _initialized = true;

     now.add_sec( sign * _sampling ); // not yet add_dsec() ! not support >1Hz data
      
     // RTS Smooth
     if ( _smooth ){
        if (now - timeT >= smt_delay && smt_delay != 0){
           _backflt();
           _vfltdata.clear();
	   timeT = now;
        }      
     }         
   
   }   
   
  _running = false;  
   
  if (_smooth && smt_delay == 0){
     _backflt();
  }

  _gmutex.unlock();

  return 1;
}

// ----------------------------------------------------------------------------------
// PROTECTED
// ----------------------------------------------------------------------------------
// 

// Update time for RDW processes
// ---------------------------
void t_gpppflt::_timeUpdate(const t_gtime& epo)
{
   t_gsppflt::_timeUpdate(epo);
   
   //tropo gradient
   _grdStoModel->updateTime(epo);
   _grdStoModel->setTcurr(epo);
}                   


// Process one epoch
// -----------------------------
int t_gpppflt::_processEpoch(const t_gtime& runEpoch)
{
   gtrace("t_gpppflt::_processEpoch");

   int irc = t_gsppflt::_processEpoch(runEpoch);
   if(irc < 0) return irc;
   
   return 1;
}


// add data to A, l, P - carrier phase
// ---------------------------
int t_gpppflt::_addObsL(t_gsatdata& satdata, unsigned int& iobs, t_gtriple& ell, Matrix& A, ColumnVector& l, DiagonalMatrix& P)
{
   t_gsys gsys(satdata.gsys());
   GSYS gs = gsys.gsys();
// int b1 = gsys.first_band(gs,1);
// int b2 = gsys.second_band(gs,b1);
   GOBSBAND b1 = BAND_1;
   GOBSBAND b2 = BAND_2;
      
   if( _phase ){	 		
                  
      bool com = true;
      if( _gnav ) com = _gnav->com();
      
      // create reduced measurement, obs. weight matrix for L3
      double L3 = satdata.L3(b1,b2);
      if(double_eq(L3, 0.0)) return -1;
      
      double weight_coef = _weightObs(satdata);
      
      if (gs == GPS)
	P(iobs,iobs) =  1 / ( (_sigPhaseGPS * _sigPhaseGPS) / weight_coef);
      if (gs == GLO)
	P(iobs,iobs) =  1 / ( (_sigPhaseGLO * _sigPhaseGLO) / weight_coef);
      if (gs == GAL)
	P(iobs,iobs) =  1 / ( (_sigPhaseGAL * _sigPhaseGAL) / weight_coef);
      if (gs == BDS)
	P(iobs,iobs) =  1 / ( (_sigPhaseBDS * _sigPhaseBDS) / weight_coef);
      if (gs == QZS)
	P(iobs,iobs) =  1 / ( (_sigPhaseQZS * _sigPhaseQZS) / weight_coef);      
      
      for( unsigned int iPar=0; iPar<=_param.parNumber()-1; iPar++ ){

	 if( _param[iPar].parType == t_gpar::AMB &&
	     _param[iPar].prn.compare( satdata.sat() ) == 0 ) {
	    
            double modObsL = _gModel->cmpObs(_epoch, _param, satdata, true, com);

	    // Check lli for cycle slip
//          GOBS type1 = satdata.pha_id(b1);
//          GOBS type2 = satdata.pha_id(b2);
            GOBS type1 = satdata.id_phase(b1);
            GOBS type2 = satdata.id_phase(b2);

	    if ( satdata.getlli(type1) >= 1 || satdata.getlli(type2) >= 1 )
	      {
		 _param[iPar].value = L3 - modObsL;
		 _Qx(iPar+1, iPar+1) =  _sigAmbig * _sigAmbig;
// cout << "Cycle slip: " << satdata.sat() << " " << gobs2str(type1) << " " << gobs2str(type2) << " " << _epoch.str_hms() << " " << _param[iPar].value << endl;		 
		 // add info about amb reset for smoothing
		 if (_smooth){
		    SymmetricMatrix Qp = _fltdata.Qp();
		    Qp(iPar+1, iPar+1) = _sigAmbig * _sigAmbig;
		    _fltdata.Qp(Qp);
		 }		  
	      }
	    	    
	    // reduced measurement for L3
	    l(iobs) = L3 - _param[iPar].value - modObsL;
	    
#ifdef DEBUG	     
	    cout << satdata.sat()    << fixed << setprecision(5)
	      << "  epoch"    << _epoch.str_ymdhms()
	      << "  lamb: "   << setw(12) << satdata.wavelength(BAND_1)
	      << "  freq: "   << setw(12) << satdata.frequency(BAND_1)
	      << "  L3: "     << setw(12) << L3
	      << "  amb: "    << setw(12) << _param[iPar].value
	      << "  cmpObs: " << setw(12) << modObsL
	      << endl;
#endif
	 }
      }
      
      // create first design matrix 
      for( unsigned int ipar=1; ipar<=_param.parNumber(); ipar++ ){
	 A(iobs, ipar) = _param[ipar-1].partial( satdata, _epoch, ell, true);
      }      
      
   iobs++; // count obs
   }

   return 1;
}


// Applying tides
// -------------------------------------------------------
int t_gpppflt::_apply_tides(t_gtime& _epoch, t_gtriple& xRec)
{
   gtrace("t_gpppflt::_apply_tides");

    if( _tides ){
      xRec = xRec + _tides->tide_searth(_epoch, xRec)
                  + _tides->load_ocean(_epoch, _site)
     	          + _tides->tide_pole()
	          + _tides->load_atmosph();
    } else return -1;
//cout << fixed << setprecision(10) << "xRec = " << xRec[0] << " " << xRec[1] << " " << xRec[2] << endl;   
   return 1;
}

// Predict state vector and covariance matrix
// =======================================================
void t_gpppflt::_predict()
{
   gtrace("t_gpppflt::_predict");

  // Predict coordinates, clock and troposphere
  // state vector is the same, covariance matrix is predicted with noise   
   int i;      
   
   t_gsppflt::_predict();   
   
   double grdInit       = 0.0005;   

   i = _param.getParam(_site, t_gpar::GRD_N, "");
   if ( i >= 0 )  {
     if (!_initialized ) { 
	_Qx(i+1,i+1) = grdInit * grdInit;
     } else if (_cntrep == 1 ) {
         _Qx(i+1,i+1) += _grdStoModel->getQ();
         if(_smooth) _Noise(i+1,i+1)  = _grdStoModel->getQ();
     }      
   }

   i = _param.getParam(_site, t_gpar::GRD_E, "");
   if ( i >= 0 )  {
     if (!_initialized ) {
	_Qx(i+1,i+1) = grdInit * grdInit;
     } else if (_cntrep == 1 ) {
        _Qx(i+1,i+1) += _grdStoModel->getQ();
        if(_smooth) _Noise(i+1,i+1)  = _grdStoModel->getQ();
     }      
   }
   
  // Predict ambiguity 
  // Add or remove ambiguity parameter and appropriate rows/columns covar. matrix
  if( _phase )  _syncAmb();
   
//   cout << "Predicted Q: \n" << _Qx << endl;
//   cout << "Predicted X: \n" << _param << endl;
   
   // Create _Noise matrix (DiagonalMatrix)
   if( _smooth ){
      Matrix tmp = _Noise;
      _Noise.ReSize(_Qx.Nrows()); _Noise = 0;
      for (int i = 1; i <= _Noise.Ncols(); i++){
	 if (i > tmp.Ncols()) break;
	 _Noise(i,i) = tmp(i,i);
      }
   }
   
} // end prediction


// Satelite position
// ----------
int t_gpppflt::_satPos(t_gtime& epo, t_gsatdata& gsatdata)
{
   gtrace("t_gpppflt::_satPos");      

   if( t_gsppflt::_satPos(epo, gsatdata) < 0 ) return -1;

   string satname  =  gsatdata.sat();
   
   // phase center offset correction
   // phase center variations are applied in gpppmodel.cpp
   if( _gobj != 0 ){
      t_gtriple crd           =  gsatdata.satcrd();

      shared_ptr<t_gobj> nullobj;
      shared_ptr<t_gobj> obj  =  _gobj->obj( satname );

      shared_ptr<t_gpcv> nullapc;
      shared_ptr<t_gpcv> apc;
      if ( obj != nullobj ) apc =  obj->pcv( epo );
      
      if ( apc != nullapc ) {
	t_gtriple dx(0,0,0);
	if( _gnav && _gnav->com() ){
	  apc->pcoS(gsatdata, dx, "L3");
          for (int i=0; i<3; i++) crd[i] += dx[i];
	}
	gsatdata.addcrd(crd);
	 
      }else{
	 t_gpcv apc_def;
	 t_gtriple dx(0,0,0);
	 if( _gnav && _gnav->com() ){
	    apc_def.pcoS(gsatdata, dx, "L3");
	    for (int i=0; i<3; i++) crd[i] += dx[i];
	 }
	 gsatdata.addcrd(crd);	 
      }
   }
   
   return 1;
}


// Synchronize ambiguity with respect to observation data.
// First new ambiguity parameters are added.
// Second old ambiguity parameters are deleted;
// -----------------------------------------------------
void t_gpppflt::_syncAmb()
{
  gtrace("t_gpppflt::_syncAmb");
   
  // Add ambiguity parameter and appropriate rows/columns covar. matrix   
  set<string> mapPRN; 
   
  vector<t_gsatdata>::iterator it;
  for( it = _data.begin(); it != _data.end(); it++){       // loop over all observations
    mapPRN.insert( it->sat() );

    t_gsys gsys(it->gsys());
//    GSYS gs = gsys.gsys();
//    int b1 = gsys.first_band(gs,1);
//    int b2 = gsys.second_band(gs,b1);     
    GOBSBAND b1 = BAND_1;
    GOBSBAND b2 = BAND_2;
     
    if ( _param.getParam(_site, t_gpar::AMB, it->sat()) < 0 ) {
       double L3 = it->L3(b1,b2);
       if(double_eq(L3, 0.0)) continue;
       
       bool com = true;
       if( _gnav ) com = _gnav->com();
       
       t_gpar newPar(it->site(), t_gpar::AMB, _param.parNumber()+1, it->sat());
       newPar.value = L3 - _gModel->cmpObs(_epoch, _param, *it, true, com);      // first ambiguity value       
       
      _param.addParam(newPar);

      Matrix_addRC(_Qx, _param.parNumber(), _param.parNumber());
       double sigAmb = _sigAmbig;

      _Qx(_param.parNumber(), _param.parNumber()) = sigAmb * sigAmb;

      if( _log )
          _log->comment(3, "gppp", "AMB was added! For Sat PRN " + it->sat() + " Epoch: " + _epoch.str_ymdhms());
	   
#ifdef DEBUG
       cout << "AMB was added! For Sat PRN " << it->sat() << " Epoch: "
            << _epoch.str_ymdhms() << endl;
#endif
      
    }
  }  // end loop over all observations

// Remove ambiguity parameter and appropriate rows/columns covar. matrix

   for (unsigned int i=0; i<=_param.parNumber()-1; i++)
     {
	if (_param[i].parType == t_gpar::AMB) {

	   set<string>::iterator prnITER = mapPRN.find(_param[i].prn);
	   if (prnITER == mapPRN.end()) {

      if( _log )
          _log->comment(3, "gppp", "AMB will be removed! For Sat PRN " + _param[i].prn + " Epoch: " + _epoch.str_ymdhms());
		  
#ifdef DEBUG
   cout << "AMB will be removed! For Sat PRN " << _param[i].prn
        << " Epoch: " << _epoch.str_ymdhms() << endl;
#endif	      

      	      Matrix_remRC(_Qx, _param[i].index, _param[i].index);
	      _param.delParam(i);
	      _param.reIndex();
	   }	  
	}	
     }

}


// Set output & return if smooth
// -----------------------------
void t_gpppflt::_setOut()
{
  string tmp;
  tmp = dynamic_cast<t_gsetout*>(_set)->outputs("flt");
  if( ! tmp.empty() ){
    substitute(tmp,"$(rec)",_site, false);
    _flt = new t_giof;
    _flt->tsys(t_gtime::GPS);
    _flt->mask(tmp);
    _flt->append( dynamic_cast<t_gsetout*>(_set)->append() );
  }

}

// Run smoothing
// ---------------------
void t_gpppflt::_backflt()
{
  gtrace("t_gpppflt::_backflt");
  return;
}

} // namespace
