
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iomanip>
#include <memory>

#include "gproc/gpreproc.h"
#include "gutils/gtypeconv.h"
#include "gmodels/gbancroft.h"

using namespace gnut;

namespace gnut {

// Constructor
// ----------
t_gpreproc::t_gpreproc(t_gallobs* obs, t_gsetbase* settings)
: t_gmonit("gpreproc"),
  _log(0)
{
   _obs = obs;
   _set = settings;
   
   _msoffset.clear();

   _sys = dynamic_cast<t_gsetgen*>(_set)->sys();
   _sat = dynamic_cast<t_gsetgnss*>(_set)->sat();
   _sigCode      = dynamic_cast<t_gsetproc*>(_set)->sigma_C_GPS();
   _sigPhase     = dynamic_cast<t_gsetproc*>(_set)->sigma_L_GPS();
   _sigCode_GLO  = dynamic_cast<t_gsetproc*>(_set)->sigma_C_GLO();
   _sigPhase_GLO = dynamic_cast<t_gsetproc*>(_set)->sigma_L_GLO();   
   
   for (set<string>::iterator it = _sys.begin(); it != _sys.end(); it++){
      _dI[*it] = 0.0;
   }

   
//   if( _set ){ // AVOID gsetproc !
//     _sigCode  = dynamic_cast<t_gsetproc*>(_set)->sigma_C();
//     _sigPhase = dynamic_cast<t_gsetproc*>(_set)->sigma_L();
//   }
}

// Destructor
// ----------
t_gpreproc::~t_gpreproc()
{ 
}


// run preprocessing from time beg until time end (main method)
// --------------------------------------------------------------
int t_gpreproc::preprocess(string site, const t_gtime& beg, const t_gtime& end, int sampl, bool save)
{
   this->_site = site;
//   cout << "start prep for " << _site << endl;
   if ( _obs->nepochs(_site) <= 1) return -1;   

   bool firstEpo = true;
   shared_ptr<t_gobsgnss> obsnull;
   vector<shared_ptr<t_gobsgnss>> epoData;
   vector<t_gobsgnss> epoDataPre;

   double CJ = 0.0;   // integer ms clk jump   
   
   double subint = pow(10,floor(log10(sampl)));   
   
   t_gtime epoch(beg); 
   while( epoch < end || epoch == end )
   {      
//     cout << epoch.str_ymdhms() << endl;
     // synchronization     
     if( int(round(epoch.sod()+epoch.dsec()))%int(sampl) != 0){
        epoch.add_dsec( subint/100 ); // add_dsec used for synchronization!
        continue;
     }
     epoch.reset_dsec();      

     epoData = _obs->obs_pt(_site, epoch);
	
     if (epoData.size() == 0) {
        epoch.add_sec( sampl ); // not yet add_dsec() ! not support >1Hz data
	continue;
     }
                  
     if (firstEpo) {
       for(unsigned int i = 0; i < epoData.size(); i++) epoDataPre.push_back(*epoData[i]);
       firstEpo = false;
     }	     

     // test data gap
     if(save){
	shared_ptr<t_gobsgnss> g = *(epoData.begin());
	double diffEpo = g->epoch() - epoDataPre.begin()->epoch();
	if(!firstEpo && diffEpo > sampl + DIFF_SEC ){
	   _gapReport(epoData);
	}
	_compare(epoDataPre, epoData);     
     }      
      
     double Ns = epoData.size();   // total number of viewed satellites
     double Na = Ns;                // number of satellite involved into clk jump detection
     double n  = 0;                 // number of detected jumps in particular epoch
     _sumS = 0;
      
     shared_ptr<t_gobsgnss> obs2;
      
     for (vector<shared_ptr<t_gobsgnss>>::iterator it = epoData.begin(); it != epoData.end(); ++it) {
	if (*it == obsnull) {	
//	   cout << "*it = 0" << endl;
	   Na--; continue;
	}	
		
	obs2 = *it;
	t_gobsgnss* obs1 = 0;	
	
	vector<t_gobsgnss>::iterator itPre;
	for (itPre = epoDataPre.begin(); itPre != epoDataPre.end(); ++itPre){
	   if ( (*it)->sat() == (itPre)->sat() ) {
	      obs1 = &(*itPre);
	   }	   
	}

	if (obs1 == 0) {
//	   cout << "obs1 = 0" << endl;
	   Na--; continue;
	} // sat not found in previous epoch
	
	t_gsys gsys(obs1->gsys());	
	if (gsys.gsys() != GPS) Na--;   // clk jump is determined only from GPS
       

        string satname  =  obs1->sat();
        if ( _sat.find(satname) == _sat.end() )  continue;  // sat in exclude in config	

	// Cycle slip detection and repair
	if( _slip(obs1, obs2) > -1 ) _transform(obs2, save);

	// Clock jump detection and repair
	int irc_jmp = _jumps(obs1, obs2);
	if ( irc_jmp ==  1 ) n++;
	if ( irc_jmp == -1 ) Na--;	
     }      // end sats loop

      epoDataPre.clear();
      
      for(unsigned int i = 0; i < epoData.size(); i++) epoDataPre.push_back(*epoData[i]);

      if (n == Na) {                                         // clk jump effect at all sats in the same way
	 double M = (1e3 * _sumS) / (CLIGHT * n);
	 double Js = abs( M-floor(M+0.5) );

	 if (Js <= 1e-5) {                                   // it is clock jump instead of common slip
	    CJ += floor(M+0.5);                               // ms jump
	    if (save) _mbreaks[epoch] = CJ;                             // store for logging
	    _remove_slip(epoData);	   
	 }
      }
      
      if( CJ != 0.0 ){
        _repair(epoData, 1e-3*CJ*CLIGHT);                      // corrected carrier phase
//        cout << "Opravil jsem epoData: " << epoch.str_hms() << " o hodnotu " << CJ << " ms" << endl;
      }
      
      epoch.add_sec( sampl ); // not yet add_dsec() ! not support >1Hz data
   } 
   
//   cout << "Preprocess for " << site << " finished" << endl;
   return 1;   
}

// Compute threshold for cycle slip
// -------------------------------------
double t_gpreproc::_slipThreshold(string LC, t_spt_gobs obs, t_gband& band1, t_gband& band2, double sampl)
{  
  double iono = 0.4/60;    // maximum iono jump/sec  ??? set via XML ???

  double f1 = obs->frequency( band1.band() ); // first  band frequency
  double f2 = obs->frequency( band2.band() ); // second band frequency
  double k1 = f1/(f1-f2);
  double k2 = f2/(f1-f2);
  double k3 = f1/(f1+f2);
  double k4 = f2/(f1+f2);

//  double alfa = (f1*f1)/(f1*f1 - f2*f2);
  
#ifdef DEBUG
   cout << "sigma phase = " << _sigPhase << endl;
   cout << "iono*sample = " << iono*sampl << endl;
   cout << "sample = " << sampl << endl;

   cout << " L4 = " << 4*sqrt(4*_sigPhase*_sigPhase) + iono*sampl 
        << " MW = " << 4*sqrt(2*k1*k1*_sigPhase*_sigPhase +
		   2*k2*k2*_sigPhase*_sigPhase +
		   2*k3*k3*_sigCode *_sigCode  +
		   2*k4*k4*_sigCode *_sigCode); 
#endif   

  if (LC.compare("L4") == 0) {
     return 4*sqrt(4*_sigPhase*_sigPhase) + iono*sampl;
  } else if (LC.compare("MW") == 0) {
     double bias = 0.05;
     double mltp = 0.5;
     return 4*sqrt( 2*k1*k1*_sigPhase*_sigPhase +
		    2*k2*k2*_sigPhase*_sigPhase +
		    2*k3*k3*_sigCode *_sigCode  +
		    2*k4*k4*_sigCode *_sigCode  + bias + mltp);
  } else if (LC.compare("WL") == 0) {
     double bias = 0.05;
     return 4*sqrt( 2*k1*k1*_sigPhase*_sigPhase +
		    2*k2*k2*_sigPhase*_sigPhase + bias);
  } else return 0.0;      
}
	  
	  
// check coherency between range and phase caused by clock jump
// --------------------------------------------------------------
int t_gpreproc::_jumps(t_gobsgnss* gobs1, t_spt_gobs gobs2)
{              
   string prn = gobs2->sat();
   GSYS GS = gobs1->gsys();
   
   if( GS != GPS ) return -1;  
   
   set<int> freq_1 = gobs1->freq_avail();
   set<int> freq_2 = gobs2->freq_avail();   
   
   if(freq_1.size() < 2 || freq_2.size() < 2) {return -1;}
   
   GOBSBAND band;
   set<int>::reverse_iterator it = freq_1.rbegin();
   // int b1_1 = t_gsys::freq2band(gobs1->gsys(), *it);
   band = t_gsys::freq2band(gobs1->gsys(), *it); 
   t_gband b1_1( band, ATTR);
   it++; 
   // int b2_1 = t_gsys::freq2band(gobs1->gsys(), *it);
   band = t_gsys::freq2band(gobs1->gsys(), *it);
   t_gband b2_1( band, ATTR);
   
   it = freq_2.rbegin();
   // int b1_2 = t_gsys::freq2band(gobs2->gsys(), *it);
   band = t_gsys::freq2band(gobs2->gsys(), *it);
   t_gband b1_2( band, ATTR);
   it++; 
   // int b2_2 = t_gsys::freq2band(gobs2->gsys(), *it);
   band = t_gsys::freq2band(gobs2->gsys(), *it);
   t_gband b2_2( band, ATTR);


  // j-epoch
  double L1j = gobs2->obs_L(b1_2); // [m]
  double L2j = gobs2->obs_L(b2_2); // [m]
  double P1j = gobs2->obs_C(b1_2); // [m]
  double P2j = gobs2->obs_C(b2_2); // [m] 

  // i-epoch
  double L1i = gobs1->obs_L(b1_1); // [m]
  double L2i = gobs1->obs_L(b2_1); // [m]
  double P1i = gobs1->obs_C(b1_1); // [m]
  double P2i = gobs1->obs_C(b2_1); // [m]

  int    nl = 0;
  double dl = 0.0;
  if (!double_eq(L1j,0.0) && !double_eq(L1i,0.0)){ dl += L1j - L1i; nl++; }
  if (!double_eq(L2j,0.0) && !double_eq(L2i,0.0)){ dl += L2j - L2i; nl++; }
  if( nl == 0 ) return -1;
  dl /= nl;

  int    np = 0;
  double dp = 0.0;
  if (!double_eq(P1j,0.0) && !double_eq(P1i,0.0)){ dp += P1j - P1i; np++; }
  if (!double_eq(P2j,0.0) && !double_eq(P2i,0.0)){ dp += P2j - P2i; np++; }
  if( np == 0 ) return -1;
  dp /= np;

  double S = dp - dl;              // jump detection observable
  double sig = 5;                  // sigma of S
  double k = 1e-3 * CLIGHT - 3*sig; // jump threshold

#ifdef DEBUG
   cout << gobs2->epoch().str_hms() << " " << gobs2->sat() << " S: " << S << " dp: " << dp << " dl: " << dl << " k: " << k << endl;
#endif
   
  if (abs(S) >= k) {   // candidate of clock jump
    _sumS += S; 
    return 1;   
  }
   return 0;
}
 
// Set navigation file (rinexn or sp3)
// --------------------------
void t_gpreproc::setNav(t_gallnav* nav)
{
   this->_nav = nav;
}

// Repair phase due to clk jump
// -----------------------------------
void t_gpreproc::_repair(vector<t_spt_gobs> epoData, double dL)
{

  for (vector<t_spt_gobs>::iterator it = epoData.begin(); it != epoData.end(); it++)
  {
     (*it)->mod_L(dL, X);  // modify all phases by dL [m]
  }
   
}

// Set site name
// -------------------
void t_gpreproc::setSite(string site)
{
   this->_site = site;
}

// Get map with cysle slips
// ---------------------------------
map<t_gtime, map<string, map<GOBS, double> > > t_gpreproc::getSlips()
{
   return _mslips;
}
   
// Get map with cysle slips : gap version
// ---------------------------------
map<t_gtime, map<string, map<GOBS, int> > > t_gpreproc::getSlipsGap()
{
   return _mslipsGap;
}

// Get map with clk jumps
// ---------------------------------
map<t_gtime, int> t_gpreproc::getClkjump()
{
   return _mbreaks;
}

// Transform dN to slips on individual band
// ----------------------------------
int t_gpreproc::_transform(t_spt_gobs gobs,  bool save)
{
   int d = _v_lcslp.size();     
 
   unsigned int maxit = 1;
   for(t_vec_slp::iterator it = _v_lcslp.begin(); it != _v_lcslp.end(); it++){
      if(it->second.size() > maxit) maxit = it->second.size();
   }
     
   for(unsigned int tmp = 1; tmp <= maxit; tmp++){
      Matrix M(d,d); M = 0;
      ColumnVector S(d); S = 0;
      
      map<GOBS, double> m_orig;      
   
      bool trans = false;

      for(int i = 1; i <= d; i++){
	 // find narr-lane slp
	 if(_v_lcslp.find(i+10) != _v_lcslp.end()) {
	    S(d) = - _v_lcslp[i+10].begin()->val;
	    if( !double_eq(S(d), 0.0) ) trans = true;	 
	    m_orig[_v_lcslp[i+10].begin()->obs1.gobs()] = 0;
	    m_orig[_v_lcslp[i+10].begin()->obs2.gobs()] = 0;

	    if(_v_lcslp[i+10].size() >= 2) {
	       _v_lcslp[i+10].erase(_v_lcslp[i+10].begin());
	    }
	    M(d,1) = 2;
	    M(d,2) = -1;
	 }
	 
	 // find wide-lane slps
	 if(_v_lcslp.find(i) != _v_lcslp.end()){

	    M(i,d) = -1;
	    M(i,d-i) = 1;

	    S(i)    = - _v_lcslp[i].begin()->val;

	    m_orig[_v_lcslp[i].begin()->obs1.gobs()] = 0;
	    m_orig[_v_lcslp[i].begin()->obs2.gobs()] = 0;
	    if(_v_lcslp[i].size() >= 2) {
	       _v_lcslp[i].erase(_v_lcslp[i].begin());
	    }

	    if( !double_eq(S(d-i), 0.0) ) trans = true;

	 }
      }      

//cout << "M = " << M << endl;
//cout << "S = " << S << endl;
      
      ColumnVector O(S); O = 0;
      
      if(trans){
	 O = M.i()*S;
	 set<int> frqs;
	 for(map<GOBS, double>::iterator it = m_orig.begin(); it != m_orig.end(); it++){
	    t_gobs s;
	    s.gobs(it->first);
	    int f = t_gsys::band2freq(gobs->gsys(), s.band());
	    frqs.insert(f);
	 }

	 int i = 1;
	 for(set<int>::iterator it = frqs.begin(); it != frqs.end(); it++, i++){
	    int b = t_gsys::freq2band(gobs->gsys(), *it);
	    for(map<GOBS, double>::iterator it = m_orig.begin(); it != m_orig.end(); it++){
	       t_gobs s;
	       s.gobs(it->first);
	       if(s.band() == b) {
		  it->second = O(i);
		  // save lli to t_gobsgnss
		  gobs->addlli( it->first, 1 );
		  if (_log ) _log->comment(2, "gpreproc", _site + ": LLI set to 1 for " + gobs->sat() + " "
					   + gobs2str( it->first ) + " "
					   +  gobs->epoch().str_hms() );		  
	       }
	       
	    }	    
	 }	 
	 
	 if (save) _save(gobs, m_orig);	 
      }

   }
   
   return 1;   
}


// Save estimated cycle slips
// ----------------------------
void t_gpreproc::_save(t_spt_gobs gobs, const map<GOBS,double>& slips)
{
   t_gtime epo = gobs->epoch();
   string  prn = gobs->sat();   
  
   for(map<GOBS, double>::const_iterator it = slips.begin(); it != slips.end(); it++){
      GOBS g  = it->first;
      double slp = it->second;
//      cout << "preproc test: " << gobs->epoch().str_hms() << " " << gobs->sat() << " " << gobs2str(g) << "  " << slp << endl;
      if( !double_eq(slp, 0.0) ) _mslips[epo][prn][g] = slp;      
   }      
}

// remove slips from gobsgnss* and _mslips
// ---------------------------------------
void t_gpreproc::_remove_slip(vector<t_spt_gobs> gobs)
{
   t_gtime epo = gobs[0]->epoch();

   map<t_gtime, map<string, map<GOBS, double> > >::iterator itEPO;
   itEPO = _mslips.find(epo);  
   if (itEPO != _mslips.end()) _mslips.erase(itEPO);
   
   for(vector<t_spt_gobs>::iterator itG = gobs.begin(); itG != gobs.end(); itG++){
      set<int> freq = (*itG)->freq_avail();      
      for(set<int>::iterator itF = freq.begin(); itF != freq.end(); itF++){
//	 int band = t_gsys::freq2band( (*itG)->gsys(), *itF);
	 GOBSBAND band = t_gsys::freq2band( (*itG)->gsys(), *itF);
	 t_gband( band, ATTR );
//	 GOBS g = (*itG)->pha_id(band);
	 GOBS g = (*itG)->id_phase(band);
	 (*itG)->addlli( g, 0 );	 
      }      
   }
}

// only common items remain
// -----------------------------------
void t_gpreproc::_common(set<int>& set1, set<int>& set2)
{
   for(set<int>::iterator it1 = set1.begin(); it1 != set1.end(); ){
      set<int>::iterator it2 = set2.find(*it1);
      if(it2 == set2.end()) {
	 set<int>::iterator tmp = it1;
	 ++it1;
	 set1.erase(tmp);
      }else ++it1;
   }
   
   for(set<int>::iterator it1 = set2.begin(); it1 != set2.end(); ){
      set<int>::iterator it2 = set1.find(*it1);
      if(it2 == set1.end()) {
	 set<int>::iterator tmp = it1;
	 ++it1;
	 set2.erase(tmp);
      }else ++it1;
   }   
}

// only common items remain
// -----------------------------------
void t_gpreproc::_common(set<GOBS>& set1, set<GOBS>& set2)
{   
   for(set<GOBS>::iterator it1 = set1.begin(); it1 != set1.end(); ){
      set<GOBS>::iterator it2 = set2.find(*it1);
      if(it2 == set2.end()) {
	 set<GOBS>::iterator tmp = it1;
	 ++it1;
	 set1.erase(tmp);
      }else ++it1;
   }
   
   for(set<GOBS>::iterator it1 = set2.begin(); it1 != set2.end(); ){
      set<GOBS>::iterator it2 = set1.find(*it1);
      if(it2 == set1.end()) {
	 set<GOBS>::iterator tmp = it1;
	 ++it1;
	 set2.erase(tmp);
      }else ++it1;
   }  
   
}

// check phase cycl slips
// --------------------------------
int t_gpreproc::_slip(t_gobsgnss* gobs1, t_spt_gobs gobs2)
{
   bool slip = false;

   set<int> frqs_t1 = gobs1->freq_avail();
   set<int> frqs_t2 = gobs2->freq_avail();   
   
   this->_common(frqs_t1, frqs_t2);      
   
   if(frqs_t1.size() != frqs_t2.size()) {
      return -1; 
      cerr << "ERROR: problem in t_gpreprocc::_common" << endl;
   }   
   
   if(frqs_t1.size() <= 1){
      if( _log ) _log->comment(2,"t_gpreproc", gobs1->epoch().str_ymdhms( "Not enough bands available: " ) );
      return -1;
   }

   int nfreq = frqs_t1.size();

   t_gobs s1;
   t_gobs s2;
   t_gobs sF;
   t_gobs s_narr;

   set<int>::reverse_iterator itFRQ = frqs_t1.rbegin();
   int b = t_gsys::freq2band(gobs1->gsys(), *itFRQ);
   set<GOBS> gf2_t1 = gobs1->obs_phase(b);
   set<GOBS> gf2_t2 = gobs2->obs_phase(b);
   this->_common(gf2_t1, gf2_t2);     // signals for wide-lane
   
   //second freq: for narrow-lane
   set<int>::iterator it_narr = frqs_t1.begin();
   if(nfreq >= 2) it_narr++;
   int b_narr = t_gsys::freq2band(gobs1->gsys(), *it_narr);
   set<GOBS> gNL_t1 = gobs1->obs_phase(b_narr);
   set<GOBS> gNL_t2 = gobs2->obs_phase(b_narr);   
   this->_common(gNL_t1, gNL_t2);     // signals for wide-lane         

   _m_lcslp.clear();
   _v_lcslp.clear();   

   for(int i = 1; i < nfreq; i++){
      ++itFRQ;
      b = t_gsys::freq2band(gobs1->gsys(), *itFRQ);
      set<GOBS> gf1_t1 = gobs1->obs_phase(b);
      set<GOBS> gf1_t2 = gobs2->obs_phase(b);            
      this->_common(gf1_t1, gf1_t2);     // reference signal - last one

      //shift reference band if no common signals
      if(gf2_t1.size() == 0){
      	 int b = t_gsys::freq2band(gobs1->gsys(), *itFRQ);
      	 gf2_t1 = gobs1->obs_phase(b);
      	 gf2_t2 = gobs2->obs_phase(b);	 
      	 this->_common(gf2_t1, gf2_t2);     // signals for wide-lane
      	 i--; nfreq--; continue;
      }
      
      // find freq for extra-wide-lane
      set<int>::reverse_iterator itFF;
      set<GOBS> gFF_t1;
      set<GOBS> gFF_t2;
      if(i>1){
	 set<int>::reverse_iterator itFF = itFRQ; --itFF;
	 b = t_gsys::freq2band(gobs1->gsys(), *itFF);
	 gFF_t1 = gobs1->obs_phase(b);
	 gFF_t2 = gobs2->obs_phase(b);	 
	 this->_common(gFF_t1, gFF_t2);     // freq for extra-widelane
      }      
      
      set<GOBS>::iterator itGOBSFF = gFF_t1.begin();
      set<GOBS>::iterator itGOBSf2 = gf2_t1.begin();
      set<GOBS>::iterator itGOBSNL = gNL_t1.begin();      
      
      bool endFF = true;
      bool endNL = true;
      for(set<GOBS>::iterator itGOBSf1 = gf1_t1.begin(); ; ){
	 if(gf1_t1.size() == 0 || gf2_t1.size() == 0) break;
	 
	 s1.gobs(*itGOBSf1);
	 s2.gobs(*itGOBSf2);
	 
	 t_gobs_pair gobs_pair(s1, s2);
	 
	 double diff = 0;
	 double lam = CLIGHT / gobs1->frequency_lc( s1.band(), 1, s2.band(), -1 );

	 if(i == 1){
//cout << "MW test " << endl;	    
	    double lct1 = gobs1->MW(s1, s2);
	    double lct2 = gobs2->MW(s1, s2);
//cout << gobs1->epoch().str_hms() << " " << gobs2->epoch().str_hms() << " " << gobs1->sat() << " " << gobs2str(s1.gobs()) << "  " << gobs2str(s2.gobs()) 
//	      << " MW_t1: " << lct1 << " MW_t2: " << lct2 << endl;
	    diff = (lct2 - lct1) / lam ;
	    t_gband band1 = s1.gband();
	    t_gband band2 = s2.gband();	    
	    double thr = _slipThreshold("MW", gobs2, band1, band2) / lam;
//cout << "diff: " << diff << " thr: " << thr << endl;
//int ooo; cin >> ooo;	    
	    double wlSlp = 0;
	    if(fabs(diff) > thr) {
	       slip = true;
	       wlSlp = round(diff);
//cout << gobs1->epoch().str_hms() << " " << gobs2->epoch().str_hms() << " " << gobs1->sat() << " " << gobs2str(s1.gobs()) << "  " << gobs2str(s2.gobs()) << "  " << wlSlp << endl;
	    }
	    gobs_pair.val = wlSlp;
	    _v_lcslp[i].push_back(gobs_pair);
	    _m_lcslp[i][gobs_pair] = wlSlp;
	 }else{
//cout << "WL test " << endl;	    
	    // compute wide-lane time differenced observations
	    double lct1 = gobs1->LWL(s1, s2);
	    double lct2 = gobs2->LWL(s1, s2);	 
	    double dWL = lct2 - lct1;
	    t_gband band1 = s1.gband();
	    t_gband band2 = s2.gband();	    
	    double thrWL = _slipThreshold("WL", gobs2, band1, band2) / lam;
//cout << gobs1->epoch().str_hms() << " " << gobs2->epoch().str_hms() << " " << gobs1->sat() << " " << gobs2str(s1.gobs()) << "  " << gobs2str(s2.gobs())
//	      << fixed << setprecision(3) << " WL_t1: " << lct1 << " WL_t2: " << lct2 << endl;
	    // compute extra-wide-lane time differenced observatins 
	    sF.gobs(*itGOBSFF);
	    lct1 = gobs1->LWL(sF, s2);
	    lct2 = gobs2->LWL(sF, s2);	    
	    double dEWL = lct2 - lct1;
	    gobs_pair.obs1 = sF;
//cout << gobs1->epoch().str_hms() << " " << gobs2->epoch().str_hms() << " " << gobs1->sat() << " " << gobs2str(sF.gobs()) << "  " << gobs2str(s2.gobs())
//	      << fixed << setprecision(3) << " EWL_t1: " << lct1 << " EWL_t2: " << lct2 << endl;
	    // find extra-wide-lane cycle slip from previous cascade
	    double ewlSlp = 0;
	    int x = i-1;
	    ewlSlp = _findSlp(x, gobs_pair);
	    double elam   = 0;
	    elam = CLIGHT / gobs1->frequency_lc(gobs_pair.obs1.band(), 1, gobs_pair.obs2.band(), -1);	    
	    t_gband bandF = sF.gband();
	    double thrEWL = _slipThreshold("WL", gobs2, bandF, band2) / elam;
	    
	    double thr = thrEWL + thrWL;
	    
	    diff = (dEWL - dWL + ewlSlp*elam) / lam;
//cout << gobs1->sat() << " dEWL: " << dEWL << " dWL: " << dWL << " ewlSlp " << ewlSlp << " diff " << diff << " thr: " << thr << " " << thrEWL << " " << thrWL << endl;
//int ooo; cin >> ooo;	    
	    gobs_pair.obs1 = s1;
	    double wlSlp = 0;
	    if(fabs(diff) > thr) {
	       slip = true;
	       wlSlp = round(diff);
//cout << gobs1->epoch().str_hms() << " " << gobs2->epoch().str_hms() << " " << gobs1->sat() << " " << gobs2str(s1.gobs()) << "  " << gobs2str(s2.gobs()) << " " << wlSlp << endl;	       
	    }
	    gobs_pair.val = wlSlp;
	    _v_lcslp[i].push_back(gobs_pair);	    
	    _m_lcslp[i][gobs_pair] = wlSlp;
	 }	 
	 
	 // narrow-lane slip
	 if(i == nfreq - 1){
//cout << "NL test " << endl;
	    if(gf1_t1.size() == 0 || gNL_t1.size() == 0) break;
	    double lct1 = gobs1->LWL(s1, s2);
	    double lct2 = gobs2->LWL(s1, s2);	    
	    double dWL = lct2 - lct1;
	    
	    // find wide-lane slip from last cascade
	    double wlSlp = 0;
	    wlSlp = _findSlp(i, gobs_pair);
	    
	    s_narr.gobs(*itGOBSNL);
	    lct1 = gobs1->LNL(s1, s_narr);
	    lct2 = gobs2->LNL(s1, s_narr);
	    double dNL = lct2 - lct1;
	    gobs_pair.obs1 = s1;
	    gobs_pair.obs2 = s_narr;	    
	    double nlam = CLIGHT / gobs1->frequency_lc(gobs_pair.obs1.band(), 2, gobs_pair.obs2.band(), -1);
	    double disf = _disf(gobs1, gobs2, s1, s_narr);
	    if (!slip) _iono(gobs1, gobs2, s1, s_narr);
	    string prn = gobs1->sat();
	    
	    diff = (dWL - dNL - disf*_dI[prn] + wlSlp*lam) / nlam;
//cout << gobs1->epoch().str_hms() << " " << gobs2->epoch().str_hms() << " " << gobs1->sat() << " " << gobs2str(s1.gobs()) << "  " << gobs2str(s_narr.gobs()) 
//       << " " << diff << " " << nlam << "  " << dWL << "  " << dNL << " " << wlSlp << "  " << disf*_dI[prn] << " " << disf << " " << _dI[prn] << endl;
//int ooo; cin >> ooo;	    
	    double nlSlp = 0;
	    if(fabs(diff) > 1) {	 
	       nlSlp = round(diff);
//cout << gobs1->epoch().str_hms() << " " << gobs2->epoch().str_hms() << " " << gobs1->sat() << " " << gobs2str(s1.gobs()) << "  " << gobs2str(s2.gobs()) << " " << nlSlp << endl;
	    }
	    gobs_pair.val = nlSlp;
	    _v_lcslp[i+10].push_back(gobs_pair);	    
	    _m_lcslp[i+10][gobs_pair] = nlSlp;
	 }

	 if(i>1) {
	    ++itGOBSFF;
	    if(itGOBSFF == gFF_t1.end()) endFF = true;
	    else endFF = false;
	 }
	 if(i==nfreq-1) {
	    ++itGOBSNL;
	    if(itGOBSNL == gNL_t1.end()) endNL = true;
	    else endNL = false;
	 }	 
	 ++itGOBSf1; ++itGOBSf2;
	 if(itGOBSf1 == gf1_t1.end() && itGOBSf2 == gf2_t1.end() && endFF && endNL) break;
	 if(itGOBSf1 == gf1_t1.end()) --itGOBSf1;
	 if(itGOBSf2 == gf2_t1.end()) --itGOBSf2;
	 if(i>1 && endFF) --itGOBSFF;
	 if(i==nfreq-1 && endNL) --itGOBSNL;	 
      }
   }
   return 1;
}

// Costructor for t_gobs_pair
// ----------------------------
t_gobs_pair::t_gobs_pair(t_gobs& gobs1, t_gobs& gobs2)
: obs1(gobs1),
  obs2(gobs2)
{   
}


// operator< for t_gobs_pair
// // --------------------------
bool t_gobs_pair::operator<(const t_gobs_pair& t) const
{
   return ( this->obs1.attr() < t.obs1.attr() ||
	    
	    this->obs2.attr() < t.obs2.attr() );   
}

// ionosphere scale factor: wl - nl
// --------------------------------------
double t_gpreproc::_disf(t_gobsgnss* gobs1, t_spt_gobs gobs2, t_gobs& s1, t_gobs& s2)
{
   double isf_wl = gobs1->isf_lc(s1.band(), 1, s2.band(), -1);
   double isf_nl = gobs1->isf_lc(s1.band(), 2, s2.band(), -1);
   double disf = isf_wl - isf_nl;
   
   return disf;
}

// ionodphere
// -----------------------------------------
void t_gpreproc::_iono(t_gobsgnss* gobs1, t_spt_gobs gobs2, t_gobs& s1, t_gobs& s2)
{
   double k  = pow(gobs1->frequency(s1.band()),2) / pow(gobs1->frequency(s2.band()),2);

   double Lb1_1 = gobs1->obs_L(s1);
   double Lb2_1 = gobs1->obs_L(s2);
   double Lb1_2 = gobs2->obs_L(s1);
   double Lb2_2 = gobs2->obs_L(s2);
   double dL = Lb1_2 - Lb2_2 - Lb1_1 + Lb2_1;   
 
   double dI = -dL/(k-1);
   
   string prn = gobs1->sat();
   
#ifdef DEBUG   
   cout << "iono: " << dL << " " << Lb1_2 << " " << Lb2_2 << " " << Lb1_1 << " " << Lb2_1 << endl;
#endif   
   
   _dI[prn] = dI;
   
//   ddI = _dI[prn] - dI;      
}

// find wl slip
// ----------------------------------
double t_gpreproc::_findSlp(int& i, t_gobs_pair& gpair)
{
   double wlSlp = 0;
   if( _m_lcslp.find(i) != _m_lcslp.end() ){
      if( _m_lcslp[i].find(gpair) != _m_lcslp[i].end() ) {
	 wlSlp = _m_lcslp[i][gpair];
      }	       
   }	    
   
   return wlSlp;
}

// report epo data as slips due to epoch data gap
// ------------------------------------   
void t_gpreproc::_gapReport(vector<shared_ptr<t_gobsgnss>> epoData)
{   
   for(vector<shared_ptr<t_gobsgnss>>::iterator it = epoData.begin(); it != epoData.end(); it++ ){
      t_gtime epo = (*it)->epoch();
      string prn  = (*it)->sat(); 
      vector<GOBS> allobs = (*it)->obs();      
      for(vector<GOBS>::iterator itGOBS = allobs.begin(); itGOBS != allobs.end(); itGOBS++){
	 if(gobs_phase(*itGOBS)) _mslipsGap[epo][prn][*itGOBS] = 1;
      }
   }
}
   
// report epo data as slips due to satellite data gap
// ------------------------------------   
void t_gpreproc::_gapReport(shared_ptr<t_gobsgnss> data)
{   
   t_gtime epo = data->epoch();
   string prn  = data->sat();
   vector<GOBS> allobs = data->obs();      
   for(vector<GOBS>::iterator itGOBS = allobs.begin(); itGOBS != allobs.end(); itGOBS++){
      if(gobs_phase(*itGOBS)) _mslipsGap[epo][prn][*itGOBS] = 2;
   }
}   

// compare two epoch data
// --------------------------------
void t_gpreproc::_compare(vector<t_gobsgnss> data1, vector<shared_ptr<t_gobsgnss>> data2)
{
   for(vector<shared_ptr<t_gobsgnss>>::iterator it2 = data2.begin(); it2 != data2.end(); it2++){
      t_gtime epo = (*it2)->epoch();
      string prn  = (*it2)->sat();	
      bool foundPRN = false;
      for(vector<t_gobsgnss>::iterator it1 = data1.begin(); it1 != data1.end(); it1++){
	 if( (*it2)->sat() == it1->sat() ) {
	    foundPRN = true;	 	    
	    vector<GOBS> vgobs1 = it1->obs();
	    vector<GOBS> vgobs2 = (*it2)->obs();
	    for(vector<GOBS>::iterator itGOBS2 = vgobs2.begin(); itGOBS2 != vgobs2.end(); itGOBS2++){
	       bool foundGOBS = false;
	       for(vector<GOBS>::iterator itGOBS1 = vgobs1.begin(); itGOBS1 != vgobs1.end(); itGOBS1++){
		  if(*itGOBS2 == *itGOBS1) {foundGOBS = true; break;}
	       }
	       if(!foundGOBS && gobs_phase(*itGOBS2)) _mslipsGap[epo][prn][*itGOBS2] = 3;
	    }
	    break;
	 }	 	 
      }
      if(!foundPRN) _gapReport(*it2);
   }
}
   
   
} // namespace
