
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/
#include "gproc/gspp.h"
#include "gmodels/gsppmodel.h"

namespace gnut {

// Constructor
t_gspp::t_gspp(string mark, t_gsetbase* set)
: _site(mark),
  _set(set),
  _log(0),
  _res(0),  
  _gobs(0),
  _gnav(0),
  _allprod(0),
  _phase(false),
  _single_frq(false),
  _tropo(true),     
  _gnss(GNS),
  _minElev(10.0),
  _initialized(false)
{
   _setLog();       // set log  (before using it later on!)
   
   _gModel   = new t_gsppmodel(_site,_log,_set);
}

// Destructor
t_gspp::~t_gspp()
{
   if( _gModel   ) delete _gModel;
   
   if( _log ){ if( _log->is_open() ){ _log->close(); }; delete _log; }
   if( _res ){ if( _res->is_open() ){ _res->close(); }; delete _res; }   
}

// Set obs, nav
// --------------------------------
void t_gspp::setDAT(t_gallobs* gobs, t_gallnav* gnav)
{      
  _gobs = gobs;
  _gnav = gnav;
}

// set output for products
// ----------------------------
void t_gspp::setOUT(t_gallprod* products)
{
   this->_setOut();
   _allprod = products;
}

// Tropo is used or not
// -------------------------
void t_gspp::tropo(bool tropo)
{
   this->_tropo = tropo;
}

// Tropo is used or not
// -------------------------
void t_gspp::phase(bool phase)
{
   this->_phase = phase;
}
   
// Tropo is used or not
// -------------------------
void t_gspp::single_frq(bool single_frq)
{
   this->_single_frq = single_frq;
}   

// set used GNSS
// --------------------------
void t_gspp::setgnss(GSYS sys)
{
   this->_gnss = sys;
}

// Get settings
// -----------------------------
int t_gspp::_get_settings()
{
  _sats          = dynamic_cast<t_gsetgnss*>(_set)->sat();
  _tropo         = dynamic_cast<t_gsetproc*>(_set)->tropo();
  _single_frq    = dynamic_cast<t_gsetproc*>(_set)->single_frq();
  _mf            = dynamic_cast<t_gsetproc*>(_set)->tropo_mf();
  _sampling      = dynamic_cast<t_gsetgen*>(_set)->sampling();
  _aprox_ztd_xml = dynamic_cast<t_gsetrec*>(_set)->ZTD();   
  _sig_init_crd  = dynamic_cast<t_gsetproc*>(_set)->sig_init_crd();
  _sig_init_ztd  = dynamic_cast<t_gsetproc*>(_set)->sig_init_ztd();
  _sig_init_glo  = dynamic_cast<t_gsetproc*>(_set)->sig_init_glo();
  _sig_init_gal  = dynamic_cast<t_gsetproc*>(_set)->sig_init_gal();
  _sig_init_bds  = dynamic_cast<t_gsetproc*>(_set)->sig_init_bds();
  _sig_init_qzs  = dynamic_cast<t_gsetproc*>(_set)->sig_init_qzs();   
  _crd_constr    = dynamic_cast<t_gsetrec*>(_set)->CRD_constr();   
  _sigCodeGPS    = dynamic_cast<t_gsetproc*>(_set)->sigma_C_GPS();
  _sigCodeGLO    = dynamic_cast<t_gsetproc*>(_set)->sigma_C_GLO();
  _sigCodeGAL    = dynamic_cast<t_gsetproc*>(_set)->sigma_C_GAL();
  _sigCodeBDS    = dynamic_cast<t_gsetproc*>(_set)->sigma_C_BDS();
  _sigCodeQZS    = dynamic_cast<t_gsetproc*>(_set)->sigma_C_QZS();
  _sigPhaseGPS   = dynamic_cast<t_gsetproc*>(_set)->sigma_L_GPS();
  _sigPhaseGLO   = dynamic_cast<t_gsetproc*>(_set)->sigma_L_GLO();
  _sigPhaseGAL   = dynamic_cast<t_gsetproc*>(_set)->sigma_L_GAL();
  _sigPhaseBDS   = dynamic_cast<t_gsetproc*>(_set)->sigma_L_BDS();
  _sigPhaseQZS   = dynamic_cast<t_gsetproc*>(_set)->sigma_L_QZS();

  if( _sampling == 0 ) _sampling = dynamic_cast<t_gsetgen*>(_set)->sampling_default();

  return 1;
}

// Return value
// ----------
bool t_gspp::_valid_crd_xml()
{
  if ( double_eq ( _aprox_crd_xml[0], 0.0 ) &&
       double_eq ( _aprox_crd_xml[1], 0.0 ) &&
       double_eq ( _aprox_crd_xml[2], 0.0 ) ) {
         return false;
  }else{ return true;}
}

// Return value
// ----------
bool t_gspp::_valid_ztd_xml()
{
  if ( _aprox_ztd_xml > 0.5 && _aprox_ztd_xml < 5.0 ) { return true;}
  else { return false;}
}

// Set Out
// -----------------------------
void t_gspp::_setOut()
{
  string tmp( dynamic_cast<t_gsetout*>(_set)->outputs("res") );
  if( !tmp.empty() ){
    substitute(tmp,"$(rec)",_site, false);
    _res = new t_glog;
    _res->tsys(t_gtime::GPS);
    _res->mask(tmp);
    _res->append( dynamic_cast<t_gsetout*>(_set)->append() );
  }
}

// Set log
// -----------------------------
void t_gspp::_setLog()
{
//   cout << " JSEM TU original _setLog !\n"; 
  string tmp( dynamic_cast<t_gsetout*>(_set)->outputs("ppp") );
  if( !tmp.empty() ){
    substitute(tmp,"$(rec)",_site, false);
    _log = new t_glog;
    _log->tsys(t_gtime::GPS);
    _log->mask(tmp);
    _log->append( dynamic_cast<t_gsetout*>(_set)->append() );
    _log->verb(   dynamic_cast<t_gsetout*>(_set)->verb() );
  }
}
   
} // namespace
