/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include "gproc/gsppflt.h"
#include "gutils/gmatrixconv.h"

namespace gnut {

// Global variables
// -----------------------------------------------------------
const unsigned int MINOBS = 5;
// -----------------------------------------------------------

// constructor
// ----------
t_gsppflt::t_gsppflt(string mark, t_gsetbase* set)
: t_gspp(mark, set),
   _filter(0),
   _smooth(false)
{
  gtrace("t_gsppflt::t_gsppflt");

   t_gspp::_get_settings();
  _cEpo_beg    = _cEpo_end = t_gtime::first_time();      
   
  string fltModStr(dynamic_cast<t_gsetflt*>(_set)->method_flt() );
  if      (fltModStr.compare("kalman") == 0) _filter = new t_kalman();
  else if (fltModStr.compare("srcf")   == 0) _filter = new t_SRF();
  else
    if( _log ) _log->comment(0, "gppp", _site + " not correct filter setting - check XML-config");
    else                        cerr << _site + " not correct filter setting - check XML-config\n";   

  _trpStoModel = new t_randomwalk();
  _trpStoModel->setq(dynamic_cast<t_gsetflt*>(_set)->rndwk_ztd() );   
  _gloStoModel = new t_whitenoise(dynamic_cast<t_gsetflt*>(_set)->noise_glo() );
  _galStoModel = new t_whitenoise(dynamic_cast<t_gsetflt*>(_set)->noise_gal() );
  _bdsStoModel = new t_whitenoise(dynamic_cast<t_gsetflt*>(_set)->noise_bds() );
  _qzsStoModel = new t_whitenoise(dynamic_cast<t_gsetflt*>(_set)->noise_qzs() );   
  _clkStoModel = new t_whitenoise(dynamic_cast<t_gsetflt*>(_set)->noise_clk() );
  _crdStoModel = new t_whitenoise(dynamic_cast<t_gsetflt*>(_set)->noise_crd() );

  _vBanc.ReSize(4); _vBanc = 0.0;   
   
  int ipar = 0;   
   
  // Add coordinates parameters
  if( _crd_constr.compare("FIX") != 0 ){
     _param.addParam( t_gpar(_site, t_gpar::CRD_X, ++ipar, "") );
     _param.addParam( t_gpar(_site, t_gpar::CRD_Y, ++ipar, "") );
     _param.addParam( t_gpar(_site, t_gpar::CRD_Z, ++ipar, "") );      
    }         

  // Add receiver clock parameter
  _param.addParam( t_gpar(_site, t_gpar::CLK, ++ipar, "") );   

  // Add tropospheric wet delay parameter
  if( _tropo ){
     t_gpar par_trp(_site, t_gpar::TRP, ++ipar, "");
     par_trp.setMF( _mf );
     _param.addParam(par_trp);
  }   
   
  // Filling init parameter covariance matrix
  _Qx.ReSize( _param.parNumber() ); _Qx = 0.0;     
 double crdInit = _sig_init_crd;
 double ztdInit = _sig_init_ztd; 
 for (unsigned int i=1; i<=_param.parNumber(); i++) {
   if (_param[i-1].parType == t_gpar::CRD_X )	
     _Qx(i,i) = crdInit * crdInit;
   else if (_param[i-1].parType == t_gpar::CRD_Y)
     _Qx(i,i) = crdInit * crdInit;
   else if (_param[i-1].parType == t_gpar::CRD_Z)
     _Qx(i,i) = crdInit * crdInit;
   else if (_param[i-1].parType == t_gpar::CLK)
     _Qx(i,i) = _clkStoModel->getQ() * _clkStoModel->getQ();
   else if (_param[i-1].parType == t_gpar::TRP)
     _Qx(i,i) = ztdInit * ztdInit;    
 }   
   
} // end constructor


// destructor
// ----------
t_gsppflt::~t_gsppflt()
{ 
  gtrace("t_gsppflt::~t_gsppflt");
   
  if( _trpStoModel ) delete _trpStoModel;   
  if( _filter      ) delete _filter;   
  if( _gloStoModel ) delete _gloStoModel;
  if( _galStoModel ) delete _galStoModel;
  if( _bdsStoModel ) delete _bdsStoModel;
  if( _qzsStoModel ) delete _qzsStoModel;   
  if( _clkStoModel ) delete _clkStoModel;
  if( _crdStoModel ) delete _crdStoModel;   
}

// Process observation batch
// ----------
int t_gsppflt::processBatch(const t_gtime& beg, const t_gtime& end)
{

  gtrace("t_gsppflt::processBatch");

  if( beg > end ){ cerr << _site << " - processing not started [beg > end]\n"; return -1; }

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif 
  _gmutex.lock();

  _dop.set_data(_gnav, _gobs, _site);   
   
  t_gtime begT( beg );
  t_gtime endT( end );

  _param.setSite(_site);
  t_gtriple xyz, neu, ell;

  double subint = pow(10,floor(log10(_sampling)));

  int sign  = fabs(_sampling)/_sampling;  
     
   t_gtime now( begT );

#ifdef DEBUG
   if( _log ) _log->comment(1,"t_gpppflt", begT.str_ymdhms( "Filtering: " ) + endT.str_ymdhms(" "));
#endif

   t_gtime timeT(now); 
     
//   cout << _site << ": Start filtering: " << _site << " " << now.str_ymdhms() << endl;
   while(now < end || now == end)
   {
     // synchronization of now != end
     if(now != end){
	if( now.sod()%_sampling != 0 ){	
	   now.add_dsec( sign * subint/100 ); // add_dsec used for synchronization!
	   continue;
	}
	now.reset_dsec();
     }

     // clean/collect/filter epoch data
     _data.erase( _data.begin(), _data.end() );
     _data = _gobs->obs( _site, now );                
      
     if( _data.size() == 0){
       if( _log && _log->verb() >= 1 )
	   _log->comment(1,"gspp", _site + now.str_ymdhms(" no observation found at epoch: "));

       now.add_sec( sign * _sampling );  // not yet add_dsec() ! not support >1Hz data
       continue;
     }

      t_gtime obsEpo = _data.begin()->epoch();       
     _timeUpdate(obsEpo); 

     int irc_epo = _processEpoch(obsEpo);

     if( irc_epo < 0 ) {
       if( _log && _log->verb() >= 1 ) 
	   _log->comment(1,"gspp", _site + now.str_ymdhms(" epoch ") + " was not calculated" );

       now.add_sec( sign * _sampling ); // not yet add_dsec() ! not support >1Hz data
       continue;
     }      

     if( _log && _log->verb() >= 1 )
	 _log->comment(1,"gspp", _site + now.str_ymdhms(" processing epoch: "));

     if( _param.getCrdParam(_site, xyz) > 0 ){
     }else { if( _log ) _log->comment(1,"gspp", _site + now.str_ymdhms(" No coordinates included in params: "));
     }      
      
     xyz2ell(xyz, ell);

     if (_allprod != 0){ 
       shared_ptr<t_gprodcrd> prd = make_shared<t_gprodcrd>(now); // now is rounded value !!
       prd->xyz(xyz);
       _allprod->add(prd, _site);
     }                  

     _initialized = true;

     now.add_sec( sign * _sampling ); // not yet add_dsec() ! not support >1Hz data
            
   }      
   
   _gmutex.unlock();

  return 1;
}

// Process one epoch
// -----------------------------
int t_gsppflt::_processEpoch(const t_gtime& runEpoch)
{
   
   gtrace("t_gsppflt::_processEpoch");

#ifdef DEBUG
  cout << "gpppflt: Epoch - full data size: " << _data.size() << " " << runEpoch.str_hms() << endl;
#endif

  _epoch = runEpoch;

  Matrix A;
  DiagonalMatrix P;
  ColumnVector l, dx, dx_el;
  SymmetricMatrix Qsav;
  ColumnVector    v;
  vector<t_gsatdata>::iterator it;
  vector<GOBSTYPE> obstypes;
   
  _cntrep = 0;    // number of iterations caused by outliers

  if( _prepareData() < 0 )  {
//    cout << "gppflt: Epoch not calculated due to prepareDate " << _epoch.str_hms() << endl;
     return -1;
  }
   
  do{     
#ifdef DEBUG
  cout << "gpppflt: Epoch - process data size: " << _data.size() << " " << runEpoch.str_hms() << endl;
#endif               
     
     if ( _data.size() < static_cast<int>(MINOBS) ) { 
//    cout << "Epoch " << _epoch.str_hms() << " not calculated due to MINOBS." << endl;
	return -1;
     }
     
    _predict();

     
    t_gtriple groundEll, groundXYZ; 

    if ( _param.getCrdParam(_site, groundXYZ) > 0){
    }else if( _valid_crd_xml() ) { groundXYZ = _aprox_crd_xml; }

    xyz2ell(groundXYZ.crd_array(), groundEll.crd_array());

    unsigned int nObs;    
    if( ! _phase ) nObs = _data.size();   // P3 used only
    else nObs = 2 * _data.size();        // P3 and L3 used together
     
    if (_sigCodeGLO >= 999) nObs -= _numSat(GLO);

    unsigned int nPar = _param.parNumber();
	
    A.ReSize(nObs, nPar); A = 0.0;	
    l.ReSize(nObs);       l = 0.0;	
    P.ReSize(nObs);       P = 0.0;	 
    dx.ReSize(nPar);     dx = 0.0;	     	
    unsigned int iobs = 1;
     
    // Create matrixes and vectors for estimation
    // ----------------------------------------------------------------------------------	     	
     for( it = _data.begin(); it != _data.end(); it++ ){

       if( _addObsP(*it, iobs, groundEll, A, l, P) > 0 ) obstypes.push_back(TYPE_C);
       if( _addObsL(*it, iobs, groundEll, A, l, P) > 0 ) obstypes.push_back(TYPE_L); 	
       
     } // end loop over all measurement in current epoch

     // delete zero rows/cols
     iobs--;
     A = A.Rows(1,iobs);
     P = P.SymSubMatrix(1,iobs);
     l = l.Rows(1,iobs);
     
// end of creation objects for estimation
// ------------------------------------------------------------------------------------------------     
/*
// Add pseudomeasurments as constrain
// --------------------------------------
    // Design Matrix
    if( _epoch > _cEpo_beg && _epoch < _cEpo_end )
    {
#ifdef DEBUG       
      cout << _epoch.str_ymdhms(" epoch: ") << endl;
      cout << _cEpo_beg.str_ymdhms(" cEpo_beg: ") << endl;
      cout << _cEpo_end.str_ymdhms("_cEpo_end: ") << endl;
#endif

      Matrix_addRC(A, A.Nrows()+1, 0);
      int i = _param.getParam(_site, t_gpar::TRP, "");
      A(A.Nrows(), i+1) = 1;

      // tropo (wet part) is constrained to this value 
      double TRP_fix = 0;
      if( _gModel->tropoModel != 0 ) TRP_fix = _aprox_ztd_xml - _gModel->tropoModel->getZHD(groundEll, _epoch);       
       
      // Reduced measurement
      Vector_add(l, l.Nrows()+1);
      l(l.Nrows()) = _param[i].value - TRP_fix;
     
       // weight matrix
      Matrix_addRC(P, P.Nrows()+1);
      P(P.Nrows(), P.Ncols()) = 99999;
    }
// -------------------------------------     
*/ 

     Qsav =  _Qx;

#ifdef DEBUG       
 if( _log )
     _log->comment(1, "t_gpppflt","before update");
#endif       

     _filter->update(A, P, l, dx, _Qx);

     // Test if _Qx is positive definite
     try {
       LowerTriangularMatrix L = Cholesky(_Qx);
     }     
     catch (Exception NPDException) {
       if( _log )
          _log->comment(1, "gppp", _site + _epoch.str_ymdhms(" epoch ")
                                         + " variance-covariance matrix is not positive definite.\n");
     }
//     int i = _param.getParam(_site, t_gpar::CLK, "");    

     v = l - A*dx;     
//     ColumnVector vtPv = v.t()*P*v;
//     _sig_unit = vtPv(1) / (A.Nrows() - A.Ncols());
     
#ifdef DEBUG
    cout << "Used satellites: ";
    for (unsigned int i = 0; i < _data.size(); i++) cout << _data[i].sat() << " ";
     cout << endl;
  cout << "Used parameters: ";
    for (unsigned int i = 0; i < _param.parNumber(); i++) cout << _param[i].str_type() << " ";     
    cout << endl;
    cout << fixed   << setprecision(3)
         << _epoch.str_ymdhms() << endl
         << " l: "  << setw(7) << l 
         << " dx: " << setw(7) << dx
         << " v.Nrows(): " << v.Nrows() 
         << " v: "  << setw(7) << v 
         << " A: "  << setw(7) << A
         << " P: "  << setw(7) << P
         << " Q: "  << setw(7) << _Qx
         << endl; cout.flush();
//int ooo; cin >> ooo;
#endif

  }while( _gModel->outlierDetect(_data, _Qx, Qsav, v, obstypes) != 0 );

  if( _data.size() < static_cast<unsigned int>(MINOBS) ){
     
    if( _log )
	_log->comment(3, "gppp", _site + _epoch.str_ymdhms(" epoch ")
		                       + " skipped (_data.size < MINOBS)");
//    cout << "Epoch " << _epoch.str_hms() << " not calculated due to MINOBS 2." << endl;     
     return -1;
  }

   // Edit and save parematres
  for( unsigned int iPar = 0; iPar < _param.parNumber(); iPar++ ) {
     _param[iPar].value += dx(_param[iPar].index);
  }   

  return 1;   
   
}

// weight coef for observations
// ---------------------------
double t_gsppflt::_weightObs(t_gsatdata& satdata)
{
   double weight_coef = 0;
   
   if (satdata.ele()*180.0/M_PI > 30) weight_coef = 1.0;
   else weight_coef = cos(satdata.ele());//* cos(satdata.ele());
   
   return weight_coef;
}


// add data to A, l, P - code
// ---------------------------
int t_gsppflt::_addObsP(t_gsatdata& satdata, unsigned int& iobs, t_gtriple& ell, Matrix& A, ColumnVector& l, DiagonalMatrix& P)
{
   gtrace("t_gsppflt::_addObsP");
   
   t_gsys gsys(satdata.gsys());
   GSYS gs = gsys.gsys();

//   int b1 = gsys.first_band(gs,1);
//   int b2 = gsys.second_band(gs,b1);
   GOBSBAND b1 = BAND_1;
   GOBSBAND b2 = BAND_2;   
   
   // Create reduced measurement, obs. weight matrix for P3
   double weight_coef = _weightObs(satdata);

   double modObsP = _gModel->cmpObs(_epoch, _param, satdata, false);
   
   double P3 = satdata.P3(b1, b2);
   if( double_eq(P3, 0.0) ){
      if(_single_frq) P3 = satdata.obs_SF();
      else return -1;
   }  
   
   if (gs == GPS){
      P(iobs,iobs) =  1 / (_sigCodeGPS * _sigCodeGPS)  / weight_coef ;      
      l(iobs) = P3 - modObsP;
   }else if (gs == GLO && _sigCodeGLO < 999){
      P(iobs,iobs) =  1 / (_sigCodeGLO * _sigCodeGLO)  / weight_coef ;
      l(iobs) = P3 - modObsP;
   }else if (gs == GAL){
      P(iobs,iobs) =  1 / (_sigCodeGAL * _sigCodeGAL)  / weight_coef ;
      l(iobs) = P3 - modObsP;      			  
   }else if (gs == BDS){				  
      P(iobs,iobs) =  1 / (_sigCodeBDS * _sigCodeBDS)  / weight_coef ;
      l(iobs) = P3 - modObsP;
   }else if (gs == QZS){
      P(iobs,iobs) =  1 / (_sigCodeQZS * _sigCodeQZS)  / weight_coef ;
      l(iobs) = P3 - modObsP;      
   }
   
   // Create first design matrix for P3       
   if ( (gs == GLO && _sigCodeGLO < 999) || 
	 gs == GPS ||
	 gs == GAL ||
	 gs == BDS ||
	 gs == QZS	 ){
      
      for( unsigned int ipar=1; ipar <= _param.parNumber(); ipar++ ){
	 A(iobs, ipar) = _param[ipar-1].partial( satdata, _epoch, ell, false);	 
      }
   } else return -1;
   
#ifdef DEBUG       
      cout << satdata.sat()     << fixed << setprecision(10)
     << "  P3: (" << BAND_1 << "," << BAND_2 << ") :"  << setw(12) << satdata.P3(b1,b2)
           << "  cmpObs: "  << setw(12) << modObsP
           << endl;
#endif   

//   if (gs == GLO && _sigCodeGLO >= 999) iobs--;                
   
   iobs++;
   
   return 1;
}

// add data to A, l, P - carrier phase
// ---------------------------
int t_gsppflt::_addObsL(t_gsatdata& satdata, unsigned int& iobs, t_gtriple& ell, Matrix& A, ColumnVector& l, DiagonalMatrix& P)
{
   gtrace("t_gsppflt::_addObsL");   
   return 1;
}


// Prepare data: 
// filter, bancroft, members in gsatdata
// ----------
int t_gsppflt::_prepareData()
{   
   gtrace("t_gpppflt::_prepareData");
   
   vector<t_gsatdata>::iterator it = _data.begin();
   while (it != _data.end())
   {
     switch (_gnss) {
       case GNS:
	 break;
       case GPS:
         if (it->gsys() != GPS) {_data.erase(it); continue;}
	 break;
       case GLO:
         if (it->gsys() != GLO) {_data.erase(it); continue;}
	 break;
       case GAL:
         if (it->gsys() != GAL) {_data.erase(it); continue;}
	 break;
       case BDS:
         if (it->gsys() != BDS) {_data.erase(it); continue;}
	 break;	
       case QZS:
         if (it->gsys() != QZS) {_data.erase(it); continue;}
	 break;	
       case SBS:
         if (it->gsys() != SBS) {_data.erase(it); continue;}
	 break;	
     }
      it++;
   }

  Matrix BB;

  BB.ReSize(_data.size(), 4);
  BB = 0.0;   
  int iobs = 0;
  vector<t_gsatdata>::iterator iter = _data.begin();
//int N_GPS = _satNum(_data, GPS);

  while( iter != _data.end() ){
          
//cout << "iter->epoch = " << iter->epoch().str_hms() << endl;
//cout << "_epoch = " << _epoch.str_hms() << endl;

//    t_gsys gsys(iter->gsys());
//    GSYS gs = gsys.gsys();
//    int b1 = gsys.first_band(gs,1);
//    int b2 = gsys.second_band(gs,b1);
//
    GOBSBAND b1 = BAND_1;
    GOBSBAND b2 = BAND_2;
     
    iter->glog(_log); 

    // check data availability
    double P3 = iter->P3(b1, b2);
    double L3 = iter->L3(b1, b2);          

    if( double_eq(L3, 0.0) ) {
	if( _log ) {
	   ostringstream str; 
	   str << "t_gpppflt::prepareData(VEC): erasing data due to no phase double bands observation, " <<
	     "epo: " << _epoch.str_hms() << ", "        <<
	     "prn: " << iter->sat() << " (" << iter->channel() << ")" ;
	   _log->comment(2, "gppp", str.str());
	}
	
	_data.erase( iter ); // !!!! OK u vektoru funguje bez ++ (erase zvysuje pointer !)		
	continue;
    }
            
    if( double_eq(P3, 0.0) ) {
       if(_single_frq){
	  double sf = iter->obs_SF();
	  if( double_eq(sf, 0.0) ){
	     if( _log ) {
		ostringstream str; 
		str << "t_gpppflt::prepareData(VEC): erasing data due to no code single bands observation, " <<
		  "epo: " << _epoch.str_hms() << ", "        <<
		  "prn: " << iter->sat();
		_log->comment(2, "gppp", str.str());
	     }
	     
	     _data.erase( iter ); // !!!! OK u vektoru funguje bez ++ (erase zvysuje pointer !)
	     continue;
	  }
       }else{
	     if( _log ) {
		ostringstream str; 
		str << "t_gpppflt::prepareData(VEC): erasing data due to no code double bands observation, " <<
		  "epo: " << _epoch.str_hms() << ", "        <<
		  "prn: " << iter->sat();
		_log->comment(2, "gppp", str.str());
	     }
	  
	  _data.erase( iter ); // !!!! OK u vektoru funguje bez ++ (erase zvysuje pointer !)		
	  continue;
       }
    }
     
    if( _satPos( _epoch, *iter ) < 0 ){
       if( _log ) {
        ostringstream str; 
        str << "t_gpppflt::prepareData(VEC): erasing data since _satPos failed, " <<
               "epo: " << _epoch.str_hms() << ", "        <<
               "prn: " << iter->sat();
        _log->comment(2, "gppp", str.str());
      }       
#ifdef DEBUG
    cout << "prepareData: sat pos not calculated "
	 << _epoch.str_ymdhms(_site+" "+iter->sat()+" "+int2str(_data.size())+" ") 
	 << endl; cout.flush();
#endif       
      _data.erase( iter ); // !!!! OK u vektoru funguje bez ++ (erase zvysuje pointer !)

    }else{// if (iter->gsys() == gnss){
      iobs++;
      BB(iobs,1) = iter->satcrd().crd(0);
      BB(iobs,2) = iter->satcrd().crd(1);
      BB(iobs,3) = iter->satcrd().crd(2);
      BB(iobs,4) = iter->P3(b1,b2) + iter->clk();
      iter++;
//    }else{
//       iter++;
       continue;
    }

  }

   if( _data.size() < static_cast<int>(MINOBS) ) {
      if( _log && _log->verb() >= 2 )
	  _log->comment(2, "gppp", _epoch.str_ymdhms( _site + " epoch ") + int2str(_data.size())
			          + " skipped (Bancroft not calculated: _data.size < MINOBS)");
//      cout << t_gsys::gsys2str(_gnss) << " " << _epoch.str_ymdhms( _site + " epoch ") << " " << int2str(_data.size()) <<  " skipped (Bancroft not calculated: _data.size < MINOBS)" << endl;
      return -1;
   }

  BB = BB.Rows(1,iobs);    // delete zero rows

  if( BB.Nrows() < static_cast<int>(MINOBS) ) {
      if( _log && _log->verb() >= 2 )
          _log->comment(2, "gppp", _epoch.str_ymdhms(_site + " epoch ")
			         + " skipped (Bancroft not calculated: BB.Nrows < MINOBS)");
//      cout << _epoch.str_ymdhms( _site + " epoch ") << int2str(_data.size()) <<  " skipped (Bancroft not calculated: _data.size < MINOBS)" << endl;
    return -1;
  }

  _vBanc = 0.0;
  gbancroft(BB, _vBanc);

   if ( _crd_constr.compare("APRIORI") == 0 ) {      
      _vBanc(1) = _aprox_crd_xml[0];
      _vBanc(2) = _aprox_crd_xml[1];
      _vBanc(3) = _aprox_crd_xml[2];
   }   

#ifdef DEBUG
   cout << _epoch.str_ymdhms() << endl;;
   cout << "\nBB\n" << fixed << setprecision(10) << BB << endl;
   cout << "vBanc\n"  << fixed << setprecision(10)<< _vBanc << endl;
   cout << "Clk [mili sec]" << fixed << setprecision(10) << _vBanc(4)*1000/CLIGHT << endl;
   cout << "_Qx\n"  << fixed << setprecision(3)<< _Qx << endl;    cout.flush();
#endif
      
  //Compute sat elevation and rho
  iter = _data.begin();
  while( iter != _data.end() ){

     t_gtriple xyz_r, xyz_s, neu;

     xyz_s = iter->satcrd();

     if ( _crd_constr.compare("FIX") == 0 && _valid_crd_xml() ){
	xyz_r = _aprox_crd_xml;
     }else if( !_initialized ){
	xyz_r.set(_vBanc.Rows(1,3));
    }else{
       _param.getCrdParam(_site, xyz_r);
    }

#ifdef DEBUG
    cout << "Coordinates: " << iter->sat() << " " << iter->epoch().str_hms()
         << fixed << setprecision(10) << endl
         << " X_R " << setw(12) << xyz_r[0]
         << " Y_R " << setw(12) << xyz_r[1]
         << " Z_R " << setw(12) << xyz_r[2] << endl
         << " X_S " << setw(12) << xyz_s[0]
         << " Y_S " << setw(12) << xyz_s[1]
         << " Z_S " << setw(12) << xyz_s[2] << endl; cout.flush();
#endif

    xyz2neu(xyz_s.crd_array(), xyz_r.crd_array(), neu.crd_array());
	
    t_gtriple xRec;
    double rho0 = sqrt(pow(xyz_r[0]-xyz_s[0],2) + pow(xyz_r[1]-xyz_s[1],2) + pow(xyz_r[2]-xyz_s[2],2));
    double dPhi = OMEGA * rho0 / CLIGHT;
    xRec[0] = xyz_r[0] * cos(dPhi) - xyz_r[1] * sin(dPhi);
    xRec[1] = xyz_r[1] * cos(dPhi) + xyz_r[0] * sin(dPhi);
    xRec[2] = xyz_r[2];  

    // Apply tides 
    _apply_tides(_epoch, xRec); 

    double tmp = (iter->satcrd().crd_cvect() - xRec.crd_cvect()).norm_Frobenius();

    iter->addrho( tmp );    
     
    double NE2 = neu[0]*neu[0]+neu[1]*neu[1];
    double ele = acos(sqrt(NE2)/iter->rho());
    if( neu[2]<0.0 ) {
       ele *= - 1.0;
    }

    if( sqrt(NE2)/iter->rho() > 1.0 )
         iter->addele(0.0);
    else iter->addele(ele);

    double azi = atan2(neu[1], neu[0]);
    if (azi < 0) azi += 2*M_PI;
    iter->addazi(azi);     
     
    // check elevation cut-off 
    if ( iter->ecl() || iter->ele_deg() < _minElev ){
      if( _log ) _log->comment(2, "gppp", "Erasing " + iter->sat() + " data due to elevation or eclipsing!" + iter->epoch().str_ymdhms() );
       _data.erase(iter); // !!!! zveda iterator !!!!
       if( _data.size() < static_cast<int>(MINOBS) ){
	  if( _log )
	    _log->comment(2, "", _site + _epoch.str_ymdhms(" epoch ") + int2str(_data.size())
			  + " skipped (data.size < MINOBS)");
//cout << _epoch.str_ymdhms( _site + " epoch ") << int2str(_data.size()) <<  " elevation cut off ->  _data.size < MINOBS" << endl;	  
	  return -1;
       }
    
    } else iter++;
               
  }

  return 1;
}


// Tides are not applied in SPP
// ---------------------------------
int t_gsppflt::_apply_tides(t_gtime& _epoch, t_gtriple& xRec)
{
   return 1;
}


// Predict state vector and covariance matrix
// =======================================================
void t_gsppflt::_predict()
{
   gtrace("t_gsppflt::_predict");
   
   _cntrep++;
  
   _Noise.ReSize(_Qx.Nrows()); _Noise = 0;
   
  // Predict coordinates, clock and troposphere
  // state vector is the same, covariance matrix is predicted with noise   
   int i;
   
   double crdInit       = _sig_init_crd;
   double ztdInit       = _sig_init_ztd;
   
   i = _param.getParam(_site, t_gpar::CRD_X, "");

   if ( i >= 0 )  {
      if ( !_initialized ) {
	 _param[i].value = _vBanc(1);
	 _Qx(i+1,i+1) = crdInit * crdInit;
      }      
      else if (_cntrep == 1   ) {
	 _Qx(i+1,i+1) += _crdStoModel->getQ() * _crdStoModel->getQ();
	 if(_smooth) _Noise(i+1,i+1) = _crdStoModel->getQ() * _crdStoModel->getQ();	
      }
   }

   i = _param.getParam(_site, t_gpar::CRD_Y, "");
   if ( i >= 0 )  {
     if ( !_initialized ) {
	_param[i].value = _vBanc(2);
	_Qx(i+1,i+1) = crdInit * crdInit;	
     }      
     else if (_cntrep == 1   ) {
	_Qx(i+1,i+1) += _crdStoModel->getQ() * _crdStoModel->getQ();
	if(_smooth) _Noise(i+1,i+1) = _crdStoModel->getQ() * _crdStoModel->getQ();	
     }      
   }   
   
   i = _param.getParam(_site, t_gpar::CRD_Z, "");
   if ( i >= 0 )  {
     if ( !_initialized ) {
	_param[i].value = _vBanc(3);
	_Qx(i+1,i+1) = crdInit * crdInit;	
     }      
     else if (_cntrep == 1   ) {
	_Qx(i+1,i+1) += _crdStoModel->getQ() * _crdStoModel->getQ();
	if(_smooth) _Noise(i+1,i+1) = _crdStoModel->getQ() * _crdStoModel->getQ();	
     }      
   }
   
   i = _param.getParam(_site, t_gpar::CLK, "");
   if ( i >= 0 )  {
     _param[i].value = _vBanc(4);
     for( unsigned int jj = 1; jj <= _param.parNumber(); jj++ ) _Qx(i+1, jj) = 0.0;      
      _Qx(i+1,i+1) = _clkStoModel->getQ() * _clkStoModel->getQ();
      if(_smooth) _Noise(i+1,i+1) = _clkStoModel->getQ() * _clkStoModel->getQ();      
   }
   
   i = _param.getParam(_site, t_gpar::GLO_ISB, "");
   if ( i >= 0 )  {
     if (!_initialized ) {
	_Qx(i+1,i+1) = _sig_init_glo * _sig_init_glo;
     } else if (_cntrep == 1 ) {
	 _Qx(i+1,i+1) += _gloStoModel->getQ() * _gloStoModel->getQ();
	 if(_smooth) _Noise(i+1,i+1)  = _gloStoModel->getQ() * _gloStoModel->getQ();
     }      
   }
   
   i = _param.getParam(_site, t_gpar::GAL_ISB, "");
   if ( i >= 0 )  {
     if (!_initialized ) {
	_Qx(i+1,i+1) = _sig_init_gal * _sig_init_gal;
     } else if (_cntrep == 1 ) {
	 _Qx(i+1,i+1) += _galStoModel->getQ() * _galStoModel->getQ();
	 if(_smooth) _Noise(i+1,i+1)  = _galStoModel->getQ() * _galStoModel->getQ();
     }      
   }
   
   i = _param.getParam(_site, t_gpar::BDS_ISB, "");
   if ( i >= 0 )  {
     if (!_initialized ) {
	_Qx(i+1,i+1) = _sig_init_bds* _sig_init_bds;
     } else if (_cntrep == 1 ) {
	 _Qx(i+1,i+1) += _bdsStoModel->getQ() * _bdsStoModel->getQ();
	 if(_smooth) _Noise(i+1,i+1)  = _bdsStoModel->getQ() * _bdsStoModel->getQ();
     }      
   }   
   
   i = _param.getParam(_site, t_gpar::QZS_ISB, "");
   if ( i >= 0 )  {
     if (!_initialized ) {
	_Qx(i+1,i+1) = _sig_init_qzs * _sig_init_qzs;
     } else if (_cntrep == 1 ) {
	 _Qx(i+1,i+1) += _qzsStoModel->getQ() * _qzsStoModel->getQ();
	 if(_smooth) _Noise(i+1,i+1)  = _qzsStoModel->getQ() * _qzsStoModel->getQ();
     }      
   }   
   
   i = _param.getParam(_site, t_gpar::TRP, "");
   if ( i >= 0 )  {
     if (!_initialized){
	// for determination elipsoidal crd of receiver
	t_gtriple Ell, XYZ;
	
	if (_param.getCrdParam(_site, XYZ) > 0){
	}else if ( _valid_crd_xml() ) { XYZ = _aprox_crd_xml; }
	xyz2ell( XYZ.crd_array(), Ell.crd_array());

        if ( _valid_ztd_xml() ) {
	   _param[i].value = _aprox_ztd_xml - _gModel->tropoModel->getZHD(Ell, _epoch);
        }  else _param[i].value = _gModel->tropoModel->getZWD(Ell, _epoch);
	_Qx(i+1,i+1) = ztdInit * ztdInit;
     }else if(_cntrep == 1) {
	_Qx(i+1,i+1) += _trpStoModel->getQ();
	if(_smooth) _Noise(i+1,i+1) = _trpStoModel->getQ();	
     }
      
   }
   
   // add/remove inter system bias
   _syncSys();       
   
}

// Satelite position
// ----------
int t_gsppflt::_satPos(t_gtime& epo, t_gsatdata& gsatdata)
{
  gtrace("t_gsppflt::_satPos");   
   
   string satname  =  gsatdata.sat();
   set<string>::iterator itSAT = _sats.find(satname);   
   
   if ( itSAT == _sats.end() ) {
      if( _log )
	_log->comment(2, "gppp"," satelite " + satname 
		      + _epoch.str_ymdhms(" Satellite excluded in configuration! "));
      return -1;
   }
   
   int i;
   i = gsatdata.addprd(_gnav); //add sat crd and clk
   if (i < 0) {
      return i;
   }
   
   return 1;     
}

// Get satellite number in epoch for specific system
// -----------------------------
int t_gsppflt::_numSat(GSYS gsys)
{
   int num = 0;
   
   vector<t_gsatdata>::iterator it;
   for (it = _data.begin(); it != _data.end(); it++){
      GSYS satSys = it->gsys();
      if (satSys == gsys) num++;
      else continue;
   }
   
   return num;
}

// Update time for RDW processes
// ---------------------------
void t_gsppflt::_timeUpdate(const t_gtime& epo)
{
   // tropo delay  
   _trpStoModel->updateTime(epo);
   _trpStoModel->setTcurr(epo);   
} 

// Sync inter-GNSS systems bias
// -----------------------------------
void t_gsppflt::_syncSys()
{
  gtrace("t_gsppflt::_syncSys() " + _epoch.str_hms());
   
//  set<string> mapPRN; 
  bool obsGlo = false;   
  bool obsGal = false;
  bool obsBds = false;
  bool obsQzs = false;   
  vector<t_gsatdata>::iterator it;
  for( it = _data.begin(); it != _data.end(); it++){       // loop over all observations
//    mapPRN.insert( it->sat() );

    /* // add GLO ISB  */
    /* if ( it->gsys() == GLO && _param.getParam(_site, t_gpar::GLO_ISB, it->sat()) < 0 ) { */
    /*    t_gpar newPar(_site, t_gpar::GLO_ISB, _param.parNumber()+1, it->sat()); */
    /*    newPar.value = 0.0; */
    /*    _param.addParam(newPar); */
    /*    Matrix_addRC(_Qx, _param.parNumber(), _param.parNumber()); */
    /*    _Qx(_param.parNumber(), _param.parNumber()) = _clkStoModel->getQ() * _clkStoModel->getQ(); // init isb noise = clk noise */
/* #ifdef DEBUG */
    /*    cout << "GLO_ISB was added!" << " Epoch: "  */
    /* 	    << _epoch.str_ymdhms() << endl; */
/* #endif        */
    /* } */
          
    if (it->gsys() == GLO) obsGlo = true;     
    if (it->gsys() == GAL) obsGal = true;
    if (it->gsys() == BDS) obsBds = true;
    if (it->gsys() == QZS) obsQzs = true;     
  }   

  bool parGlo = false;   
  bool parGal = false;
  bool parBds = false;
  bool parQzs = false; 
  for (unsigned int i = 0; i < _param.parNumber(); i++){
     /* // remove GLO ISB */
     /* if (_param[i].parType == t_gpar::GLO_ISB) {      */
     /* 	set<string>::iterator itPRN = mapPRN.find(_param[i].prn); */
     /* 	if (itPRN == mapPRN.end()) {	 */
     /* 	   Matrix_remRC(_Qx, _param[i].index, _param[i].index); */
     /* 	   _param.delParam(i); */
     /* 	   _param.reIndex(); */
/* #ifdef DEBUG	 */
     /* 	   cout << "GLO_ISB was removed!" << " Epoch: "  */
     /* 	     << _epoch.str_ymdhms() << endl; */
/* #endif		 */
     /* 	}      */
     /* } */
     
    if (_param[i].parType == t_gpar::GLO_ISB) parGlo = true;     
    if (_param[i].parType == t_gpar::GAL_ISB) parGal = true;
    if (_param[i].parType == t_gpar::BDS_ISB) parBds = true;
    if (_param[i].parType == t_gpar::QZS_ISB) parQzs = true;     
  }        

  // Add GLO ISB aprameter
  if ( ! parGlo && obsGlo){
      t_gpar newPar(_data.begin()->site(), t_gpar::GLO_ISB, _param.parNumber()+1, "");
      newPar.value = 0.0;
      _param.addParam(newPar);
      Matrix_addRC(_Qx, _param.parNumber(), _param.parNumber());
      _Qx(_param.parNumber(), _param.parNumber()) = _clkStoModel->getQ() * _clkStoModel->getQ(); // init isb noise = clk noise
#ifdef DEBUG	
       cout << "GLO_ISB was added!" << " Epoch: " 
            << _epoch.str_ymdhms() << endl;
#endif
  }      
   
  // Add GAL ISB aprameter
  if ( ! parGal && obsGal){
      t_gpar newPar(_data.begin()->site(), t_gpar::GAL_ISB, _param.parNumber()+1, "");
      newPar.value = 0.0;
      _param.addParam(newPar);
      Matrix_addRC(_Qx, _param.parNumber(), _param.parNumber());
      _Qx(_param.parNumber(), _param.parNumber()) = _clkStoModel->getQ() * _clkStoModel->getQ(); // init isb noise = clk noise
#ifdef DEBUG	
       cout << "GAL_ISB was added!" << " Epoch: " 
            << _epoch.str_ymdhms() << endl;
#endif
  }   

  // Add BDS ISB aprameter
  if ( ! parBds && obsBds){
      t_gpar newPar(_data.begin()->site(), t_gpar::BDS_ISB, _param.parNumber()+1, "");
      newPar.value = 0.0;
      _param.addParam(newPar);
      Matrix_addRC(_Qx, _param.parNumber(), _param.parNumber());
      _Qx(_param.parNumber(), _param.parNumber()) = _clkStoModel->getQ() * _clkStoModel->getQ(); // init isb noise = clk noise     
#ifdef DEBUG	
       cout << "BDS_ISB was added!" << " Epoch: " 
            << _epoch.str_ymdhms() << endl;
#endif
  }

  // Add QZS ISB aprameter
  if ( ! parQzs && obsQzs){
      t_gpar newPar(_data.begin()->site(), t_gpar::QZS_ISB, _param.parNumber()+1, "");
      newPar.value = 0.0;
      _param.addParam(newPar);
      Matrix_addRC(_Qx, _param.parNumber(), _param.parNumber());
      _Qx(_param.parNumber(), _param.parNumber()) = _clkStoModel->getQ() * _clkStoModel->getQ(); // init isb noise = clk noise
#ifdef DEBUG	
       cout << "QZS_ISB was added!" << " Epoch: " 
            << _epoch.str_ymdhms() << endl;
#endif
  }      

  // Remove GLO ISB paremeter
  if (parGlo && ! obsGlo){
     int i = _param.getParam(_data.begin()->site(), t_gpar::GLO_ISB, "");
     Matrix_remRC(_Qx, _param[i].index, _param[i].index);
     _param.delParam(i);
     _param.reIndex();
#ifdef DEBUG	
       cout << "GLO_ISB was removed!" << " Epoch: " 
            << _epoch.str_ymdhms() << endl;
#endif     
  }   
   
  // Remove GAL ISB paremeter
  if (parGal && ! obsGal){
     int i = _param.getParam(_data.begin()->site(), t_gpar::GAL_ISB, "");
     Matrix_remRC(_Qx, _param[i].index, _param[i].index);
     _param.delParam(i);
     _param.reIndex();
#ifdef DEBUG	
       cout << "GAL_ISB was removed!" << " Epoch: " 
            << _epoch.str_ymdhms() << endl;
#endif     
  }

  // Remove BDS ISB paremeter
  if (parBds && ! obsBds){
     int i = _param.getParam(_data.begin()->site(), t_gpar::BDS_ISB, "");
     Matrix_remRC(_Qx, _param[i].index, _param[i].index);
     _param.delParam(i);
     _param.reIndex();     
#ifdef DEBUG	
       cout << "BDS_ISB was removed!" << " Epoch: " 
            << _epoch.str_ymdhms() << endl;
#endif     
  }   

  // Remove QZS ISB paremeter
  if (parQzs && ! obsQzs){
     int i = _param.getParam(_data.begin()->site(), t_gpar::QZS_ISB, "");
     Matrix_remRC(_Qx, _param[i].index, _param[i].index);
     _param.delParam(i);
     _param.reIndex();
#ifdef DEBUG	
       cout << "QZS_ISB was removed!" << " Epoch: " 
            << _epoch.str_ymdhms() << endl;
#endif     
  }   

} // void

} // namespace
