/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iomanip>
#include <sstream>

#include "gutils/gnss.h"
#include "gutils/gsys.h"
#include "gutils/gtypeconv.h"
#include "gset/gsetgen.h"

using namespace std;
using namespace pugi;

namespace gnut {

// Constructor
// ----------
t_gsetgen::t_gsetgen() 
 : t_gsetbase()
{ 
  _set.insert(XMLKEY_GEN);
   
  // initiate STRINGS
  set<string>::const_iterator itSAT;

  t_map_sats gnss_sats = GNSS_SATS();
  t_map_sats::const_iterator itGNS;
  for( itGNS = gnss_sats.begin(); itGNS != gnss_sats.end(); ++itGNS ){
   _sys += " " + t_gsys::gsys2str( itGNS->first );
   for( itSAT = itGNS->second.begin(); itSAT != itGNS->second.end(); ++itSAT ){
    _sat += " " + (*itSAT);
  }
 }
}

// Destructor
// ----------
t_gsetgen::~t_gsetgen()
{}


// Return gtime
// ----------
t_gtime t_gsetgen::beg()
{
#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif   
  _gmutex.lock();

  string str = _doc.child(XMLKEY_ROOT).child(XMLKEY_GEN).child_value( "beg" );
  substitute(str,"\"","");
  t_gtime gt(t_gtime::GPS);
   
  if( str.empty() ){
    _gmutex.unlock();
    t_gtime tmp = t_gtime::first_time(t_gtime::GPS);
    return tmp;
  }
   
  gt.from_str( "%Y-%m-%d %H:%M:%S", trim(str) );
  _gmutex.unlock(); return gt;
}


// Return gtime
// ----------
t_gtime t_gsetgen::end()
{
#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif   
  _gmutex.lock();
   
  string str = _doc.child(XMLKEY_ROOT).child(XMLKEY_GEN).child_value( "end" );
  substitute(str,"\"","");
  t_gtime gt(t_gtime::GPS);
   
  if( str.empty() ){
    _gmutex.unlock();
    t_gtime tmp = t_gtime::last_time(t_gtime::GPS);
    return tmp;
  }
      
  gt.from_str( "%Y-%m-%d %H:%M:%S", trim(str) );
  _gmutex.unlock(); return gt;
}


// Return sampling
// ----------
int t_gsetgen::sampling()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  string str  = _doc.child(XMLKEY_ROOT).child(XMLKEY_GEN).child_value( "int" );
  int tmp = str2int(str);

  _gmutex.unlock(); return tmp;
}


// Return set
// ----------
set<string> t_gsetgen::sys()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif   
  _gmutex.lock();
   
  set<string> tmp = t_gsetbase::_setval( XMLKEY_GEN, "sys" );
  _gmutex.unlock(); return tmp;
}


// Return set
// ----------
set<string> t_gsetgen::rec()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif   
  _gmutex.lock();
   
  set<string> tmp = t_gsetbase::_setval( XMLKEY_GEN, "rec" );
  _gmutex.unlock(); return tmp;
}


// Return set
// ----------
set<string> t_gsetgen::sat()
{
#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif   
  _gmutex.lock();
   
  set<string> tmp = t_gsetbase::_setval( XMLKEY_GEN, "sat" );
  _gmutex.unlock(); return tmp;
}


// Return set
// ----------
set<string> t_gsetgen::obs()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif   
  _gmutex.lock();
   
  set<string> tmp = t_gsetbase::_setval( XMLKEY_GEN, "obs" );
  _gmutex.unlock(); return tmp;
}


// settings check
// ----------
void t_gsetgen::check()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
   
 _gmutex.lock();

  // check existence of nodes/attributes
  xml_node parent = _doc.child(XMLKEY_ROOT);
  xml_node node = _default_node(parent, XMLKEY_GEN);

  _default_node(node, "BEG", "");            // all!
  _default_node(node, "END", "");	     // all!
  _default_node(node, "SYS", "");	     // all!
  _default_node(node, "SAT", _sat.c_str());  // default
  _default_node(node, "REC", "");	     // none
  _default_node(node, "INT", int2str(DEF_SAMPLING).c_str());  // default

#ifdef BMUTEX   
   lock.unlock();
#endif

  // TO CHECK USER-SETUP CONSISTENCY
  // support >= 1Hz
  _gmutex.unlock();
  if( floor( sampling() ) < 0 ){
#ifdef BMUTEX
    boost::mutex::scoped_lock lock(_mutex);
#endif     
     
    _default_node(node, "INT", "0", true); // reset!
    cerr << "gsetgen: sampling rate settings negative: reset to 0. \n";
  }

  if( floor( sampling() ) < 1 && floor( sampling() ) != 0  ){
#ifdef BMUTEX
    boost::mutex::scoped_lock lock(_mutex);
#endif     
     
    _default_node(node, "INT", "1", true); // reset!
    cerr << "gsetgen: sampling rate settings above 1Hz is not supported: reset to 1. \n";
  }

  return;   
   
}


// help body
// ----------
void t_gsetgen::help()
{
 #ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  cerr << "\n <gen>\n"
       << "   <beg> \"1980-01-01 00:00:00\" </beg>\n"  // t_gtime::first_time().str("%Y-%m-%d %H:%M:%S")
       << "   <end> \"2079-12-31 23:59:59\" </end>\n"  // t_gtime::last_time().str("%Y-%m-%d %H:%M:%S")
       << "   <sys> GPS GLO GAL BDS QZS SBS </sys>\n"  // _sys
       << "   <rec> </rec>\n"
       << "   <int>"+int2str(DEF_SAMPLING)+"</int>\n"
       << " </gen>\n";

  cerr << "\t<!-- general description:\n"
       << "\t beg    .. beg time          (default: all)\n"
       << "\t end    .. end time          (default: all)\n"
       << "\t int    .. data sampling     (default: 30s)\n"
       << "\t sys    .. GNSS system(s)    (default: all)\n"
       << "\t rec    .. GNSS receiver(s)  (rec active list, e.g.: GOPE ONSA WTZR ... )\n"
       << "\t -->\n\n";

#ifdef BMUTEX   
  lock.unlock();
#endif

   _gmutex.unlock(); return;
}

} // namespace
