
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <math.h>

#include "gall/gallnav.h"
#include "gutils/gtime.h"
#include "gutils/gtriple.h"
#include "gutils/gsatview.h"
#include "gutils/gcommon.h"

using namespace std;

namespace gnut {

// calculate satellite visibility
// ----------
int sat_view( t_map_sky_sat&   sat_view,
	      t_gtriple        xyz_rec,
 	      t_gallnav*       all_nav,
	      const t_gtime&   beg,       // should be setup reasonably !
	      const t_gtime&   end        // should be setup reasonably !
	    )
{
  gtrace("sat_view");
  if( beg > end || beg == t_gtime::first_time() || end == t_gtime::last_time() ||
      ! all_nav ) return -1;

  t_gtriple xyz_sat;
  t_gtriple top_sat;
  double sec = 120.0; // add seconds as a function of elevation (120s ~ 1 deg)
  double r_radi, s_radi, s_topo;

  // loop over all satellites
  t_map_sats gnss_sats = GNSS_SATS();
  t_map_sats::const_iterator  itSYS;
  set<string>::const_iterator itSAT;

  for( itSYS = gnss_sats.begin(); itSYS != gnss_sats.end(); ++itSYS )
  { 
    GSYS gs = itSYS->first;

    for( itSAT = gnss_sats[gs].begin(); itSAT != gnss_sats[gs].end(); ++itSAT )
    { 
      string prn = *itSAT;

#ifdef DEBUG
      t_gtime beg_eph = all_nav->beg_time( prn );
      t_gtime end_eph = all_nav->end_time( prn );
       
      if( beg_eph == t_gtime::last_time()  ||
	  end_eph == t_gtime::first_time() )
      {
	cout << prn << " no eph - sat skipped "
	     << " per: " << beg_eph.str_ymdhms()
	     <<    " - " << end_eph.str_ymdhms() << endl;
        continue;
      }
#endif	 
       
      t_gtime beg_lim = beg;
      t_gtime end_lim = end;

#ifdef DEBUG
      double dt_lim = end_lim - beg_lim;
      if( dt_lim > 86400 || beg_lim == t_gtime::first_time() ) beg_lim = beg_eph;
      if( dt_lim > 86400 || end_lim == t_gtime::last_time()  ) end_lim = end_eph;

      cout << " sat: " << prn
	   << " per: " << beg_lim.str_ymdhms() 
	   <<    " - " << end_lim.str_ymdhms()
	   << endl;
#endif

      t_gtime tt     = beg_lim;
      t_gtime tt_sav = beg_lim;
      double ele     = 0.0;
      double ele_sav = 0.0;
      int  direction = 0;
      int  iter = 0;
      bool init = true;
      
      while( tt < end_lim + 1 && ++iter < 70 )
      {
        if( all_nav->pos( prn, tt, xyz_sat.crd_array(), NULL, NULL ) < 0 ){
//	  tt.add_sec(3600);  ele = 1.0;
//	  continue; // risky to skip nav gap
	   
	  break; // better STOP and skip this SAT, but a problem with GLONASS when missing NAV
	}

        top_sat = xyz_sat - xyz_rec;

        r_radi = xyz_rec.crd_cvect().norm_Frobenius()/1000.0; // km
        s_radi = xyz_sat.crd_cvect().norm_Frobenius()/1000.0; // km
        s_topo = top_sat.crd_cvect().norm_Frobenius()/1000.0; // km
     
        ele = -90 + acos( ( pow(s_topo, 2.0) + pow(r_radi, 2.0) - pow(s_radi, 2.0) )
	  		   / ( 2 * r_radi * s_topo ) ) / PI * 180;

	if( gs == QZS || gs == SBS ){
//	  cout << "skip geostacionary satellites QZS and SBS !\n";
	  break;
	}

	// horizon control
//	bool horizon_found = ( fabs(ele) < 0.01 && !init ); // not below 0.01! for 1s increase at least
	bool horizon_found = ( fabs(tt-tt_sav) < 2 && !init ); // MORE ROBUST, GLONASS Particularly!
	bool horizon_cross = ( (ele > 0) - (ele_sav > 0) && !init ); // alternative to a sign check
	 
	if( !init && !horizon_found ) //  !horizon_cross &&
	{
  	  if( ele_sav < ele ) direction = +1;
	  if( ele_sav > ele ) direction = -1;
	}

 	init = false;

#ifdef DEBUG	 
	cout << fixed << setprecision(3)
	     << " " << prn  
	     << "  " << tt.str_ymdhms()  
	     << "  " << tt_sav.str_ymdhms()
	     << " ele: " << setw(7) << ele  
	     << " sav: " << setw(7) << ele_sav
	     << " " << init << ":" << horizon_cross << ":" << horizon_found
	     << " " << setw(2) << direction
             << "  " << endl;
#endif

	// horizon found
	if( horizon_found ){ // && direction != 0 ){
#ifdef DEBUG
	  cout << prn << " add H 1800.000";
	  if( direction > 0 ){  cout << "  ascend horizon\n\n"; }
	  else{                 cout << " descend horizon\n\n"; }
#endif
	  // FILL 
          map<t_gtime, t_gtime>::reverse_iterator itEND = sat_view[prn].rbegin();
	  if( sat_view[prn].size() > 0 )                          // multiple visibility record
	  { if( direction < 0 )     itEND->second = tt;           // add visibility end time
	    else                sat_view[prn][tt] = end_lim;      // initiate new start with end time!
	  }
	  else                                                    // initial visibility record
	  { if( direction < 0 ) sat_view[prn][beg_lim] = tt;      // add visibility end time
	    else                sat_view[prn][tt] = end_lim;      // initiate new start with end time!
	  }

	  iter   = 0;
          tt_sav = tt;
	  tt.add_sec(1800);
	  ele_sav = direction*0.01;
	  continue;
 	}
	  
        // horizon roughly crossing
	else if( horizon_cross ){
	  
	  if( fabs(ele_sav) > fabs(ele) || fabs(ele) > 1.0 ){
//            cout << prn << " add M " << setw(8) << -sec*fabs(ele) << endl;
	    tt.add_sec( -sec*fabs(ele) );
	  }else{            
//            cout << prn << " add P " << setw(8) << sec*fabs(ele_sav) << endl;
	    if( fabs(ele_sav) < 0.2 ) tt = tt_sav + sec*fabs(ele_sav/2);
	    else                      tt = tt_sav + sec*fabs(ele_sav);
	  }
	}

	else{
  	  ele_sav = ele;
          tt_sav = tt;

	  double add = sec*fabs(ele);
	  if( fabs(ele) > 10 ) add *= 0.6;

//	  cout << prn << " add A " << setw(8) << add << endl;
          tt.add_sec(add);

	}
      }       
    }
  }

#ifdef DEBUG 
    t_map_sky_sat::iterator itPRN;
    for( itPRN = sat_view.begin(); itPRN != sat_view.end(); ++itPRN )
    {	 
      string prn = itPRN->first;

      map<t_gtime,t_gtime>::iterator itBEG;
      for( itBEG = sat_view[prn].begin(); itBEG != sat_view[prn].end(); ++itBEG )
      {
        cout << " PRN: " << prn 
	     << "   "    << itBEG->first.str_ymdhms()
	     << " - "    << itBEG->second.str_ymdhms()
	     << " duration: " << itBEG->second-itBEG->first
	     << endl;
      }  
      cout << endl;
    }
#endif

  return 1;
}

} // namespace
